#! /usr/bin/env python

from getopt import getopt
from sys import argv, stdin, stderr, stdout, exit
from types import StringTypes, IntType
from copy import deepcopy

# Parse options
options, arguments = getopt(argv[1:], "mkv")
ms_format = False
known_mrca = False
verbose = False
def parse_positiveint(s):
  i = int(s)
  if i < 0:
    raise ValueError, "Negative value encountered"
  return i
    
for o in options:
  if o[0][1] == 'm':
    ms_format = True
  elif o[0][1] == 'k':
    known_mrca = True
  elif o[0][1] == 'v':
    verbose = True

# Read data
if arguments != []:
  infile = open(arguments[0], 'r')
else:
  infile = stdin
data = []
if ms_format:
  while True:
    s = infile.readline()
    if s[:2] == "//":
      for i in range(3):
        s = infile.readline()
      while s != "" and s[:2] != "//":
        s = s.strip()
        if s != "":
          data.append(s)
          if len(data[-1]) != len(data[0]):
            stderr.write("Data lines are not of the same length\n")
            exit(1)
        s = infile.readline()
    elif s != "":
      continue
    break
else:
  # getline(f): returns next non-empty, non-comment line from f, None if EOF
  def getline(f):
    while True:
      s = f.readline()
      if s == "":
        return None
      # Strip off trailing comment
      escaped = 0
      for i in range(0, len(s)):
        if s[i] == '#' and not escaped:
          s = s[:i]
          break
        if s[i] == '\\':
          escaped = not escaped
      # Strip off trailing whitespace
      s = s.rstrip()
      if s != "":
        break
    return s

  count = 1
  while True:
    line = getline(infile)
    if line == None:
      break
    if line[0] == '>':
      # This line contains count of next sequence
      for i in xrange(1, len(line)):
        if line[i] != '>':
          count = parse_positiveint(line[i:])
        break
      else:
        # No count in line - reset
        count = 1
      continue
    data.append((line, count))
    count = 1
    if len(data[-1][0]) != len(data[0][0]):
      stderr.write("Data lines are not of the same length\n")
      exit(1)
  if arguments != []:
    infile.close()

# Remove duplicates, and check for symbols not in {0, 1}
data.sort()
i = 0
while i < len(data):
  if i > 0 and data[i][0] == data[i - 1][0]:
    # Previous copy of this sequence
    data[i - 1] = (data[i - 1][0], data[i - 1][1] + data[i][1])
    del data[i]
  else:
    for j in data[i][0]:
      if j not in "01":
        # Non-binary symbol encountered - delete sequence and issue warning
        stderr.write("WARNING: sequence '" + data[i][0] + "' contains non-binary symbols\n")
        del data[i]
        break
    else:
      # Kept this sequence - proceed with next one
      i += 1
if i == 0:
  # No sequences left in data set
  print "No sequences in data set; no ACs can be encountered"
  exit(0)
# Determine sequence length
n = len(data[0][0])

# Return iterator over legal values of r0, r1 given n0, n1
def I(n0, n1 = None):
  # If only one argument is present, assume it is a pair
  if n1 == None:
    n1 = n0[1]
    n0 = n0[0]
  class Iterator:
    def __init__(self, n0, n1):
      self.n0 = n0
      self.n1 = n1
      if known_mrca and n0 > 0:
        # We can never entirely get rid of 0s
        self.p = 1
        self.q = -1
      else:
        self.p = 0
        self.q = 0
      self.known_mrca = known_mrca
    def __iter__(self):
      return self
    def next(self):
      self.q += 1
      if self.q <= self.n1 and self.p <= self.n0 or self.q == 0:
        # We can always observe at least as many 1s as we initially had
        return (self.p, self.q)
      if not self.known_mrca and self.q == self.n1 + 1 and self.p == 0 and self.n0 > 0:
        # If we do not know the MRCA, an extra 1 can be obtained by mutation
        # of a 0 once all 0s have coalesced into one
        return (self.p, self.q)
      # We can never observe p 0s and q 1s
      self.p += 1
      self.q = -1
      if self.p <= self.n0 or (self.p == self.n0 + 1) and self.n1 > 0:
        return self.next()
      raise StopIteration
  return Iterator(n0, n1)

# Simple iterator for running through all possible ways to assign block[1]
# columns with columns sums block[0] and row sums bounded by r. A more
# efficient iterator would e.g. require the columns to be sorted and return
# the multiplicity of distinct permutations of the columns. The code using
# this iterator is prepared for such an extension, as the iterator is
# expected to return a list containing non-zero entries together with
# a multiplicity.
def resolve_block(block, r):
  class Iterator:
    def __init__(self, block, r):
      self.value = block[0]
      self.count = block[1]
      self.r = r
      self.n = []
      # Create first block
      for i in xrange(self.count):
        if self.value <= r[0]:
          # This column can be initialised to all 0s
          self.n.extend([self.value, 0, 0])
          r[0] -= self.value
        else:
          # Use remaining 0s available
          self.n.append(r[0])
          if self.value - r[0] <= r[1]:
            # This column can be initialised without *s
            self.n.extend([self.value - r[0], 0])
            r[1] -= (self.value - r[0])
          else:
            # Use remaining 1s available, and set rest to *s
            self.n.extend([r[1], self.value - r[0] - r[1]])
            r[2] -= (self.value - r[0] - r[1])
            r[1] = 0
          # Used remaining 0s
          r[0] = 0
      # Decrement this state such that first call to next will return it
      self.n[-3] += 1
      self.r[0] -= 1
      self.n[-2] -= 1
      self.r[1] += 1
    def __iter__(self):
      return self
    def next(self):
      for i in xrange(self.count - 1, -1, -1):
        if self.n[3 * i] > 0 and self.r[1] > 0:
          # Change a 0 to a 1
          self.n[3 * i] -= 1
          self.n[3 * i + 1] += 1
          self.r[0] += 1
          self.r[1] -= 1
          break
        elif self.r[2] > 0 and self.n[3 * i + 2] < self.value:
          # Increment the number of *s
          self.n[3 * i + 2] += 1
          self.r[2] -= 1
          self.r[0] += self.n[3 * i]
          self.r[1] += self.n[3 * i + 1]
          if self.value - self.n[3 * i + 2] <= self.r[0]:
            # Remaining symbols can be set to all 0s
            self.n[3 * i] = self.value - self.n[3 * i + 2]
            self.r[0] -= self.n[3 * i]
            self.n[3 * i + 1] = 0
          else:
            # Set as many symbols as possible to 0, the rest to 1
            self.n[3 * i] = self.r[0]
            self.n[3 * i + 1] = self.value - self.r[0] - self.n[3 * i + 2]
            self.r[0] = 0
            self.r[1] -= self.n[3 * i + 1]
          break
        else:
          # Cannot increment this column; reset and continue with next column
          for j in xrange(3):
            self.r[j] += self.n[3 * i + j]
            self.n[3 * i + j] = 0
      else:
        # We did not find a column that could be incremented further
        raise StopIteration
      # Initialise visited columns
      for j in xrange(i + 1, self.count):
        if self.value <= self.r[0]:
          # This column can be initialise to all 0s
          self.n[3 * j] = self.value
          self.r[0] -= self.value
        else:
          # Use remaining 0s available
          self.n[3 * j] = self.r[0]
          if self.value - self.r[0] <= self.r[1]:
            # This column can be initialised without *s
            self.n[3 * j + 1] = self.value - self.r[0]
            self.r[1] -= (self.value - self.r[0])
          else:
            # Use remaining 1s available, and set rest to *s
            self.n[3 * j + 1] = self.r[1]
            self.n[3 * j + 2] = self.value - self.r[0] - self.r[1]
            self.r[2] -= (self.value - self.r[0] - self.r[1])
            self.r[1] = 0
          # Used remaining 0s
          self.r[0] = 0
      # Return sorted list of non-zero entries in n
      n = deepcopy(self.n)
      n.sort()
      i = 0
      while n[i] == 0:
        i += 1
      return (tuple(n[i:]), 1)
  return Iterator(block, r)

# Extend newtable with columns with column sums taken from table and remaining
# row sums taken from r. When newtable has been finalised, newtables is updated
# with the final multiplicity of newtable, with m being the current
# multiplicity of newtable.
def build_table(table, newtable, r, m, newtables):
  if table == ():
    # Reached bottom of recursion
    newtable = deepcopy(newtable)
    newtable.sort()
    i = 0
    while newtable[i] == 0:
      i += 1
    while i < len(newtable):
      j = i + 1
      while j < len(newtable) and newtable[i] == newtable[j]:
        j += 1
      table += ((newtable[i], j - i),)
      i = j
    if newtables.has_key(table):
      newtables[table] += m
    else:
      newtables[table] = m
  else:
    # Handle next block of column sums
    tmpr = deepcopy(r)
    for i in resolve_block(table[0], r):
      build_table(table[1:], newtable + list(i[0]), r, m * i[1], newtables)

# Count the number of 0s and 1s in column i of data
def counts(data, i):
  c = [0, 0]
  for s in data:
    c[int(s[0][i])] += s[1]
  return c

# Initialise contingency tables for first site
tables = {}
# Run through all allowed counts of 0s and 1s
for r in I(counts(data, 0)):
  r = list(r)
  r.sort()
  if r[0] == 0:
    # Ignore zero entries
    r = ((r[1], 1),)
  elif r[0] == r[1]:
    r = ((r[0], 2),)
  else:
    r = [(i, 1) for i in r]
    r = tuple(r)
  if tables.has_key(r):
    tables[r] += 1
  else:
    tables[r] = 1

# Now extend column by column
for i in xrange(1, len(data[0][0])):
  newtables = {}
  # Run through all column sums from previous site
  for table in tables.items():
    table, m = table[0], table[1]
    c = reduce(lambda x, y: x + y[0] * y[1], table, 0)
    # Run through all possible counts of 0s and 1s for this site
    for r in I(counts(data, i)):
      r = list(r)
      for rstar in xrange(max(0, c - sum(r)), c + 1):
        # Start by handling the star column
        cstar = sum(r) + rstar - c
        for j in xrange(max(0, cstar - r[1]), min(r[0], cstar) + 1):
          newtable = [j, cstar - j]
          tmpr = [r[0] - j, r[1] - cstar + j, rstar]
          # Now run through all normal columns in contingency table
          build_table(table, newtable, tmpr, m, newtables)
  tables = newtables

print sum(tables.values())
