/* cob.c: Implementation of a branch and bound based algorithm to
 * generate equation system for the probability of a SNP data set
 * under coalescent, infinite sites model of mutations and a mid-point
 * model of recombinations, where only ancestral states that can be
 * visited in a history with a bounded number of recombinations are
 * considered. The space of ancestral configurations is searched in a
 * breadth first manner, and the equation system is stored on disk.
 *
 * Rune Lyngsoe, March 2006
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <limits.h>
#include <math.h>
#include <string.h>

#include "enumerate.h"
#include "elist.h"
#include "common.h"
#include "bounds.h"
#include "exact.h"
#include "bitfunctions.h"
#include "mergesort.h"
#include "topological_sort.h"
#include "hashtable.h"
#include "umfpack.h"

/* Select method for computing a lower bound on the number of recombinations */
#ifdef GREVEN_HAPLOTYPE
#define LOWERBOUND(g) haplotype_bound_genes(g)
#elif defined(GREVEN_HAPLOTYPEHEURISTIC)
#define LOWERBOUND(g) haplotype_heuristic_genes(g, INT_MAX, INT_MAX, 1)
#else
#define LOWERBOUND(g) hudson_kaplan_genes(g)
#endif

/* We need to store an index as well as number of recombinations for each AC */
typedef struct _IntPair {
  int index;
  int value;
} IntPair;
static int current_slice;

/* A term in a linear equation is a coefficient a, possibly times a
 * variable (with -1 indicating that the coefficient should be added
 * directly). Symbolic terms are used when equation system has to be
 * solved for more than one theta,rho pair.
 */
typedef struct _Term {
  int variable;
  double a;
} Term;
typedef enum { CONSTANT, THETA, RHO } TermType;
typedef struct _SymbolicTerm {
  int variable;
  TermType type;
  union {
    int i;
    double x;
  } e;
  int d;
  double dtheta;
  double drho;
} SymbolicTerm;

/* The search tactic for MLE is determined by the range and
 * granularity of the grid for which likelihoods are computed, and
 * stopping criteria.
 */
typedef struct _MLEExpression {
  double thetamin;
  double thetamax;
  unsigned int thetan;
  double rhomin;
  double rhomax;
  unsigned int rhon;
  unsigned int maxiter;
  double ldiff;
  double logldiff;
} MLEExpression;

/* Segregating sites in data set */
static LList *positions;
/* Variables specifying level of verbosity and method for computing
 * weight of recombinations, whether a loop free equation system
 * should be constructed, and collecting the total number of reachable
 * ACs.
 */
static int vlevel = 0, aprime = 0, noloops = 0, n_ac, n_components,
  max_component;
/* Evolutionary parameters */
static EList *thetas, *rhos, *pairs, *mle;
static double theta, rho;
/* Should equation system be solved for more than one data point */
static int multiple_dps = 0;
/* Are the shortcuts heuristics for quick and dirty computation in
 * original model or an alternative model?
 */
typedef enum { _HEURISTIC, _FULLY_MODEL, _STATES_MODEL, _SITES_MODEL }
  _View;
static _View _model_view = _HEURISTIC;
/* Arrays holding positions of segregating sites and mid-points */
static double *s, *r;

static void print_usage(FILE *f, char *name)
{
  int linelength = 80;

  fprintf(f, "Usage: %s [options] [input]\n", name);
  pretty_print(f, "The program reads data from the input file specified (or from standard input if no input file is specified) and enumerates the ancestral states that can be visited in histories with at most a specified number of recombinations, where the default is to enumerate only the ancestral states that can be visited in histories with a minimum number of recombinations. A reduced equation system for data probability, only including the enumerated ancestral states, is then constructed and solved.", linelength, 0);
  pretty_print(f, "Legal options are:", linelength, 0);
  print_option(f, "-cN", "Count states visitable in histories with at most N recombinations", linelength, -1);
  print_option(f, "-rN", "Count states visitable with at most N more recombinations than the minimum number of recombinations required for the data", linelength, -1);
  print_option(f, "-k", "Assume that the common ancestral sequence is known, i.e. that we know which is the wild type and which is the mutant in each site. In binary format the all-0 sequence is assumed to be the common ancestral sequence, while the common ancestral sequence has to be specified directly for amino acid and nucleic sequences (see options -a and -n)", 80, -1);
  print_option(f, "-o", "Assume input data is in own format (default is to first try to parse data in own format, and if that fails to try to parse it in fasta format).", linelength, -1);
  print_option(f, "-f", "Assume input data is in fasta format.", linelength, -1);
  print_option(f, "-a", "Assume input consists of amino acid (protein) sequences using the one letter amino acid alphabet (default is to assume sequences in binary, i.e. 0/1, format). If the most recent common ancestor is assumed known, the first sequence in the input data is considered to specify the wild type of each site and is not included as part of the sample set.", linelength, -1);
  print_option(f, "-n", "Assume input consists of nucleic sequences; anything but a/c/g/t/u is considered an unresolved site. If the most recent common ancestor is assumed known, the first sequence in the input data is considered to specify the wild type of each site and is not included as part of the sample set.", linelength, -1);
  print_option(f, "-eF,G", "Set positions of left and right end points of sequences to F and G, respectively. Default is to use positions of first and last site in data as left and right end points, and if F and/or G is omitted they will default to these positions.", linelength, -1);
  print_option(f, "-i[F]", "Rescale sequence to interval of length F. If no length is specified the sequence will be rescaled to the unit interval, i.e. positions are rescaled so all are points between 0 and 1.", linelength, -1);
  print_option(f, "-tF", "Set scaled mutation rate to theta = F. If no value is specified, Watterson's estimate is used. If more than one value is specified, the equation system is solved with each value.", linelength, -1);
  print_option(f, "-sF", "Set scaled recombination rate to rho = F. If no value is specified, a default value of 1.0 is used. If more than one value is specified, the equation system is solved with each value. Combined with multiple theta values this allows finding probabilities for a grid of theta and rho values.", linelength, -1);
  print_option(f, "-pF,G", "Set scaled mutation rate to theta = F and scaled recombination rate to rho = G. If more than one value is specified, the equation system is solved with each pair of values. This allows finding probabilities for pairs of theta of rho values not necessarily forming a grid, compare with -t and -s options. If used in conjunction with -t and/or -s options, both the grid specified by these and the pairs specified by this option are used.", linelength, -1);
  print_option(f, "-TL", "Specify a list of theta values. L should be of the form t1,t2,...,tn where t1, t2,..., tn are the theta values and at least one theta value is specified. Equivalent to -tt1 -tt2 ... -ttn.", linelength, -1);
  print_option(f, "-SL", "Specify a list of rho values. L should be of the form r1,r2,...,rn where r1, r2,..., rn are the rho values and at least one rho value is specified. Equivalent to -sr1 -sr2 ... -srn.", linelength, -1);
  print_option(f, "-PL", "Specify a list of pairs of theta, rho values. L should be of the form t1,r1:t2,r2:...:tn,rn where t1,r1,..., tn,rn are the theta, rho pairs and at least one pair is specified. Equivalent to -pt1,r1 -pt2,r2 ... -ptn,rn.", linelength, -1);
  print_option(f, "-mE", "Determine maximum likelihood estimate of theta and rho. Initially the MLE is found as the theta, rho pair in a prespecified grid attaining the highest likelihood. This procedure is iterated for the subrange of theta and rho values bounded by the neighbours to the MLE until the convergence criteria specified is met. The search is determined by the expression E that should be on the form t1-t2,tn:r1-r2,rn:n:d:ld where\n - t1 and t2 specifies the allowed range for theta and tn the granularity of the grid on the theta axis (tn is the number of different theta values used in an iteration - hence tn is required to be at least 1, 2 if t1 and t2 are not equal, and iteration only makes sense if tn is at least 4)\n - r1 and r2 specifies the allowed range for rho and rn the granularity of the grid on the rho axis\n - n specifies the number of iterations\n - d specifies that iteration should be terminated if the maximum difference in likelihood between the current MLE and its eight grid neighbours is at most d\n - ld specifies that iteration should be terminated if the maximum difference in loglikelihood between the current MLE and its eight grid neighbours is at most ld\nIf t1 is omitted the range of theta values is taken to be from 0 to t2 times Watterson's estimate. If more than one stopping criteria is specified, iteration stops as soon as one of the criteria is met. If -m occurs more than once, a MLE is determined according to each specification, in turn.", linelength, -1);
  print_option(f, "-y", "Compute a as the fraction of sites carrying ancestral material, i.e. ignoring the segregating sites positions.", linelength, -1);
  print_option(f, "-l", "Set up a loop free equation system by not allowing two sequences to coalesce unless they have at least one site with ancestral material in common.", linelength, -1);
  print_option(f, "-z[s|l]", "The approximations of only including some ancestral states in the equation system constructed and if the -l option is specified of only allowing sequences to coalesce if they share at least one site with ancestral material, can be viewed as either a computational approximation (heuristic) or a modelling approximation. The default view is as a computational approximation, where the data probability computed is guaranteed to lower bound the true data probability in the infinite sites mutation/finite sites recombination model. With the modelling approximation view, the normalising Z factor in the equations should be modified to reflect that some events are not possible. This option invokes this modified computation of the normalising Z factor. For added control, an optional single character argument to this option allows the modelling view to being applied only to the ancestral state space reduction (-zs) or only to the sharing of ancestral sites requirement (-zl).", linelength, -1);
  print_option(f, "-v[N]", "Set verbosity level to N. The levels are\n0 - only output probability of data set; if probability is computed for more than one pair of theta, rho values, each probability is followed by the theta and rho value it was based on\n1 - furthermore output number of ancestral configurations and stages in computation\n2 - furthermore output detailed stage information\n3 - furthermore output probabilities of all ancestral configurations\n4 - deluge of information\nIf N is omitted verbosity level is set to 2. The default verbosity level when this option is not used is 0.", linelength, -1);
  print_option(f, "-h, -H", "Print this information and stop.", linelength, -1);}

/* Check whether g is a novel ancestral state that can be explained
 * with at most n recombinations. If this is the case, insert it in
 * the right slice of states so branches leading back in time from g
 * will be pursued. Otherwise g is deallocated.
 */
static void check_ancestralstate(Genes *g, int n, HashTable *t,
				 HashTable *states, EList *slice, int *index)
{
  IntPair *lookup;
  int am = ancestral_material(g);
#ifdef ENABLE_VERBOSE
  int v = verbose();
  set_verbose(v - 1);
#endif
  PackedGenes *rep, *p = pack_genes(g);

  if ((rep = hashtable_lookuprepresentative((void *)p, states + am - g->length,
					    (void **)&lookup)) == NULL){
    /* We reached a novel ancestral state - determine whether we
     * can extend the history without using more than n further
     * recombinations.
     */
    if (beagle_reusable_bounded(g, NULL, 0, n, t) > n)
      /* Too many recombinations required to visit this configuration */
      free_packedgenes(p);
    else{
      /* The ancestral state g is visitable */
      lookup = (IntPair *)xmalloc(sizeof(IntPair));
      lookup->index = (*index)++;
      lookup->value = n;
      hashtable_insert((void *)p, (void *)lookup, states + am - g->length);
      elist_append(slice + am - g->length, (void *)p);
#ifdef ENABLE_VERBOSE
      if (v){
	printf("New visitable ancestral state:\n");
	output_genes_indexed(g, NULL);
      }
#endif
    }
    free_genes(g);
  }
  else{
    /* This state has been encountered previously */
    if (n > lookup->value){
      /* But with fewer recombinations allowed */
      lookup->value = n;
      if (am == current_slice)
	/* We may already have handled this state - insert it for rehandling */
	elist_append(slice + am - g->length, (void *)rep);
    }
    free_genes(g);
    free_packedgenes(p);
  }

#ifdef ENABLE_VERBOSE
  set_verbose(v);
#endif
}

/* The recursion actually enumerating ancestral states in histories
 * ending with g and going back in time using at most n
 * recombinations; lower is a guaranteed lower bound on the number of
 * recombinations required to explain g, t is an auxiliary hash table
 * used for determining the minimum number of recombinations required
 * for a given ancestral state, and states is the array of hash tables
 * the enumerated ancestral states are stored in.
 */
static void handle_ancestralstate
(Genes *g, int n, HashTable *t, HashTable *states, EList *slice, int *index)
{
  int i, j;
  Genes *h;
  EList *forced;

  /* Try all possible coalesces */
  for (i = 0; i < g->n - 1; i++)
    for (j = i + 1; j < g->n; j++)
      if (compatible(g, i, j)){
	h = copy_genes(g);
	coalesce(h, i, j);
#ifdef ENABLE_VERBOSE
	if (verbose() > 1){
	  printf("Coalescing sequences %d and %d", i, j);
	  output_genes_indexed(h, NULL);
	}
#endif
	if (!noloops || pairwise_am_overlap(g, i, j)
	    || (_model_view == _STATES_MODEL))
	  check_ancestralstate(h, n, t, states, slice, index);
      }

  /* Try all possible mutations */
#ifdef ENABLE_VERBOSE
  set_verbose(verbose() - 1);
#endif
  forced = force_mutation(g, NULL);
#ifdef ENABLE_VERBOSE
  set_verbose(verbose() + 1);
#endif
  for (i = 0; i < elist_length(forced); i++){
    h = elist_get(forced, i);
    check_ancestralstate(h, n, t, states, slice, index);
  }
  elist_destroy(forced);

  if (n > 0)
    /* Try all possible recombinations */
    for (i = 0; i < g->n; i++){
#ifdef ENABLE_VERBOSE
      set_verbose(verbose() - 1);
#endif
      forced = force_split(g, i, NULL);
#ifdef ENABLE_VERBOSE
      set_verbose(verbose() + 1);
#endif
      for (j = 0; j < elist_length(forced); j++){
	h = elist_get(forced, j);
	check_ancestralstate(h, n - 1, t, states, slice, index);
      }
      elist_destroy(forced);
    }
}

/* Allocate and initialise a term of the right type according to
 * multiple_dps, and insert it into elist; v is the index of the
 * variable, a is coefficient (when it is known, i.e. when system is
 * only solved for one set of parameters), i or x is multiplicative
 * factor, d, dt, and dr are multiplicative factors determining
 * normalising constant (i or x, d, dt, and dr are necessary for
 * computing coefficient of variable when multiple theta, rho values
 * need to be considered), and type is the type of term.
 */
static void *allocate_term(int v, double a, int i, double x, int d, double dt,
			   double dr, TermType type, EList *elist)
{
  Term *term;
  SymbolicTerm *sterm;

  if (multiple_dps){
    sterm = (SymbolicTerm *)xcalloc(1, sizeof(SymbolicTerm));
    sterm->variable = v;
    sterm->type = type;
    if (type == CONSTANT)
      sterm->e.i = i;
    else
      sterm->e.x = x;
    sterm->d = d;
    sterm->dtheta = dt;
    sterm->drho = dr;
    elist_append(elist, sterm);
    return sterm;
  }
  else{
    term = (Term *)xmalloc(sizeof(Term));
    term->variable = v;
    term->a = a;
    elist_append(elist, term);
    return term;
  }
}

/* Determine immediate predecessors of ancestral state g that are also
 * present in hash table states. Determine the corresponding terms of
 * the equation system and insert them into terms.
 */
static void find_equations(PackedGenes *state, IntPair *value, va_list args)
{
  HashTable *states = va_arg(args, HashTable *);
  EList *terms = va_arg(args, EList *);
  int i, j, k, l, m, q, am;
  double Z, a, b;
  EList *sites;
  Genes *h, *g = unpack_genes((PackedGenes *)state);
  IntPair *lookup;
  PackedGenes *p;

  if (vlevel > 3){
    printf("Ancestral state %d:\n", value->index + 1);
    output_genes_indexed(g, NULL);
  }

  /* Insert marker in terms to signal new variable */
  /* Variable occurs with coefficient -1 in own equation */
  allocate_term(value->index, -1, -1, 0, 1, 0, 0, CONSTANT, terms);

  if (ancestral_material(g) == g->length){
    /* Reached an MRCA */
    j = 1;
    for (i = 2; i <= g->n; i++)
      j *= i;
    allocate_term(-1, j, 1, j, 1, 0, 0, CONSTANT, terms);
    free_genes(g);
    return;
  }

  /* Compute normalising factor */
  /* Start by computing normalising factor for when restrictions are
   * purely for computational efficiency - all pairs of compatible
   * sequences are allowed to coalesce.
   */
  Z = g->n * (g->n - 1);
  if (aprime)
    a = (r[g->length] - r[0])
      * (double)ancestral_material(g) / (double)g->length;
  else
    a = 0;
  b = 0;
  for (i = 0; i < g->n; i += k){
    /* Determine number of sequences of this type */
    for (k = 1; (i + k < g->n) && (compare_sequences(g, i, i + k) == 0);
	 k++);
    /* Get list of sites carrying ancestral material in this type */
    sites = ancestral_sites(g, i);
    /* Tally this types contribution to a and b */
    b += k * (s[(int)elist_get(sites, elist_length(sites) - 1)]
	      - s[(int)elist_get(sites, 0)]);
    if (!aprime)
      /* Compute contribution to a as sum of weights of intervals
       * carrying ancestral material.
       */
      for (j = 0; j < elist_length(sites); j = l){
	l = j + 1;
	/* Find maximal interval of contiguous ancestral material
	 * starting at ancestral site number j.
	 */
	while ((l < elist_length(sites)) &&
	       ((int)elist_get(sites, l) == (int)elist_get(sites, l - 1) + 1))
	  l++;
	/* Add contribution from this interval to a */
	a += k * (r[(int)elist_get(sites, l - 1) + 1]
		  - r[(int)elist_get(sites, j)]);
      }

    if (_model_view != _HEURISTIC){
      /* Actually (some of) the restrictions are considered to be of a
       * modelling nature - modify Z and b to reflect this.
       */
      /* Check for coalescence events disallowed by modelling approximations */
      for (j = i + k; j < g->n; j += l){
	/* Determine number of sequences of this type */
	for (l = 1; (j + l < g->n) && (compare_sequences(g, j, j + l) == 0);
	     l++);
	if (!compatible(g, i, j))
	  /* Incompatible pair of sequences so restrictions have no effect */
	  continue;
	if ((_model_view != _STATES_MODEL) && !pairwise_am_overlap(g, i, j)){
	  /* Coalescence disallowed by ancestral material overlap requirement,
	   * and this is considered a modelling approximation.
	   */
	  Z -= k * l;
	}
	else if (_model_view != _SITES_MODEL){
	  /* Restriction to ancestral states in (near) minimal ARG is
	   * considered a modelling approximation.
	   */
	  h = copy_genes(g);
	  coalesce(h, i, j);
	  p = pack_genes(h);
	  if (!hashtable_lookup(p, states + ancestral_material(h) - h->length,
				NULL))
	    /* And this restriction has been invoked for this pair of
	     * sequences.
	     */
	    Z -= k * l;
	  free_genes(h);
	  free_packedgenes(p);
	}
      }
      /* Check for recombination events disallowed by modelling assumptions */
      if (_model_view != _SITES_MODEL){
	for (j = 1; j < elist_length(sites); j++){
	  h = copy_genes(g);
	  split(h, i, (int)elist_get(sites, j));
	  p = pack_genes(h);
	  if (!hashtable_lookup(p, states + ancestral_material(h) - h->length,
				NULL))
	    b -= k *
	      (s[(int)elist_get(sites, j)] - s[(int)elist_get(sites, j - 1)]);
	  free_genes(h);
	  free_packedgenes(p);
	}
      }
    }
    elist_destroy(sites);
  }

  if (!multiple_dps){
    Z += a * theta + b * rho;
    if (vlevel > 3)
      printf(" a = %lf, b = %lf, Z = %lf\n", a, b, Z);
  }
  else if (vlevel > 3)
    printf(" a = %lf, b = %lf, n = %d\n", a, b, g->n);

  /* Check for mutations */
  h = copy_genes(g);
  /* All states reachable by mutation carry same amount of ancestral
   * material.
   */
  am = ancestral_material(h) - h->length;
  for (i = 0; i < g->length; i++){
    /* If ancestral sequence is known, only mutate 1s to 0s; if
     * ancestral sequence is not known, mutate 0s to 1s and 1s to
     * 0s
     */
    for (k = 0; k <= (gene_knownancestor ? 0 : 1); k++)
      if ((j = mutate(h, i, k)) >= 0){
	p = pack_genes(h);
	if (hashtable_lookup(p, states + am , (void **)&lookup)){
	  if (vlevel > 3)
	    printf(" Ancestral state %d reachable by mutation of site %s in sequence %d\n", lookup->index + 1, (char *)GetByIndex(positions, i), j + 1);
	  /* Determine number of sequences identical to the mutated sequence */
	  m = 0;
	  for (l = 0; l < h->n; l++)
	    if (identical(h, j, l))
	      m++;
	  /* Insert entry into list of terms */
	  allocate_term(lookup->index, (r[h->length] - r[0]) * m * theta / Z,
			0, (r[h->length] - r[0]) * m, Z, a, b, THETA, terms);
	}
	free_packedgenes(p);
	free_genes(h);
	h = copy_genes(g);
      }
  }

  /* Check for coalesces */
  for (i = 0; i < g->n - 1; i++){
    /* Is this the first encounter of sequence i? */
    /* Genes unpacked from canonical representation are sorted */
    if ((i > 0) && (identical(h, i, i - 1)))
      continue;

    /* Determine number of identical sequences */
    for (m = 0; (i + 1 + m < g->n) && identical(h, i, i + 1 + m); m++);
    if (m > 0){
      coalesce(h, i, i + 1);
      p = pack_genes(h);
      am = ancestral_material(h);
      if (hashtable_lookup(p, states + am - h->length, (void **)&lookup)){
	/* More than one occurrence of this sequence */
	allocate_term(lookup->index, g->n * m / Z, g->n * m, 0, Z, a, b,
		      CONSTANT, terms);
	if (vlevel > 3)
	  printf(" Ancestral state %d reachable by coalescence of sequence %d with one of %d\n identical sequences\n", lookup->index + 1, i + 1, m);
      }
      free_packedgenes(p);
      free_genes(h);
      h = copy_genes(g);
    }

    /* Check for coalesces with compatible but non-identical sequences */
    for (j = i + 1 + m; j < g->n; j++)
      if ((compatible(g, i, j)) && (!noloops || pairwise_am_overlap(g, i, j))){
	/* Found compatible and non-identical sequence - check for
	 * prior occurrences of this sequence.
	 */
	if (identical(g, j - 1, j))
	  continue;
	/* First occurrence of sequence j - compute entry for this
	 * coalescence.
	 */
	coalesce(h, i, j);
	am = ancestral_material(h);
	p = pack_genes(h);
	if (hashtable_lookup(p, states + am - h->length, (void **)&lookup)){
	  /* Determine number of sequences in h that are identical to
	   * the result of the coalescence.
	   */
	  m = 0;
	  for (l = 0; l < h->n; l++)
	    if (identical(h, i, l))
	      m++;
	  /* Insert entry in list of terms */
	  allocate_term(lookup->index, 2 * g->n * m / Z, 2 * g->n * m, 0, Z,
			a, b, CONSTANT, terms);
	  if (vlevel > 3)
	    printf(" Ancestral state %d reachable by coalescence of sequences %d and %d\n (%d possible coalescent sequences)\n", lookup->index + 1, i + 1, j + 1, m);
	}
	free_packedgenes(p);
	free_genes(h);
	h = copy_genes(g);
      }
  }

  /* Check recombinations */
  /* Recombinations do not change the amount of ancestral material */
  am = ancestral_material(g) - g->length;
  for (i = 0; i < g->n; i++){
    /* Skip forward to last occurrence of sequence i */
    for (; (i < g->n - 1) && identical(g, i, i + 1); i++);
    j = next_ancestral(g, i, -1);
    while ((l = next_ancestral(g, i, j)) != -1){
      /* Check recombination between sites j and l in sequence i */
      split(h, i, l);
      p = pack_genes(h);
      if (hashtable_lookup(p, states + am, (void **)&lookup)){
	if (vlevel > 3)
	  printf(" Ancestral state %d reachable by split between %s and %s in sequence %d\n", lookup->index + 1, (char *)GetByIndex(positions, j), (char *)GetByIndex(positions, l), i + 1);
	/* Determine multiplicity of split sequences in h */
	k = q = 0;
	for (m = 0; m < h->n; m++){
	  if (identical(h, i, m))
	    k++;
	  if (identical(h, h->n - 1, m))
	    q++;
	}
	/* Insert entry in list of terms */
	allocate_term(lookup->index,
		      b * rho * ((s[l] - s[j]) / (r[g->length] - r[0]))
		      * k * q / (g->n * h->n * Z), 0,
		      b * ((s[l] - s[j]) / (r[g->length] - r[0]))
		      * k * q / (g->n * h->n), Z, a, b, RHO, terms);
      }
      free_packedgenes(p);
      free_genes(h);
      h = copy_genes(g);
      j = l;
    }
  }

  /* Clean up */
  free_genes(h);
  free_genes(g);
}

/* Find the index of a in the sorted array l of n integers; return -1
 * if a is not in l.
 */
static int get_index(int a, int n, int *l)
{
  int i = 0,
    j = n, k;

  if (n == 0)
    /* No elements in array */
    return -1;

  while (i + 1 < j){
    k = (i + j) / 2;
    if (l[k] > a)
      j = k;
    else
      i = k;
  }

  if (l[i] == a)
    return i;
  else
    return -1;
}

/* Check whether variable occurs in indeces before current_slice; if
 * so, add a times the corresponding value in x to the value in bx.
 */
static int known_value(double *bx, double a, int variable, int *slice_size,
		       double **x, int ** indeces, int current_slice,
		       int gene_length)
{
  int i, j;

  for (i = (current_slice < 2 * gene_length ? 0
	    : current_slice - 2 * gene_length);
       i < current_slice - gene_length; i++)
    if ((j = get_index(variable, slice_size[i], indeces[i])) >= 0){
      /* Found variable */
      *bx -= a * x[i][j];
      return 1;
    }

  /* Did not find variable */
  return 0;
}

/* Enumerate the different ancestral states visitable in histories
 * with at most n recombinations.
 */
static void enumerate_states(Genes *genes, int bound, HashTable *t,
			     FILE **storage,
			     void (*action)(HashTable *, HashTable *, FILE *))
{
  HashTable *states;
  EList *slice;
  int i, current_slice, table_size, index = 0,
    am = ancestral_material(genes);
  double time_point;
  Genes *g;
  IntPair *lookup;
  PackedGenes *p;

  /* Determine sizes of and allocate hash tables and elists */
  table_size = msb(am) + bound;
  states = (HashTable *)xmalloc((am - genes->length + 1) * sizeof(HashTable));
  slice = (EList *)xmalloc((am - genes->length + 1) * sizeof(EList));
  /* Prepare first genes->length + 1 hash tables and elists */
  for (i = (am < 2 * genes->length ? 0 : am - 2 * genes->length);
       i <= am - genes->length; i++){
    init_packedgeneshashtable(states + i, table_size);
    elist_init(slice + i);
  }

  /* Enumerate all ancestral states visitable with at most bound
   * recombinations.
   */
  /* Start by inserting data set as reachable ancestral configuration */
  check_ancestralstate(copy_genes(genes), bound, t, states, slice, &index);
  /* Now run through reachable ancestral configurations in order of
   * decreasing ancestral material content.
   */
  if (vlevel == 1){
    time_point = umfpack_timer();
    printf("Enumerating ancestral configurations");
    fflush(stdout);
  }
  for (current_slice = am; current_slice >= genes->length; current_slice--){
    if (vlevel > 1){
      time_point = umfpack_timer();
      printf("Enumerating configurations with %d ancestral sites\n",
	     current_slice);
      fflush(stdout);
    }
    for (i = 0; i < elist_length(slice + current_slice - genes->length); i++){
      p	= elist_get(slice + current_slice - genes->length, i);
      hashtable_lookup((void *)p, states + current_slice - genes->length,
		       (void **)&lookup);
      g = unpack_genes(p);
      handle_ancestralstate(g, lookup->value, t, states, slice, &index);
      free_genes(g);
    }

    /* Perform action on the slice just handled */
    action(states + current_slice - genes->length, states,
	   storage[current_slice - genes->length]);

    if (vlevel > 1){
      printf(" (found %d) took %.1f seconds\n", hashtable_size(states + current_slice - genes->length), umfpack_timer() - time_point);
      fflush(stdout);
    }

    /* Free the hash table and elist used for the slice just handled */
    n_ac += hashtable_size(states + current_slice - genes->length);
    hashtable_free(states + current_slice - genes->length,
		   (void (*)(void *))free_packedgenes,
		   (void (*)(void *))free, (void (*)(void *))free);
    elist_free(slice + current_slice - genes->length);
    if (current_slice > 2 * genes->length){
      /* Prepare next hash table and elist */
      init_packedgeneshashtable
	(states + current_slice - 2 * genes->length - 1, table_size);
      elist_init(slice + current_slice - 2 * genes->length - 1);
    }
  }
  if (vlevel == 1){
    printf(" took %.1f seconds\n", umfpack_timer() - time_point);
    fflush(stdout);
  }
  /* Clean up */
  free(states);
  free(slice);
}

static void build_equations(HashTable *current, HashTable *states,
			    FILE *storage)
{
  EList *terms = elist_make();
  int i;
  void *term;
  size_t termsize;

  /* Determine contributions to equation system from the slice just
   * handled.
   */
  hashtable_map(current, (void (*)(void *, void *, va_list))find_equations,
		states, terms);
  /* Write these to disk */
  if (multiple_dps)
    termsize = sizeof(SymbolicTerm);
  else
    termsize = sizeof(Term);
  fwrite(&(terms->count), sizeof(int), 1, storage);
  for (i = 0; i < elist_length(terms); i++){
    term = elist_get(terms, i);
    fwrite(term, termsize, 1, storage);
    free(term);
  }

  /* Clean up */
  elist_destroy(terms);
}

static void output_probability(PackedGenes *p, IntPair *value, va_list args)
{
  int n = va_arg(args, int);
  double *x = va_arg(args, double *);
  int *indeces = va_arg(args, int *);
  Genes *g;

  printf("Ancestral state %d has likelihood %lE:\n", value->index,
	 x[get_index(value->index, n, indeces)]);
  g = unpack_genes(p);
  output_genes_indexed(g, NULL);
  free_genes(g);
}

static void output_probabilities(HashTable *current, HashTable *states,
				 FILE *storage)
{
  int n, *indeces;
  double *x;

  if (storage == NULL)
    /* No configurations in current slice */
    return;

  /* Read in probabilities for current slice */
  fseek(storage, 0, SEEK_SET);
  fread(&n, sizeof(int), 1, storage);
  if (n > 0){
    x = (double *)xmalloc(n * sizeof(double));
    fread(x, sizeof(double), n, storage);
    indeces = (int *)xmalloc(n * sizeof(int));
    fread(indeces, sizeof(int), n, storage);

    /* Output probabilities */
    hashtable_map(current,
		  (void (*)(void *, void *, va_list))output_probability,
		  n, x, indeces);
    /* Clean up */
    free(x);
    free(indeces);
  }

  fclose(storage);
}

static SymbolicTerm *sterm;
static Term *term;
static int compare_indeces(int *i, int *j)
{
  if (multiple_dps)
    return (sterm[*i].variable < sterm[*j].variable);
  else
    return (term[*i].variable < term[*j].variable);
}

static int compare_terms(Term *a, Term *b)
{
  return (a->variable < b->variable);
}

static int compare_sterms(SymbolicTerm *a, SymbolicTerm *b)
{
  return (a->variable < b->variable);
}

static double solve_equations(Genes *genes, FILE **configurations,
			      FILE **probabilities)
{
  int i, j, k, m, n, var, current_slice, *slice_size, **indeces, *variable,
    *size, *new, *old, AAindex, *Ap, *Ai, *AAp, *AAi, slice_max,
    am = ancestral_material(genes);
  double a, *Ax, *bx, **x, *AAx, *bbx, *xx, time_point, avg_comp;
  void *symbolic, *numeric;

  slice_size = (int *)xmalloc((am - genes->length + 1) * sizeof(int));
  indeces = (int **)xmalloc((am - genes->length + 1) * sizeof(int *));
  x = (double **)xmalloc((am - genes->length + 1) * sizeof(double *));
  if (vlevel > 0)
    n_components = max_component = 0;

  /* Solve equation system for current theta, rho pair */
  for (current_slice = genes->length; current_slice <= am; current_slice++){
    /* Read terms for this slice */
    fseek(configurations[current_slice - genes->length], 0, SEEK_SET);
    fread(&n, sizeof(int), 1, configurations[current_slice - genes->length]);
    if (n == 0){
      /* No variables in this slice */
      slice_size[current_slice - genes->length] = 0;
      if (!multiple_dps)
	fclose(configurations[current_slice - genes->length]);
      if (vlevel > 2)
	probabilities[current_slice - genes->length] = NULL;
      continue;
    }
    if (vlevel > 1){
      time_point = umfpack_timer();
      printf("Solving equation systems for configurations with %d ancestral sites\n", current_slice);
      fflush(stdout);
      /* If detailed probabilities of all reachable ancestral
       * configurations is required, store probabilities on disk.
       */
      if (vlevel > 2){
	if ((probabilities[current_slice - genes->length] = (FILE *)tmpfile())
	    == NULL){
	  fprintf(stderr,
		  "Could not open temporary file for storing probabilities\n");
	  exit(1);
	}
      }
    }
    if (multiple_dps){
      sterm = (SymbolicTerm *)xmalloc(n * sizeof(SymbolicTerm));
      fread(sterm, sizeof(SymbolicTerm), n,
	    configurations[current_slice - genes->length]);
    }
    else{
      term = (Term *)xmalloc(n * sizeof(Term));
      fread(term, sizeof(Term), n,
	    configurations[current_slice - genes->length]);
      fclose(configurations[current_slice - genes->length]);
    }
    /* Determine number of variables in this slice */
    m = 0;
    for (i = 0; i < n; i++)
      if ((multiple_dps && (sterm[i].type == CONSTANT) && (sterm[i].e.i < 0))
	  || (!multiple_dps && (term[i].a < 0)))
	m++;
    slice_size[current_slice - genes->length] = m;
    
    /* Determine variables in this slice and sort them according to index */
    variable = (int *)xmalloc((m + 1) * sizeof(int));
    size = (int *)xmalloc(n * sizeof(int));
    j = 1;
    variable[0] = 0;
    for (i = 1; i < n; i++)
      if ((multiple_dps && (sterm[i].type == CONSTANT) && (sterm[i].e.i) < 0)
	  || (!multiple_dps && (term[i].a < 0))){
	size[variable[j - 1]] = i - variable[j - 1];
	variable[j++] = i;
      }
    size[variable[m - 1]] = n - variable[m - 1];
    variable[m] = n;
    merge_sort(variable, m, sizeof(int),
	       (int (*)(void *, void *))compare_indeces);

    /* Build equation system with variables ordered according to index */
    indeces[current_slice - genes->length] = (int *)xmalloc(m * sizeof(int));
    x[current_slice - genes->length] = (double *)xmalloc(m * sizeof(double));
    Ax = (double *)xmalloc(n * sizeof(double));
    Ap = (int *)xmalloc((m + 1) * sizeof(int));
    Ap[0] = 0;
    Ai = (int *)xmalloc(n * sizeof(int));
    bx = (double *)xmalloc(m * sizeof(double));
    j = 0;
    for (i = 0; i < m; i++){
      Ap[i] = j;
      bx[i] = 0;
      if (multiple_dps){
	/* Description of i'th variable starts at variable[i] in t */
	indeces[current_slice - genes->length][i]
	  = sterm[variable[i]].variable;
	/* Sort terms */
	merge_sort(sterm + variable[i], size[variable[i]],
		   sizeof(SymbolicTerm),
		   (int (*)(void *, void *))compare_sterms);
      }
      else{
	/* Description of i'th variable starts at variable[i] in t */
	indeces[current_slice - genes->length][i]
	  = term[variable[i]].variable;
	/* Sort terms */
	merge_sort(term + variable[i], size[variable[i]], sizeof(Term),
		   (int (*)(void *, void *))compare_terms);
      }
      /* Run through terms until end of t or start of new variable
       * description.
       */
      for (k = variable[i]; k < variable[i] + size[variable[i]]; k++){
	if (multiple_dps)
	  var = sterm[k].variable;
	else
	  var = term[k].variable;
	if (var < 0){
	  /* Variable is an MRCA */
	  if (multiple_dps)
	    bx[i] = -sterm[k].e.x;
	  else
	    bx[i] = -term[k].a;
	}
	else{
	  /* Normal term - check whether we know value of
	   * corresponding variable
	   */
	  if (multiple_dps){
	    /* Determine coefficient for current value of theta and rho */
	    switch (sterm[k].type){
	    case CONSTANT:
	      a = sterm[k].e.i;
	      break;
	    case THETA:
	      a = sterm[k].e.x * theta;
	      break;
	    case RHO:
	      a = sterm[k].e.x * rho;
	      break;
	    }
	    a = a
	      / (sterm[k].d + theta * sterm[k].dtheta + rho * sterm[k].drho);
	  }
	  else
	    a = term[k].a;

	  if (!known_value(bx + i, a, var, slice_size, x, indeces,
			   current_slice, genes->length)){
	    /* The corresponding variable is from current slice -
	     * insert term in equation system, and merge with
	     * subsequent duplicates (possible if a sequence in
	     * configuration is subsumed in two or more other
	     * sequences in the configuration), if any.
	     */
	    Ax[j] = a;
	    Ai[j] = var;
	    if (multiple_dps)
	      for (; (k + 1 < variable[i] + size[variable[i]])
		     && (sterm[k + 1].variable == var); k++){
		/* Duplicates should always be of type CONSTANT, but
		 * let us check just to make sure.
		 */
		switch (sterm[k + 1].type){
		case CONSTANT:
		  a = sterm[k + 1].e.i;
		  break;
		case THETA:
		  a = sterm[k + 1].e.x * theta;
		  fprintf(stderr, "WARNING: duplicate substitution term\n");
		  break;
		case RHO:
		  a = sterm[k + 1].e.x * rho;
		  fprintf(stderr, "WARNING: duplicate recombination term\n");
		  break;
		}
		Ax[j] += a / (sterm[k + 1].d + sterm[k + 1].dtheta * theta
			      + sterm[k + 1].drho * rho);
	      }
	    else
	      for (; (k + 1 < variable[i] + size[variable[i]])
		     && (term[k + 1].variable == Ai[j]); k++)
		Ax[j] += term[k + 1].a;
	    j++;
	  }
	}
      }
    }
    Ap[m] = j;
    /* Clean up */
    if (multiple_dps)
      free(sterm);
    else
      free(term);
    free(variable);
    free(size);

    /* Change indeces to local slice indeces and sort connected
     * components in equation system topologically.
     */
    for (i = 0; i < Ap[m]; i++)
      Ai[i] = get_index(Ai[i], m, indeces[current_slice - genes->length]);
    topological_sort(m, Ap, Ai, Ax, bx, &new, &old, &size);
    
    /* Set up auxiliary structures for iterative solution of
     * equation system.
     */
    AAp = (int *)xmalloc((m + 1) * sizeof(int));
    AAi = (int *)xmalloc(Ap[m] * sizeof(int));
    AAx = (double *)xmalloc(Ap[m] * sizeof(double));
    bbx = (double *)xmalloc(m * sizeof(double));
    xx = (double *)xmalloc(m * sizeof(double));
    
    /* Solve equation system */
    slice_max = 0;
    for (i = 0; size[i] < m; i++){
      /* Handle configurations in i'th component */
      AAindex = 0;
      AAp[0] = 0;
      for (j = 0; j < size[i + 1] - size[i]; j++){
	/* Handle AC j */
	bbx[j] = bx[size[i] + j];
	for (k = Ap[size[i] + j]; k < Ap[size[i] + j + 1]; k++){
	  if (Ai[k] < size[i])
	    /* This term involves a resolved variable */
	    bbx[j] -= Ax[k] * xx[Ai[k]];
	  else{
	    /* This term involves a variable in component i */
	    AAi[AAindex] = Ai[k] - size[i];
	    AAx[AAindex++] = Ax[k];
	  }
	}
	AAp[j + 1] = AAindex;
      }
      if (vlevel > 0){
	slice_max = (j > slice_max ? j : slice_max);
      }

      /* Solve this system */
      umfpack_di_symbolic(j, j, AAp, AAi, AAx, &symbolic, NULL, NULL);
      umfpack_di_numeric(AAp, AAi, AAx, symbolic, &numeric, NULL, NULL);
      umfpack_di_free_symbolic(&symbolic);
      umfpack_di_solve(UMFPACK_At, AAp, AAi, AAx, xx + size[i], bbx, numeric,
		       NULL, NULL);
      umfpack_di_free_numeric(&numeric);
    }
    if (vlevel > 0){
      avg_comp = (double)m / (double)i;
      n_components += i;
    }
    /* Move resolved values to their right place in x */
    for (i = 0; i < m; i++)
      x[current_slice - genes->length][old[i]] = xx[i];

    /* Clean up */
    free(AAp);
    free(AAi);
    free(AAx);
    free(bbx);
    free(Ap);
    free(Ai);
    free(Ax);
    free(bx);
    free(xx);
    free(old);
    free(new);
    free(size);

    if (vlevel > 2){
      fwrite(&m, sizeof(int), 1,
	     probabilities[current_slice - genes->length]);
      fwrite(x[current_slice - genes->length], sizeof(double), m,
	     probabilities[current_slice - genes->length]);
      fwrite(indeces[current_slice - genes->length], sizeof(int), m,
	     probabilities[current_slice - genes->length]);
    }

    /* Free slice that is now no longer needed */
    if ((current_slice >= 2 * genes->length)
	&& (slice_size[current_slice - 2 * genes->length] > 0)){
      free(x[current_slice - 2 * genes->length]);
      free(indeces[current_slice - 2 * genes->length]);
    }
    if (vlevel > 0){
      max_component = (slice_max > max_component ?
		       slice_max : max_component);
      if (vlevel > 1){
	printf(" took %.1f seconds (max. component size %d, avg. component size %.0f)\n", umfpack_timer() - time_point, slice_max, avg_comp);
	fflush(stdout);
      }
    }
  }
  /* Probability of data set is probability of first configuration in
   * last slice.
   */
  a = x[am - genes->length][0];

  /* Clean up */
  for (i = (am >= 2 *genes->length ? am - 2 * genes->length + 1 : 0);
       i <= am - genes->length; i++)
    if (slice_size[i] > 0){
      free(x[i]);
      free(indeces[i]);
    }
  free(x);
  free(indeces);
  free(slice_size);

  return a;
}

/* Compute Watterson's theta estimate */
double watterson(Genes *g)
{
  Genes *h = copy_genes(g);
  double an, t;
  int i;

  remove_nonsegregating(h);
  an = 0;
  for (i = h->n - 1; i >= 1; i--)
    an += 1.0 / i;
  if (h->length > 0)
    t = (double)h->length / an;
  else
    /* No segregating sites in data */
    t = 0;

  /* Clean up */
  free_genes(h);

  return t;
}

static double *equation_core(Genes *genes, int bound, HashTable *t)
{
  FILE **configurations, **probabilities;
  int i, j, ti, ri, bestt, bestr,
    datapoint = 0,
    am = ancestral_material(genes);
  double a, *dp, *result, time_point, bestl, worstl, maxt, maxr, maxl;
  MLEExpression *mleexp;
#ifdef ENABLE_VERBOSE
  int v = verbose();
  set_verbose(v - 1);
#endif

  /* Open temporary files for storing slices */
  configurations = (FILE **)xmalloc((am - genes->length + 1) * sizeof(FILE));
  for (i = 0; i <= am - genes->length; i++)
    if ((configurations[i] = (FILE *)tmpfile()) == NULL){
      fprintf(stderr, "Could not open temporary file for storing equations\n");
      exit(1);
    }
  if (vlevel > 2)
    probabilities = (FILE **)xmalloc((am - genes->length + 1) * sizeof(FILE));

  /* Determine terms in full equation system */
  enumerate_states(genes, bound, t, configurations, build_equations);

  /* Now read in terms in reverse order and solve equation system */
  result = (double *)xmalloc((elist_length(thetas) * elist_length(rhos)
			      + elist_length(pairs) + 3 * elist_length(mle))
			     * sizeof(double));

  /* Set up iteration through all theta and rho values */
  ti = 0;
  if (elist_length(pairs) > 0)
    ri = -1;
  else
    ri = 0;
  while (ri < elist_length(rhos)){
    /* Determine current theta and rho value and update iteration indeces */
    if (ri < 0){
      /* Still dealing with specific pairs */
      dp = elist_get(pairs, ti);
      theta = dp[0];
      rho = dp[1];
      if (++ti == elist_length(pairs))
	ti = ri = 0;
    }
    else{
      /* Next pair is from the grid */
      dp = elist_get(thetas, ti);
      theta = *dp;
      dp = elist_get(rhos, ri);
      rho = *dp;
      if (++ti == elist_length(thetas)){
	ti = 0;
	ri++;
      }
    }
    if (multiple_dps && (vlevel > 0))
      printf("Determining probability for theta = %lf and rho = %lf\n", theta,
	     rho);
    if (vlevel == 1){
      time_point = umfpack_timer();
      if (!multiple_dps){
	printf("Solving equation systems");
	fflush(stdout);
      }
    }

    result[datapoint++] = solve_equations(genes, configurations,
					  probabilities);
    if (vlevel == 1){
      printf(" took %.1f seconds\n", umfpack_timer() - time_point);
      fflush(stdout);
    }

    if (vlevel > 2){
      /* Output probabilities of individual configurations */
      n_ac = 0;
      enumerate_states(genes, bound, t, probabilities, output_probabilities);
    }
  }

  /* Search for MLE */
  for (i = 0; i < elist_length(mle); i++){
    /* Create working copy of search strategy */
    mleexp = (MLEExpression *)xmalloc(sizeof(MLEExpression));
    memcpy(mleexp, elist_get(mle, i), sizeof(MLEExpression));
    if (mleexp->thetamin < 0){
      /* Search from 0 to thetamax times Watterson's estimate */
      mleexp->thetamin = 0;
      mleexp->thetamax *= watterson(genes) / (r[genes->length] - r[0]);
    }
    if (vlevel > 0){
      printf("Finding MLE for theta in range %lf - %lf\nand rho in range %lf - %lf\n", mleexp->thetamin, mleexp->thetamax, mleexp->rhomin, mleexp->rhomax);
      if (vlevel == 1)
	time_point = umfpack_timer();
    }
    /* Allocate grid for storing likelihoods */
    dp = (double *)xmalloc(mleexp->thetan * mleexp->rhon * sizeof(double));
    maxl = 0;

    /* Repeatedly compute likelihoods for grid, until stopping
     * criteria is met.
     */
    for (j = 0;;j++) {
      bestl = bestt = bestr = 0;
      for (ti = 0; ti < mleexp->thetan; ti++){
	if (mleexp->thetan == 1)
	  theta = mleexp->thetamin;
	else
	  theta = mleexp->thetamin +
	    ti * (mleexp->thetamax - mleexp->thetamin) / (mleexp->thetan - 1);
	for (ri = 0; ri < mleexp->rhon; ri++){
	  if (mleexp->rhon == 1)
	    rho = mleexp->rhomin;
	  else
	    rho = mleexp->rhomin +
	      ri * (mleexp->rhomax - mleexp->rhomin) / (mleexp->rhon - 1);
	  dp[ti * mleexp->rhon + ri] = solve_equations(genes, configurations,
						       probabilities);
	  if (dp[ti * mleexp->rhon + ri] > bestl){
	    /* New best theta,rho pair */
	    bestt = ti;
	    bestr = ri;
	    bestl = dp[ti * mleexp->rhon + ri];
	  }
	  if (vlevel > 2){
	    /* Output probabilities of individual configurations */
	    n_ac = 0;
	    enumerate_states(genes, bound, t, probabilities,
			     output_probabilities);
	  }
	  if (vlevel > 0)
	    printf("Data set has likelihood %lE for theta = %lf, rho = %lf\n",
		   dp[ti * mleexp->rhon + ri], theta, rho);
	}
      }

      /* Update MLE */
      if (dp[bestt * mleexp->rhon + bestr] > maxl){
	maxl = dp[bestt * mleexp->rhon + bestr];
	if (mleexp->thetan == 1)
	  maxt = mleexp->thetamin;
	else
	  maxt = mleexp->thetamin + bestt
	    * (mleexp->thetamax - mleexp->thetamin) / (mleexp->thetan - 1);
	if (mleexp->rhon == 1)
	  maxr = mleexp->rhomin;
	else
	  maxr = mleexp->rhomin +
	    bestr * (mleexp->rhomax - mleexp->rhomin) / (mleexp->rhon - 1);
      }

      /* Check stopping criterias */
      if (((mleexp->thetan < 3) || ((mleexp->thetan == 3) && (bestt == 1)))
	  && ((mleexp->rhon < 3) || ((mleexp->rhon == 3) && (bestr == 1))))
	/* Search range is too small to allow finer subdivision */
	break;
      if (j + 1 == mleexp->maxiter)
	break;
      /* Determine lowest likelihood in neighbourhood of maximum */
      worstl = bestl;
      for (ti = (bestt > 0 ? -1 : 0);
	   ti <= (bestt < mleexp->thetan - 1 ? 1 : 0); ti++)
	for (ri = (bestr > 0 ? -1 : 0);
	     ri <= (bestr < mleexp->rhon - 1 ? 1 : 0); ri++)
	  if (((ti != 0) || (ri != 0))
	      && (dp[(bestt + ti) * mleexp->rhon + bestr + ri] < worstl))
	    worstl = dp[(bestt + ti) * mleexp->rhon + bestr + ri];
      if (dp[bestt * mleexp->rhon + bestr] - worstl < mleexp->ldiff)
	break;
      if ((worstl > 0)
	  && (log10(dp[bestt * mleexp->rhon + bestr]) - log10(worstl)
	      < mleexp->logldiff))
	break;

      /* Update search strategy */
      a = mleexp->thetamax - mleexp->thetamin;
      if (bestt != mleexp->thetan - 1)
	mleexp->thetamax = mleexp->thetamin
	  + (bestt + 1) * a / (mleexp->thetan - 1);
      if (bestt != 0)
	mleexp->thetamin = mleexp->thetamin
	  + (bestt - 1) * a / (mleexp->thetan - 1);
      a = mleexp->rhomax - mleexp->rhomin;
      if (bestr != mleexp->rhon - 1)
	mleexp->rhomax = mleexp->rhomin
	  + (bestr + 1) * a / (mleexp->rhon - 1);
      if (bestr != 0)
	mleexp->rhomin = mleexp->rhomin
	  + (bestr - 1) * a / (mleexp->rhon - 1);
      if (vlevel > 1)
	printf("Exploring subrange of theta %lf - %lf and rho %lf - %lf\n", mleexp->thetamin, mleexp->thetamax, mleexp->rhomin, mleexp->rhomax);
    }
    if (vlevel == 1)
      printf("MLE search took %.1f seconds\n", umfpack_timer() - time_point);
    /* Store maximum likelihood estimate in result */
    result[datapoint + 3 * i] = maxl;
    result[datapoint + 3 * i + 1] = maxt;
    result[datapoint + 3 * i + 2] = maxr;
    /* Clean up */
    free(mleexp);
    free(dp);
  }

  /* Clean up */
  if (multiple_dps)
    for (i = 0; i <= am - genes->length; i++)
      fclose(configurations[i]);
  free(configurations);
  if (vlevel > 2)
    free(probabilities);

  return result;
}

double *equation_absolute(Genes *genes, int bound)
{
  HashTable *t;
  Genes *g = copy_genes(genes);
  int minimum, table_size;
  double *p;

  /* Compute a good lower bound on the number of recombinations */
  implode_genes(g);
  if (no_recombinations_required(g))
    minimum = 0;
  else{
    reallocate_genes(g);
    minimum = LOWERBOUND(g);
  }
  if (minimum > bound){
    /* More recombinations required than we have available */
    free_genes(g);
    return NULL;
  }

  /* Determine sizes of and allocate hash tables and elists */
  table_size = msb(g->n * g->length) + bound;
  t = new_packedgeneshashtable(table_size);
  /* Compute minimum number of recombinations required */
  if (!no_recombinations_required(g))
    minimum = beagle_reusable_bounded(g, NULL, minimum, bound, t);
  free_genes(g);

  if (minimum <= bound)
    p = equation_core(genes, bound, t);

  /* Clean up */
  hashtable_destroy(t, (void (*)(void *))free_packedgenes, NULL,
		    (void (*)(void *))free);

  return p;
}

double *equation_relative(Genes *genes, int bound)
{
  HashTable *t;
  Genes *g;
  int minimum, table_size;
  double *p;

  if (bound < 0)
    /* No ancestral configurations are reachable with less than
     * minimum number of recombinations.
     */
    return NULL;

  /* Compute a good lower bound on the number of recombinations */
  g = copy_genes(genes);
  implode_genes(g);
  if (no_recombinations_required(g))
    minimum = 0;
  else{
    reallocate_genes(g);
    minimum = LOWERBOUND(g);
  }

  /* Determine sizes of and allocate hash tables and elists */
  table_size = msb(g->n * g->length) + minimum + bound;
  t = new_packedgeneshashtable(table_size);
  /* Compute minimum number of recombinations required */
  if (!no_recombinations_required(g))
    minimum = beagle_reusable_bounded(g, NULL, minimum, minimum + bound, t);
  free_genes(g);

  p = equation_core(genes, minimum + bound, t);

  /* Clean up */
  hashtable_destroy(t, (void (*)(void *))free_packedgenes, NULL,
		    (void (*)(void *))free);

  return p;
}

/* Extract numeric positions from position labels */
void convert_position(char *label, va_list args)
{
  int *i = va_arg(args, int *);

  if (sscanf(label, "%lf", s + (*i)++) != 1){
    fprintf(stderr, "Could not interpret %s as position\n", label);
    print_usage(stderr, "");
    exit(1);
  }
}

/* Parse an unsigned integer option argument and store it in value.
 * Return value states whether the argument could be parsed in full.
 */
static int parse_uint(char *s, unsigned int *value)
{
  int i;

  if (sscanf(s, "%u%n", value, &i) != 1)
    return 0;

  return s[i] == '\0';
}

/* Parse an integer option argument and store it in value. Return
 * value states whether the argument could be parsed in full.
 */
static int parse_int(char *s, int *value)
{
  int i;

  if (sscanf(s, "%d%n", value, &i) != 1)
    return 0;

  return s[i] == '\0';
}

/* Parse a floating point option argument and store it in value.
 * Return value states whether the argument could be parsed in full.
 */
static int parse_double(char *s, double *value)
{
  int i;

  if (sscanf(s, "%lf%n", value, &i) != 1)
    return 0;

  return s[i] == '\0';
}

/* Parse comma separated pair of doubles and store in first two
 * positions of value (that must be an array of at least two
 * doubles). Return value states whether the pair could be parsed in
 * full.
 */
static int parse_pair(char *s, double *value)
{
  int i;

  if (sscanf(s, "%lf,%lf%n", value, value + 1, &i) != 2)
    return 0;

  return s[i] == '\0';
}

/* Parse comma separated pair of doubles and store in first two
 * positions of value (that must be an array of at least two
 * doubles). Either value can be absent, in which case the
 * corresponding entry in value is not changed. Return value states
 * whether the pair could not be parsed in full (negative), no value
 * was present (0), only first value was present (1), only second
 * value was present (2) or both values were present (3).
 */
static int parse_fuzzy_pair(char *s, double *value)
{
  int i, n = 0;
  double t1, t2;

  /* Try to read first double */
  n = sscanf(s, "%lf%n", &t1, &i);
  if (s[i] != ',')
    return -1;

  /* Try to read second double */
  n += 2 * sscanf(s + i + 1, "%lf%n", &t2, &i);
  if (s[i] != ',')
    return -1;

  /* Assign values to value array */
  if (n & 1)
    value[0] = t1;
  if (n & 2)
    value[1] = t2;

  return n;
}

/* Parse a comma separated list of doubles and insert values in
 * elist. Return value states whether the list could be parsed in
 * full.
 */
static int parse_commalist(char *s, EList *elist)
{
  int i;
  double a, *b;
  char *format = "%lf%n";

  while (sscanf(s, format, &a, &i) == 1){
    b = (double *)xmalloc(sizeof(double));
    *b = a;
    elist_append(elist, b);
    s += i;
    format = ",%lf%n";
  }

  return s[0] == '\0';
}

/* Parse a semicolon separated list of pairs of doubles and insert
 * values in elist. Return value states whether the list could be
 * parsed in full.
 */
static int parse_colonlist(char *s, EList *elist)
{
  int i, j;
  double a, b, *c;
  char *format = "%lf,%lf%n";

  while ((j = sscanf(s, format, &a, &b, &i)) == 2){
    c = (double *)xmalloc(2 * sizeof(double));
    c[0] = a;
    c[1] = b;
    elist_append(elist, c);
    s += i;
    format = ":%lf,%lf%n";
  }

  return s[0] == '\0';
}

/* Parse a specification of an MLE search and insert expression in
 * elist. Return value states whether the expression could be parsed
 * in full.
 */
static int parse_mleexpression(char *s, EList *elist)
{
  int i;
  MLEExpression *mle = xmalloc(sizeof(MLEExpression));

  if (s[0] == '-'){
    mle->thetamin = -1;
    if (sscanf(s + 1, "%lf,%u:%lf-%lf,%u:%u:%lf:%lf%n", &mle->thetamax,
	       &mle->thetan, &mle->rhomin, &mle->rhomax, &mle->rhon,
	       &mle->maxiter, &mle->ldiff, &mle->logldiff, &i) != 8){
      free(mle);
      return 0;
    }
    i++;
  }
  else if (sscanf(s, "%lf-%lf,%u:%lf-%lf,%u:%u:%lf:%lf%n", &mle->thetamin,
		  &mle->thetamax, &mle->thetan, &mle->rhomin, &mle->rhomax,
		  &mle->rhon, &mle->maxiter, &mle->ldiff, &mle->logldiff, &i)
	   != 9){
    free(mle);
    return 0;
  }

  /* Check sanity of specification */
  if ((mle->thetamax < 0) || (mle->thetamin > mle->thetamax)
      || (mle->thetan < 1)
      || ((mle->thetan == 1) && (mle->thetamin >= 0)
	  && (mle->thetamin != mle->thetamax))
      || ((mle->thetan == 1) && (mle->thetamin < 0) && (mle->thetamax != 0))
      || (mle->rhomin < 0) || (mle->rhomin > mle->rhomax)
      || (mle->rhon < 1) || ((mle->rhon == 1) && (mle->rhomin != mle->rhomax))
      || ((mle->maxiter == 0) && (mle->ldiff <= 0) && (mle->logldiff <= 0))
      || (s[i] != '\0')){
    free(mle);
    return 0;
  }

  elist_append(elist, mle);
  return 1;
}

int main(int argc, char **argv)
{
  AnnotatedGenes *g;
  Gene_Format geneformat = GENE_ANY;
  Gene_SeqType seqtype = GENE_BINARY;
  /* Default behaviour is to enumerate minACs */
  int i, j, k, relative = 1, e_option = 0;
  unsigned int bound = 0;
  double *t, *p, ends[2], interval = 0, time_point = umfpack_timer();

  /* Analyse command line options */
  thetas = elist_make();
  rhos = elist_make();
  pairs = elist_make();
  mle = elist_make();
  while ((i = getopt(argc, argv, "c:r:ofkanhHt:s:v::ylp:T:S:P:m:e:i::z::")) >= 0)
    switch(i){
    case 'h':
    case 'H':
      print_usage(stdout, argv[0]);
      exit(0);
    case '?':
    case ':':
      print_usage(stderr, argv[0]);
      exit(1);
    case 'c':
      /* Enumerate with an absolute number of recombinations */
      if (!parse_uint(optarg, &bound)){
	fprintf(stderr, "Could not parse %s as absolute number of recombinations allowed\n", optarg);
	print_usage(stderr, argv[0]);
	exit(1);
      }
      relative = 0;
      break;
    case 'r':
      /* Enumerate with a relative number of recombinations */ 
      if (!parse_uint(optarg, &bound)){
	fprintf(stderr, "Could not parse %s as relative number of recombinations allowed\n", optarg);
	print_usage(stderr, argv[0]);
	exit(1);
      }
      relative = 1;
      break;
    case 'k':
      gene_knownancestor = 1;
      break;
    case 'o':
      geneformat = GENE_BEAGLE;
      break;
    case 'f':
      geneformat = GENE_FASTA;
      break;
    case 'a':
      seqtype = GENE_AMINO;
      break;
    case 'n':
      seqtype = GENE_NUCLEIC;
      break;
    case 't':
      t = (double *)xmalloc(sizeof(double));
      if (!parse_double(optarg, t)){
	fprintf(stderr, "Could not parse %s as a theta value\n", optarg);
	print_usage(stderr, argv[0]);
	exit(1);
      }
      elist_append(thetas, t);
      break;
    case 's':
      t = (double *)xmalloc(sizeof(double));
      if (!parse_double(optarg, t)){
	fprintf(stderr, "Could not parse %s as a rho value\n", optarg);
	print_usage(stderr, argv[0]);
	exit(1);
      }
      elist_append(rhos, t);
      break;
    case 'p':
      t = (double *)xmalloc(2 * sizeof(double));
      if (!parse_pair(optarg, t)){
	fprintf(stderr, "Could not parse %s as a theta,rho value\n", optarg);
	print_usage(stderr, argv[0]);
	exit(1);
      }
      elist_append(pairs, t);
      break;
    case 'T':
      if (!parse_commalist(optarg, thetas)){
	fprintf(stderr, "Could not parse %s as list of theta values\n",
		optarg);
	print_usage(stderr, argv[0]);
	exit(1);
      }
      break;
    case 'S':
      if (!parse_commalist(optarg, rhos)){
	fprintf(stderr, "Could not parse %s as list of rho values\n", optarg);
	print_usage(stderr, argv[0]);
	exit(1);
      }
      break;
    case 'P':
      if (!parse_colonlist(optarg, pairs)){
	fprintf(stderr, "Could not parse %s as list of theta,rho values\n",
		optarg);
	print_usage(stderr, argv[0]);
	exit(1);
      }
      break;
    case 'm':
      if (!parse_mleexpression(optarg, mle)){
	fprintf(stderr, "Could not parse %s as list MLE search expression\n",
		optarg);
	print_usage(stderr, argv[0]);
	exit(1);
      }
      break;
    case 'e':
      if ((i = parse_fuzzy_pair(optarg, ends)) < 0){
	print_usage(stderr, argv[0]);
	fprintf(stderr, "Could not parse %s as pair of positions\n", optarg);
	exit(1);
      }
      e_option |= i;
      break;
    case 'i':
      if (optarg != 0){
	if ((!parse_double(optarg, &interval)) || (interval < 0)){
	  print_usage(stderr, argv[0]);
	  fprintf(stderr, "Could not parse %s as interval length\n", optarg);
	  exit(1);
	}
      }
      else
	interval = 1;
      break;
    case 'y':
      aprime = 1;
      break;
    case 'l':
      noloops = 1;
      break;
    case 'z':
      if (optarg != 0){
	if ((strlen(optarg) != 1) || ((*optarg != 's') && (*optarg != 'l'))){
	  print_usage(stderr, argv[0]);
	  fprintf(stderr,
		  "Argument should be single character s or l (was %s)\n",
		  optarg);
	  exit(1);
	}
	else if (*optarg == 's')
	  _model_view = _STATES_MODEL;
	else
	  _model_view = _SITES_MODEL;
      }
      else
	_model_view = _FULLY_MODEL;
      break;
    case 'v':
      if (optarg != 0){
	if (!parse_int(optarg, &i))
	  fprintf(stderr, "WARNING: could not parse %s as verbosity level\n",
		  optarg);
	else
	  vlevel = i;
      }
      else
	vlevel = 2;
      break;
    }

  /* Sanitice model view - if ancestral material overlap is not
   * required, then only the restriction to ancestral states in (near)
   * minimal ARGs can be viewed as eitheer computational or modelling
   * approximation.
   */
  if (!noloops){
    if (_model_view == _SITES_MODEL)
      _model_view = _HEURISTIC;
    else if (_model_view == _FULLY_MODEL)
      _model_view = _STATES_MODEL;
  }

#ifdef ENABLE_VERBOSE
  set_verbose(vlevel - 3);
#endif

  /* Read data */
  if (argc > optind){
    if ((g = read_genes(argv[optind], geneformat, seqtype)) == NULL){
      fprintf(stderr, "Could not parse %s as valid data\n", argv[optind]);
      exit(1);
    }
  }
  else{
    if ((g = read_genes(NULL, geneformat, seqtype)) == NULL){
      fprintf(stderr, "Could not parse input as valid data\n");
      exit(1);
    }
  }
  if (g->positions == NULL){
    if (argc > optind)
      fprintf(stderr, "No site positions available in %s\n", argv[optind]);
    else
      fprintf(stderr, "No site positions available in data\n");
    exit(1);
  }
  if ((gene_knownancestor) && (seqtype != GENE_BINARY))
    /* First sequnce only included to specify known common ancestor */
    remove_annotatedgene(g, 0);

  /* Convert positions to numerical values, and check consistency */
  i = 0;
  s = (double *)xmalloc(g->g->length * sizeof(double));
  LListMap(g->positions, (void (*)(void *, va_list))convert_position, &i);
  for (i = 1; i < g->g->length; i++){
    if (s[i - 1] > s[i]){
      print_usage(stderr, argv[optind]);
      fprintf(stderr, "Positions %s and %s are not well ordered\n",
	      (char *)GetByIndex(positions, i - 1),
	      (char *)GetByIndex(positions, i));
      exit(1);
    }
  }
  if (e_option & 1){
    if (s[0] < ends[0]){
      print_usage(stderr, argv[optind]);
      fprintf(stderr, "Leftmost position is prior to specified left end point\n");
      exit(1);
    }
  }
  else
    ends[0] = s[0];
  if (e_option & 2){
    if (s[g->g->length - 1] > ends[1]){
      print_usage(stderr, argv[optind]);
      fprintf(stderr, "Rightmost position is posterior to specified right end point\n");
      exit(1);
    }
  }
  else
    ends[1] = s[g->g->length - 1];

  /* We cannot estimate rates for an interval of length 0 */
  if (ends[0] == ends[1]){
    print_usage(stderr, argv[optind]);
    fprintf(stderr, "Sequences' left and right end points are identical\n");
    exit(1);
  }

  /* Rescale positions */
  if (interval > 0){
    for (i = 0; i < g->g->length; i++)
      s[i] = interval * (s[i] - ends[0]) / (ends[1] - ends[0]);
    for (i = 1; i < g->g->length; i++)
      if (s[i - 1] >= s[i]){
	fprintf(stderr, "WARNING: after rescaling positions %s and %s became identical\nRecombination between these positions will have probability 0\n",
	      (char *)GetByIndex(positions, i - 1),
	      (char *)GetByIndex(positions, i));
	s[i - 1] = s[i];
      }
    ends[0] = 0;
    ends[1] = interval;
  }

  /* Determine recombination points */
  r = (double *)xmalloc((g->g->length + 1) * sizeof(double));
  for (i = 1; i < g->g->length; i++)
      r[i] = (s[i] + s[i - 1]) / 2;
  r[0] = ends[0];
  r[g->g->length] = ends[1];

  /* Make positions globally available (for output purposes) */
  positions = g->positions;

  /* Determine theta, rho points at which to calculate probability */
  if (elist_length(thetas) == 0){
    /* No theta value specified */
    if (((elist_length(pairs) == 0) && (elist_length(mle) == 0))
	|| (elist_length(rhos) > 0)){
      /* But we need a theta value - use Watterson's estimate */
      if (elist_length(rhos) > 0){
	t = (double *)xmalloc(sizeof(double));
	elist_append(thetas, t);
      }
      else{
	/* No values specified, use default rho of 1 */
	t = (double *)xmalloc(2 * sizeof(double));
	t[1] = 1;
	elist_append(pairs, t);
      }
      /* Compute default theta (Watterson's estimate) */
      *t = watterson(g->g) / (r[g->g->length] - r[0]);
      if (vlevel > 0)
	printf("Watterson's theta estimate theta_W = %lg\n", *t);
    }
  }
  else if (elist_length(rhos) == 0){
    t = (double *)xmalloc(sizeof(double));
    *t = 1;
    elist_append(rhos, t);
  }
  if ((elist_length(thetas) > 1) || (elist_length(rhos) > 1)
      || (elist_length(pairs) > 1)
      || ((elist_length(thetas) != 0) && (elist_length(pairs) != 0))
      || (elist_length(mle) != 0))
    /* More than one theta, rho pair specified */
    multiple_dps = 1;
  else
    /* Only one theta, rho pair specified */
    if (elist_length(thetas) != 0){
      theta = *(double *)elist_get(thetas, 0);
      rho = *(double *)elist_get(rhos, 0);
    }
    else{
      t = (double *)elist_get(pairs, 0);
      theta = t[0];
      rho = t[1];
    }

  /* Set up and solve equation system(s) */
  if (relative)
    p = equation_relative(g->g, bound);
  else
    p = equation_absolute(g->g, bound);

  /* Output results */
  if (multiple_dps){
    if (vlevel > 0)
      printf("Data set has %d reachable configurations\n", n_ac);
    k = i = 0;
    if (elist_length(pairs) > 0)
      j = -1;
    else
      j = 0;
    while (j < elist_length(rhos)){
      /* Determine current theta and rho value and update iteration indeces */
      if (j < 0){
	/* Still dealing with specific pairs */
	t = elist_get(pairs, i);
	theta = t[0];
	rho = t[1];
	if (++i == elist_length(pairs))
	  i = j = 0;
      }
      else{
	/* Next pair is from the grid */
	t = elist_get(thetas, i);
	theta = *t;
	t = elist_get(rhos, j);
	rho = *t;
	if (++i == elist_length(thetas)){
	  i = 0;
	  j++;
	}
      }
      /* Output probability and data point information */
      if (vlevel > 0)
	printf(" and likelihood %lE for theta = %lf, rho = %lf\n",
	       (p != NULL ? p[k++] : 0), theta, rho);
      else
	printf("%lE %lf %lf\n", (p != NULL ? p[k++] : 0), theta, rho);
    }
  }
  else
    if (vlevel > 0)
      printf("Data set has %d reachable configurations and likelihood %lE\n",
	     n_ac, (p != NULL ? *p : 0));
    else
      printf("%lE\n", (p != NULL ? *p : 0));
  if (elist_length(mle) > 0){
    if (vlevel > 0)
      printf("\nMaximum likelihood estimates:\n\n");
    for (i = 0; i < elist_length(mle); i++)
      if (vlevel > 0)
	printf("Data set has likelihood %lE for theta = %lf, rho = %lf\n",
	       p[k + 3 * i], p[k + 3 * i + 1], p[k + 3 * i + 2]);
      else
	printf("%lE %lf %lf\n", p[k + 3 * i], p[k + 3 * i + 1],
	       p[k + 3 * i + 2]);
  }

  /* Clean up */
  if (p != NULL)
    free(p);
  for (i = 0; i < elist_length(pairs); i++)
    free(elist_get(pairs, i));
  elist_destroy(pairs);
  for (i = 0; i < elist_length(thetas); i++)
    free(elist_get(thetas, i));
  elist_destroy(thetas);
  for (i = 0; i < elist_length(rhos); i++)
    free(elist_get(rhos, i));
  for (i = 0; i < elist_length(mle); i++)
    free(elist_get(mle, i));
  elist_destroy(mle);
  elist_destroy(rhos);
  free_annotatedgenes(g);
  free(r);
  free(s);

  /* Print total time usage */
  if (vlevel > 0)
    printf("Total time used is %.1f seconds\nLargest equation system component had %d variables\nAverage number of variables in equations system component was %.0f\n", umfpack_timer() - time_point, max_component, (double)n_ac / (double)n_components);

  return 0;
}
