/* common.c: Implementation of functions that should be commonly available
 * to all source files.
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "common.h"
#include "gene.h"
#include "llist.h"
#include "arg.h"

#ifdef ENABLE_VERBOSE
static int verbosity = 0;
int verbose()
{
  return (verbosity > 0 ? verbosity : 0);
}

void set_verbose(int v)
{
  verbosity = v;
}
#endif
#ifdef HAPLOTYPE_BLOCKS
LList *representativeness = NULL;
LListCounter *representativeness_counter;
int **haploblocks = NULL;
#endif
LList *eventlist;
#ifdef DEBUG
/* Define structure for storing trace of ancestral states as we return
 * from a successful path.
 */
HashTable *ancestral_state_trace = NULL;
#endif

/* xmalloc(n): Allocate n bytes of memory, checking for successful allocation.
 */
void *xmalloc(int n)
{
  void *adr;

  if (n <= 0){
    fprintf(stderr,
	    "Erroneous memory allocation - please email error report\n");
    return NULL;
  }

  if ((adr = malloc(n)) == NULL){
    fprintf(stderr, "Unable to allocate sufficient amount of memory\n");
    exit(2);
  }

  return adr;
}

/* xcalloc(n): Allocate n words of size m of memory intialised to
 * zero, checking for successful allocation.
 */
void *xcalloc(int m, int n)
{
  void *adr;

  if ((n <= 0) || (m <= 0)){
    fprintf(stderr,
	    "Erroneous memory allocation - please email error report\n");
    return NULL;
  }

  if ((adr = calloc(m, n)) == NULL){
    fprintf(stderr, "Unable to allocate sufficient amount of memory\n");
    exit(2);
  }

  return adr;
}

/* xrealloc(n): Resize memory allocated at oldadr to n bytes of memory,
 * checking for successful allocation.
 */
void *xrealloc(void *oldadr, int n)
{
  void *adr;

  if (n <= 0){
    fprintf(stderr,
	    "Erroneous memory allocation - please email error report\n");
    return NULL;
  }

  if ((adr = realloc(oldadr, n)) == NULL){
    fprintf(stderr, "Unable to allocate sufficient amount of memory\n");
    exit(2);
  }

  return adr;
}

/* Interface to random number generator. This interface is used
 * throughout, so only this needs to be replaced to change random
 * number generator.
 */
/* initialise_xrandom(): Initialise random number generator, using the
 * current time.
 */
void initialise_xrandom()
{
#ifndef DEBUG
  /* Use time to set random seed */
  srandom(time(NULL));
#else
  /* Make sure random sequence is the same for every run */
  srandom(1);
#endif
}

/* xrandom(): Return (pseudo-)random number between 0 and XRAND_MAX */
long int xrandom()
{
  return random();
}

/* Convert n to string */
char *i2a(int n)
{
  char *s, *t;

  if (n > 0)
    s = t = xmalloc(((int)log10(n) + 2) * sizeof(char));
  else if (n == 0)
    s = t = xmalloc(2 * sizeof(char));
  else{
    s = xmalloc(((int)log10(n) + 2) * sizeof(char));
    s[0] = '-';
    t = s + 1;
  }

  sprintf(t, "%d", n);

  return s;
}

/* Output s to fp with line length l and indentation i */
void pretty_print(FILE *fp, char *s, int l, int i)
{
  int j;
  char *last = s + strlen(s) - 1;

  if (fp == NULL)
    fp = stdout;

  /* Remove initial stretches of white space */
  while (isspace(*s))
    s++;
  if (s > last)
    /* s contains only white space */
    return;
  while (isspace(*last))
    last--;

  while (s <= last){
    /* Output indentation */
    for (j = 0; j < i; j++)
      fputc(' ', fp);

    /* Is there a new line within reach? */
    for (j = 0; (j <= l - i - 2) && (s + j <= last); j++)
      if (s[j] == '\n'){
	fwrite(s, sizeof(char), j + 1, fp);
	s += j + 1;
	break;
      }
    if ((j <= l - i - 2) && (s + j <= last))
      /* New line encountered - continue with next line */
      continue;

    if (s + l - i <= last){
      /* Find good place for next line break */
      for (j = l - i - 2; j > 0; j--){
	if ((s[j] == '-') && (s[j - 1] != ' ')){
	  /* Break after hyphen */
	  j++;
	  break;
	}
	if ((s[j] == ' ') && (s[j + 1] != '-')){
	  /* Break at space, unless it borders a dash */
	  while ((j > 0) && (s[j] == ' '))
	    j--;
	  if ((j > 0) && (s[j - 1] != '-'))
	    break;
	}
      }
      if (j == 0){
	/* No good line break found - look for acceptable line break */
	for (j = l - i - 2; (j > 0) && (s[j] != ' '); j--);
	while (s[j] == ' ')
	  j--;
	if (j == 0)
	  j = l - i - 2;
      }
    }
    else
      /* Rest of text fits on one line */
      j = last - s;
    fwrite(s, sizeof(char), j + 1, fp);
    fputc('\n', fp);
    s += j + 1;
    /* Remove initial stretch of white space */
    while (isspace(*s))
      s++;
  }
}

/* Print an option description to fp with line length l and subsequent
 * line indentation i (length of option if i negative).
 */
void print_option(FILE *fp, char *option, char *description, int l, int i)
{
  int n, m, j;

  if (fp == NULL)
    fp = stdout;

  /* Remove initial stretches of white space */
  while (isspace(*option))
    option++;
  n = strlen(option);
  if (n > 0)
    while (isspace(option[n - 1]))
      n--;
  while (isspace(*description))
    description++;
  m = strlen(description);
  if (m > 0)
    while (isspace(description[m - 1]))
      m--;

  /* Output option */
  fputc(' ', fp);
  fwrite(option, sizeof(char), n, fp);
  fputc(' ', fp);

  if (n + 2 + m <= l){
    /* Whole description fits on first line - output it */
    fwrite(description, sizeof(char), m, fp);
    fputc('\n', fp);
  }
  else{
    /* Check whether there is a new line within reach */
    for (j = 0; j <= l - n - 2; j++)
      if (description[j] == '\n'){
	fwrite(description, sizeof(char), j + 1, fp);
	pretty_print(fp, description + j + 1, l, (i < 0 ? n + 2 : i));
	return;
      }

    /* Output first line of description */
    /* Find good place for next line break */
    for (j = l - n - 3; j > 0; j--){
      if ((description[j] == '-') && (description[j - 1] != ' ')){
	/* Break after hyphen */
	j++;
	break;
      }
      if ((description[j] == ' ') && (description[j + 1] != '-')){
	/* Break at space, unless it borders a dash */
	while ((j > 0) && (description[j] == ' '))
	  j--;
	if ((j > 0) && (description[j - 1] != '-'))
	  break;
      }
    }
    if (j == 0){
      /* No good line break found - look for acceptable line break */
      for (j = l - n - 4; (j > 0) && (description[j] != ' '); j--);
      while (description[j] == ' ')
	j--;
      if (j == 0)
	j = l - n - 4;
    }
    fwrite(description, sizeof(char), j + 1, fp);
    fputc('\n', fp);
    description += j + 1;
    /* Output remainder of description */
    pretty_print(fp, description, l, (i < 0 ? n + 2 : i));
  }
}

#ifdef HAPLOTYPE_BLOCKS
/* From a matrix local of local bounds on recombinations in an
 * imploded data set and a list r describing what sites in the
 * original data set a site in the imploded data set corresponds to,
 * construct, in local, the matrix of local bounds on recombinations
 * in the original data set. It is assumed that local has been
 * allocated large enough to hold the expanded matrix.
 */
void explode_local(int **local, LList *r, int n)
{
  int i, j, a, b;
  SuperColumn *f, *l, *f2;
  LListCounter *first, *last;

  first = MakeCounter(r, LAST);
  last = MakeCounter(r, LAST);
  f = (SuperColumn *)Prev(first);
  /* Run through all pairs of sites in imploded sequence set */
  while (f != NULL){
    f2 = (SuperColumn *)Prev(first);
    a = (f2 != NULL ? f2->right : -1);
    b = n;
    l = (SuperColumn *)Prev(last);
    while (l != f){
      /* Run through all pairs of sites in original sequence set the
       * pair of sites in the imploded sequence set corresponds to.
       */
      for (i = f->right; i > a; i--)
	for (j = b - 1; j >= l->left; j--)
	  local[i][j - i - 1] = local[GetPosition(first) + 1]
	    [GetPosition(last) - GetPosition(first) - 2];
      b = l->left;
      l = Prev(last);
    }
    /* Make sure to reset entries corresponding to bounds between
     * sites in the same collapsed position to 0.
     */
    for (i = b - 2; i > a; i--)
      for (j = b - 1; j > i; j--)
	local[i][j - i - 1] = 0;
    f = f2;
    InitCounter(last, r, LAST);
  }
  DestroyCounter(first);
  DestroyCounter(last);
}
#endif
