/* eagl.c: Main source file for computing lower bounds by applying the
 * copmposite method to exact minimum number of recombinations
 * required for intervals of the data.
 *
 * Rune Lyngsoe, December 2004
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <time.h>
#include <limits.h>
#ifdef HAPLOTYPE_BLOCKS
#include <math.h>
#endif

#include "gene.h"
#include "bounds.h"
#include "common.h"

void print_usage(FILE *f, char *name)
{
  int linelength = 80;

  fprintf(f, "Usage: %s [options] [input]\n", name);
  pretty_print(f, "The program reads data from the input file specified (or standard input if no input is specified) and computes two bounds on the minimum number of recombinations required to explain the data set under the infinite sites assumption. The first bound is the haplotype bound, while the second bound is based on computing exact values of the minimum number of recombinations for (small) regions of the data. In both cases the composite method is used to combine local bounds to global bounds.", linelength, 0);
  pretty_print(f, "Legal options are:", linelength, 0);
#ifdef HAPLOTYPE_BLOCKS
  print_option(f, "-m[file]", "Output matrix of local bounds. The haplotype bounds will be above the diagonal, while the bounds based on exact values will be below the diagonal. The matrix is written to file, or to standard output if no file is specified.", linelength, -1);
#endif
  print_option(f, "-k", "Assume that the common ancestral sequence is known, i.e. that we know which is the wild type and which is the mutant in each site. In binary format the all-0 sequence is assumed to be the common ancestral sequence, while the common ancestral sequence has to be specified directly for amino acid and nucleic sequences (see options -a and -n)", 80, -1);
  print_option(f, "-o", "Assume input data is in own format (default is to first try to parse data in own format, and if that fails to try to parse it in fasta format).", 80, -1);
  print_option(f, "-f", "Assume input data is in fasta format.", 80, -1);
  print_option(f, "-a", "Assume input consists of amino acid (protein) sequences using the one letter amino acid alphabet (default is to assume sequences in binary, i.e. 0/1, format). If the most recent common ancestor is assumed known, the first sequence in the input data is considered to specify the wild type of each site and is not included as part of the sample set.", 80, -1);
  print_option(f, "-n", "Assume input consists of nucleic sequences; anything but a/c/g/t/u is considered an unresolved site. If the most recent common ancestor is assumed known, the first sequence in the input data is considered to specify the wild type of each site and is not included as part of the sample set.", 80, -1);
  print_option(f, "-lN", "Only compute exact values for regions of length at most N (default is to compute exact values for all regions, including the exact minimum number of recombinations required for the entire data set", linelength, -1);
  print_option(f, "-q[N]", "Regions are handled in order of increasing length. If the -q option is specified, query the user after each N length increases whether to continue expanding regions (the default value of N is one)", linelength, -1);
  print_option(f, "-tN", "Stop expanding regions after N seconds", linelength, -1);
  print_option(f, "-sN", "Only consider subsets of at most N different types of sites", linelength, -1);
  print_option(f, "-wN", "Only consider two sites in conjunction if they are contained in a region of no more than N sites in the reduced data set, i.e. where all non-informative sites, neighbouring identical columns, and sequence duplicates have been iteratively removed", linelength, -1);
  print_option(f, "-iN", "Only continue with a subset of site types if the last type added increased the number of composite types by at least N", linelength, -1);
  print_option(f, "-h, -H", "Print this information and stop", linelength, -1);
}

static int query = 0;
static int query_user(int n, int m, int **B)
{
  int answer;

  if (n % query == 0){
    printf("Regions up to length %d handled, bound is now %d - continue (Y/N)?\n", n, local2global(m, B));
    for (;;){
      answer = getchar();
      switch(answer){
      case 'y':
      case 'Y':
	return 0;
      case 'n':
      case 'N':
	return 1;
      }
    }
  }

  return 0;
}

static int timelimit = 0;
static int timeout()
{

  printf("%ld %ld\n", (long int)clock(), CLOCKS_PER_SEC);
  if (clock() > timelimit * CLOCKS_PER_SEC)
    return 1;
  else
    return 0;
}

int main(int argc, char **argv)
{
  AnnotatedGenes *a;
  Genes *g;
  Sites *s;
  int i, maxlength = 0;
  int (*query_function)(int, int, int **) = NULL;
  int (*timeout_function)(void) = NULL;
  Gene_Format geneformat = GENE_ANY;
  Gene_SeqType seqtype = GENE_BINARY;
#ifdef HAPLOTYPE_BLOCKS
  int j, k, **B, **E, n, entry_width;
  FILE *matrix = NULL;
  SuperColumn *c;
  char *format;
#endif
  int set = INT_MAX, width = INT_MAX, increase = 0, heuristic = 0;

#ifdef ENABLE_VERBOSE
  int v = 0, j;
#endif
  gene_knownancestor = 0;

  /* Initialise random number generator */
  initialise_xrandom();

  /* Analyse command line options */
#define EAGL_OPTIONS1 "hHl:q::t:ofans:w:i:"
#ifdef HAPLOTYPE_BLOCKS
#define EAGL_OPTIONS EAGL_OPTIONS1 "m::"
#else
#define EAGL_OPTIONS EAGL_OPTIONS1
#endif

  /* Parse command line options */
  while ((i = getopt(argc, argv, EAGL_OPTIONS)) >= 0)
    switch(i){
    case 'h':
    case 'H':
      print_usage(stdout, argv[0]);
      exit(0);
      break;
    case '?':
    case ':':
      print_usage(stderr, argv[0]);
      exit(1);
      break;
    case 'k':
      gene_knownancestor = 1;
      break;
    case 'o':
      geneformat = GENE_BEAGLE;
      break;
    case 'f':
      geneformat = GENE_FASTA;
      break;
    case 'a':
      seqtype = GENE_AMINO;
      break;
    case 'n':
      seqtype = GENE_NUCLEIC;
      break;
    case 'l':
      if ((maxlength = atoi(optarg)) < 0)
	maxlength = 0;
      break;
    case 'q':
      if ((optarg == 0) || ((query = atoi(optarg)) < 0))
	query = 1;
      break;
    case 't':
      if ((timelimit = atoi(optarg)) < 0)
	timelimit = 0;
      break;
#ifdef HAPLOTYPE_BLOCKS
    case 'm':
      /* Compute matrix of local minimum number of recombinations required */
      /* Check whether we already opened an output file that now needs
       * to be closed.
       */
      if ((matrix != NULL) && (matrix != stdout))
	fclose(matrix);
      /* Was a file name specified? */
      if (optarg != 0){
	if ((matrix = fopen(optarg, "w")) == NULL){
	  fprintf(stderr, "Could not open file %s for output\n", optarg);
	  print_usage(stderr, argv[0]);
	  exit(1);
	}
      }
      else
	matrix = stdout;
      break;
#endif
    case 's':
      if ((set = atoi(optarg)) < 0)
	set = INT_MAX;
      else
	heuristic = 1;
      break;
    case 'w':
      if ((width = atoi(optarg)) < 0)
	width = INT_MAX;
      else
	heuristic = 1;
      break;
    case 'i':
      if ((increase = atoi(optarg)) < 0)
	increase = 0;
      else
	heuristic = 1;
      break;
    }

  if (query > 0)
    query_function = query_user;
  if (timelimit > 0)
    timeout_function = timeout;

  /* Read data */
  if (argc > optind){
    if ((a = read_genes(argv[optind], geneformat, seqtype)) == NULL){
      fprintf(stderr, "Could not parse %s as valid data\n", argv[optind]);
      exit(1);
    }
  }
  else{
    if ((a = read_genes(NULL, geneformat, seqtype)) == NULL){
      fprintf(stderr, "Could not parse input as valid data\n");
      exit(1);
    }
  }
  if ((gene_knownancestor) && (seqtype != GENE_BINARY))
    /* First sequnce only included to specify known common ancestor */
    remove_annotatedgene(a, 0);
  g = a->g;

#ifdef HAPLOTYPE_BLOCKS
  if (matrix != NULL){
    /* Set up structures for recovering matrix of bounds for full data
     * set from matrix of bounds for imploded data set.
     */
    n = g->length;
    representativeness = MakeLList();
    for (i = 0; i < g->length; i++){
      c = (SuperColumn *)xmalloc(sizeof(SuperColumn));
      c->left = c->right = i;
      Enqueue(representativeness, (void *)c);
    }
    representativeness_counter = MakeCounter(representativeness, FIRST);
  }
#endif

  implode_genes(g);
#ifdef ENABLE_VERBOSE
  if (v){
    printf("Reduced data set:\n");
    output_genes_indexed(g, NULL);
  }
#endif
  if ((g->length > 0) && (g->n > 0)){
    /* Start by computing haplotype bounds */
    s = genes2sites(g);
    if (heuristic)
      B = haplotype_heuristic_local(s, set, width, increase);
    else
    B = haplotype_bound_local(s);
    local2locals(s->length - 1, B);
    printf("Haplotype bound: %d\n", B[0][s->length - 2]);
    free_sites(s);
#ifdef HAPLOTYPE_BLOCKS
    if (matrix == NULL){
#endif
      printf("Local exact based bound: %d\n",
	     eagl(g, (maxlength ? maxlength : g->length), B, timeout_function,
		  query_function));
#ifdef HAPLOTYPE_BLOCKS
    }
    else{
      /* Replicate B */
      E = (int **)xmalloc((g->length - 1) * sizeof(int *));
      for (i = 0; i < g->length - 1; i++){
	E[i] = (int *)xmalloc((s->length - i - 1) * sizeof(int));
	for (j = 0; j < s->length - i - 1; j++)
	  E[i][j] = B[i][j];
      }
      /* Compute matrix of bounds for reduced data set */
      E = eagl_local(g, (maxlength ? maxlength : g->length), B,
		     timeout_function, query_function);
      local2locals(g->length - 1, E);
      printf("Local exact based bound: %d\n", E[0][g->length - 2]);
      /* Convert to matrices of bounds for original data set */
      /* Start by creating space for the larger matrices */
      B = (int **)xrealloc(B, (n - 1) * sizeof(int *));
      E = (int **)xrealloc(E, (n - 1) * sizeof(int *));
      for (i = 0; i < g->length - 1; i++){
	B[i] = (int *)xrealloc(B[i], (n - i - 1) * sizeof(int));
	E[i] = (int *)xrealloc(E[i], (n - i - 1) * sizeof(int));
      }
      for (; i < n - 1; i++){
	B[i] = (int *)xmalloc((n - i - 1) * sizeof(int));
	E[i] = (int *)xmalloc((n - i - 1) * sizeof(int));
      }
      /* Now do the conversion */
      explode_local(B, representativeness, n);
      explode_local(E, representativeness, n);

      /* Output matrices to file matrix */
      entry_width = (B[0][n - 2] > 0 ? (int)floor(log10(B[0][n - 2])) + 1 : 1);
      format = (char *)xmalloc(5 + (int)floor(log10(entry_width)));
      sprintf(format, "%%%dd ", (int)entry_width);
      for (i = 0; i < n; i++){
	for (j = 0; j < n; j++)
	  if (j < i)
	    /* Print bound between sites j and i from E */
	    fprintf(matrix, format, E[j][i - j - 1]);
	  else if (j > i)
	    /* Print bound between sites i and j from B */
	    fprintf(matrix, format, B[i][j - i - 1]);
	  else{
	    /* Print centered dash */
	    for (k = 0; k < entry_width / 2; k++)
	      fputc(' ', matrix);
	    fputc('-', matrix);
	    for (k = 0; k < (entry_width - 1)/ 2 + 1; k++)
	      fputc(' ', matrix);
	  }
	fputc('\n', matrix);
      }
      /* Clean up */
      for (i = 0; i < n - 1; i++){
	free(B[i]);
	free(E[i]);
      }
      free(B);
      free(E);
      free(format);
    }
#endif
  }
  else{
    printf("Hudson-Kaplan bound: 0\n");
    printf("Haplotype bound: 0\n");
#ifdef HAPLOTYPE_BLOCKS
    if (matrix != NULL){
      for (i = 0; i < n; i++){
	for (j = 0; j < n; j++)
	  fprintf(matrix, (i != j ? "0 " : "- "));
	fprintf(matrix, "\n");
      }
    }
#endif
  }

  /* Clean up */
  free_annotatedgenes(a);
#ifdef HAPLOTYPE_BLOCKS
  if (matrix != NULL){
    while (Length(representativeness) > 0)
      free(Pop(representativeness));
    DestroyLList(representativeness);
    DestroyCounter(representativeness_counter);
  }
#endif

  return 0;
}
