/*******************************************************************

    expression.c
  
    Implementation of methods for constructing and resolving data
    structures holding expressions
		
    Rune Lyngsoe (lyngsoe@stats.ox.ac.uk), June 2007

    Revision History:

********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "expression.h"
#include "elist.h"
#include "llist.h"
#include "lex.yy.c"
#include "common.h"

/* DEBUG */
static void _print_expression(Expression *e)
{
  int i;

  switch(e->type){
  case EXPRESSION_NUMBER:
    printf("Number: %lf\n", e->content.number);
    break;
  case EXPRESSION_BOOLEAN:
    printf("Boolean: %s\n", (e->content.boolean ? "True" : "False"));
    break;
  case EXPRESSION_VARIABLE:
    if (e->content.variable < 0)
      printf("Stack variable at position %d\n", e->content.variable);
    else
      printf("Normal variable at position %d\n", e->content.variable);
    break;
  case EXPRESSION_BUILTIN:
    printf("Built-in function of arity %d\n", e->content.builtin.arity);
    break;
  case EXPRESSION_CONSTANT:
    printf("Built-in constant with value %lf\n", *(e->content.constant));
    break;
  case EXPRESSION_DECLARATION:
    printf("Declaration of function of arity %d\n",
	   e->content.declaration.arity);
    printf("Body is\n");
    _print_expression(e->content.declaration.body);
    printf("\n");
    break;
  case EXPRESSION_APPLICATION:
    printf("Application of function of arity %d\n",
	   e->content.application.arity);
    for (i = 0; i < e->content.application.arity; i++){
      printf("Argument %d is\n", i + 1);
      _print_expression(e->content.application.arguments[i]);
      printf("\n");
    }
    break;
  case EXPRESSION_CHOICE:
    printf("Choice operator with test\n");
    _print_expression(e->content.choice.test);
    printf("\nconsequence\n");
    _print_expression(e->content.choice.consequence);
    printf("\nand alternate\n");
    _print_expression(e->content.choice.alternate);
    printf("\n");
    break;
  case EXPRESSION_ERROR:
    printf("Error with message %s\n", e->content.error);
    break;
  }
}

/* Check whether expression is Boolean; if so, return NULL, otherwise
 * return Expression of type error but with no error message set.
 */
static Expression *_isboolean(Expression *a)
{
  Expression *e = NULL;

  if (a->type != EXPRESSION_BOOLEAN){
    e = (Expression *)xmalloc(sizeof(Expression));
    e->type = EXPRESSION_ERROR;
  }

  return e;
}

/* Check whether expression is numerical; if so, return NULL,
 * otherwise return Expression of type error but with no error message
 * set.
 */
static Expression *_isnumerical(Expression *a)
{
  Expression *e = NULL;

  if (a->type != EXPRESSION_NUMBER){
    e = (Expression *)xmalloc(sizeof(Expression));
    e->type = EXPRESSION_ERROR;
  }

  return e;
}

/* Auxiliary function for handling disjunction */
static Expression *_or(Expression **a)
{
  Expression *e = _isboolean(a[0]);

  if (e == NULL){
    e = _isboolean(a[1]);
    if (e == NULL){
      /* Types match */
      e = (Expression *)xmalloc(sizeof(Expression));
      e->type = EXPRESSION_BOOLEAN;
      e->content.boolean = (a[0]->content.boolean || a[1]->content.boolean);
    }
    else
      /* Type error */
      e->content.error
	= strdup("Disjunction (or) where second argument is non-boolean");
  }
  else
    /* Type error */
    e->content.error
      = strdup("Disjunction (or) where first argument is non-boolean");

  return e;
}

/* Auxiliary function for handling conjunction */
static Expression *_and(Expression **a)
{
  Expression *e = _isboolean(a[0]);

  if (e == NULL){
    e = _isboolean(a[1]);
    if (e == NULL){
      /* Types match */
      e = (Expression *)xmalloc(sizeof(Expression));
      e->type = EXPRESSION_BOOLEAN;
      e->content.boolean = a[0]->content.boolean && a[1]->content.boolean;
    }
    else
      /* Type error */
      e->content.error
	= strdup("Conjunction (and) where second argument is non-boolean");
  }
  else
    /* Type error */
    e->content.error
      = strdup("Conjunction (and) where first argument is non-boolean");

  return e;
}

/* Auxiliary function for handling exclusive or */
static Expression *_xor(Expression **a)
{
  Expression *e = _isboolean(a[0]);

  if (e == NULL){
    e = _isboolean(a[1]);
    if (e == NULL){
      /* Types match */
      e = (Expression *)xmalloc(sizeof(Expression));
      e->type = EXPRESSION_BOOLEAN;
      e->content.boolean = a[0]->content.boolean ^ a[1]->content.boolean;
    }
    else
      /* Type error */
      e->content.error
	= strdup("Exclusive or where second argument is non-boolean");
  }
  else
    /* Type error */
    e->content.error
      = strdup("Exclusive or where first argument is non-boolean");

  return e;
}

/* Auxiliary function for handling Boolean negation */
static Expression *_not(Expression **a)
{
  Expression *e = _isboolean(*a);

  if (e == NULL){
    /* Types match */
    e = (Expression *)xmalloc(sizeof(Expression));
    e->type = EXPRESSION_BOOLEAN;
    e->content.boolean = !(*a)->content.boolean;
  }
  else
    /* Type error */
    e->content.error = strdup("Negation of non-boolean value");

  return e;
}

/* Auxiliary function for handling less than operator */
static Expression *_less(Expression **a)
{
  Expression *e = _isnumerical(a[0]);

  if (e == NULL){
    e = _isnumerical(a[1]);
    if (e == NULL){
      /* Types match */
      e = (Expression *)xmalloc(sizeof(Expression));
      e->type = EXPRESSION_BOOLEAN;
      e->content.boolean = (a[0]->content.number < a[1]->content.number);
    }
    else
      /* Type error */
      e->content.error
	= strdup("Second argument in less than comparison is not a number");
  }
  else
    /* Type error */
    e->content.error
      = strdup("First argument in less than comparison is not a number");

  return e;
}

/* Auxiliary function for handling greater than operator */
static Expression *_greater(Expression **a)
{
  Expression *e = _isnumerical(a[0]);

  if (e == NULL){
    e = _isnumerical(a[1]);
    if (e == NULL){
      /* Types match */
      e = (Expression *)xmalloc(sizeof(Expression));
      e->type = EXPRESSION_BOOLEAN;
      e->content.boolean = (a[0]->content.number > a[1]->content.number);
    }
    else
      /* Type error */
      e->content.error
	= strdup("Second argument in greater than comparison is not a number");
  }
  else
    /* Type error */
    e->content.error
      = strdup("First argument in greater than comparison is not a number");

  return e;
}

/* Auxiliary function for handling less than or equal operator */
static Expression *_lessequal(Expression **a)
{
  Expression *e = _isnumerical(a[0]);

  if (e == NULL){
    e = _isnumerical(a[1]);
    if (e == NULL){
      /* Types match */
      e = (Expression *)xmalloc(sizeof(Expression));
      e->type = EXPRESSION_BOOLEAN;
      e->content.boolean = (a[0]->content.number <= a[1]->content.number);
    }
    else
      /* Type error */
      e->content.error
	= strdup("Second argument in less than or equal comparison is not a number");
  }
  else
    /* Type error */
    e->content.error
      = strdup("First argument in less than or equal comparison is not a number");

  return e;
}

/* Auxiliary function for handling greater than or equal operator */
static Expression *_greaterequal(Expression **a)
{
  Expression *e = _isnumerical(a[0]);

  if (e == NULL){
    e = _isnumerical(a[1]);
    if (e == NULL){
      /* Types match */
      e = (Expression *)xmalloc(sizeof(Expression));
      e->type = EXPRESSION_BOOLEAN;
      e->content.boolean = (a[0]->content.number >= a[1]->content.number);
    }
    else
      /* Type error */
      e->content.error
	= strdup("Second argument in greater than or equal comparison is not a number");
  }
  else
    /* Type error */
    e->content.error
      = strdup("First argument in greater than or equal comparison is not a number");

  return e;
}

/* Auxiliary function for handling equal to operator */
static Expression *_equal(Expression **a)
{
  Expression *e = _isnumerical(a[0]);

  if (e == NULL){
    e = _isnumerical(a[1]);
    if (e == NULL){
      /* Types match */
      e = (Expression *)xmalloc(sizeof(Expression));
      e->type = EXPRESSION_BOOLEAN;
      e->content.boolean = (a[0]->content.number == a[1]->content.number);
    }
    else
      /* Type error */
      e->content.error
	= strdup("Second argument in equal to comparison is not a number");
  }
  else
    /* Type error */
    e->content.error
      = strdup("First argument in equal to comparison is not a number");

  return e;
}

/* Auxiliary function for handling not equal to operator */
static Expression *_notequal(Expression **a)
{
  Expression *e = _isnumerical(a[0]);

  if (e == NULL){
    e = _isnumerical(a[1]);
    if (e == NULL){
      /* Types match */
      e = (Expression *)xmalloc(sizeof(Expression));
      e->type = EXPRESSION_BOOLEAN;
      e->content.boolean = (a[0]->content.number != a[1]->content.number);
    }
    else
      /* Type error */
      e->content.error
	= strdup("Second argument in not equal to comparison is not a number");
  }
  else
    /* Type error */
    e->content.error
      = strdup("First argument in not equal to comparison is not a number");

  return e;
}

/* Auxiliary function for handling addition */
static Expression *_add(Expression **a)
{
  Expression *e = _isnumerical(a[0]);

  if (e == NULL){
    e = _isnumerical(a[1]);
    if (e == NULL){
      /* Types match */
      e = (Expression *)xmalloc(sizeof(Expression));
      e->type = EXPRESSION_NUMBER;
      e->content.number = a[0]->content.number + a[1]->content.number;
    }
    else
      /* Type error */
      e->content.error
	= strdup("Second argument in addition is not a number");
  }
  else
    /* Type error */
    e->content.error
      = strdup("First argument in addition is not a number");

  return e;
}

/* Auxiliary function for handling subtraction */
static Expression *_sub(Expression **a)
{
  Expression *e = _isnumerical(a[0]);

  if (e == NULL){
    e = _isnumerical(a[1]);
    if (e == NULL){
      /* Types match */
      e = (Expression *)xmalloc(sizeof(Expression));
      e->type = EXPRESSION_NUMBER;
      e->content.number = a[0]->content.number - a[1]->content.number;
    }
    else
      /* Type error */
      e->content.error
	= strdup("Second argument in subtraction is not a number");
  }
  else
    /* Type error */
    e->content.error
      = strdup("First argument in subtraction is not a number");

  return e;
}

/* Auxiliary function for handling multiplication */
static Expression *_mul(Expression **a)
{
  Expression *e = _isnumerical(a[0]);

  if (e == NULL){
    e = _isnumerical(a[1]);
    if (e == NULL){
      /* Types match */
      e = (Expression *)xmalloc(sizeof(Expression));
      e->type = EXPRESSION_NUMBER;
      e->content.number = a[0]->content.number * a[1]->content.number;
    }
    else
      /* Type error */
      e->content.error
	= strdup("Second argument in multiplication is not a number");
  }
  else
    /* Type error */
    e->content.error
      = strdup("First argument in multiplication is not a number");

  return e;
}

/* Auxiliary function for handling division */
static Expression *_div(Expression **a)
{
  Expression *e = _isnumerical(a[0]);

  if (e == NULL){
    e = _isnumerical(a[1]);
    if (e == NULL){
      /* Types match */
      e = (Expression *)xmalloc(sizeof(Expression));
      e->type = EXPRESSION_NUMBER;
      e->content.number = a[0]->content.number / a[1]->content.number;
    }
    else
      /* Type error */
      e->content.error
	= strdup("Second argument in division is not a number");
  }
  else
    /* Type error */
    e->content.error
      = strdup("First argument in division is not a number");

  return e;
}

/* Auxiliary function for handling numerical negation */
static Expression *_neg(Expression **a)
{
  Expression *e = _isnumerical(*a);

  if (e == NULL){
    /* Type match */
    e = (Expression *)xmalloc(sizeof(Expression));
    e->type = EXPRESSION_NUMBER;
    e->content.number = -(*a)->content.number;
  }
  else
    /* Type error */
    e->content.error
      = strdup("Negated value is not a number");

  return e;
}

/* Auxiliary function for handling exponentiation */
static Expression *_exp(Expression **a)
{
  Expression *e = _isnumerical(a[0]);

  if (e == NULL){
    e = _isnumerical(a[1]);
    if (e == NULL){
      /* Types match */
      e = (Expression *)xmalloc(sizeof(Expression));
      e->type = EXPRESSION_NUMBER;
      e->content.number = pow(a[0]->content.number, a[1]->content.number);
    }
    else
      /* Type error */
      e->content.error
	= strdup("Exponent in exponentiation is not a number");
  }
  else
    /* Type error */
    e->content.error
      = strdup("Base in exponentiation is not a number");

  return e;
}

/* Add name to namespace */
static void _addname(EList *namespace, char *name)
{
  Declaration *d = (Declaration *)xmalloc(sizeof(Declaration));

  d->name = name;
  elist_append(namespace, d);
}

/* Add function to heap */
static void _addfunction(EList *heap, Expression *(*f)(Expression **),
			 int arity)
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));

  e->type = EXPRESSION_BUILTIN;
  e->content.builtin.f = f;
  e->content.builtin.arity = arity;
  elist_append(heap, e);
}

/* Add numerical value to heap */
static void _addvalue(EList *heap, double *value)
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));

  e->type = EXPRESSION_CONSTANT;
  e->content.constant = value;
  elist_append(heap, e);
}

/* Set up a builtin function f for program p. Functions taking no
 * parameters are treated as builtin variables, and hence can be used
 * for variables that are only computed on demand.
 */
void expression_buildinfunction(Program *p, Expression *(*f)(Expression **),
				int arity, char *name)
{
  _addfunction(p->builtin, f, arity);
  _addname(p->namespace, name);
}

/* Set up a list f of builtin functions for p; f should be a list of
 * three-tuples, where the first element in each tuple is the name of
 * the function, the second element is a pointer to the function, and
 * the third element is the arity of the function. The list should end
 * with a dummy entry where the name is NULL.
 */
void expression_buildinfunctions(Program *p, void *f[][3])
{
  int i;

  for (i = 0; f[i][0] != NULL; i++)
    expression_buildinfunction(p, (Expression *(*)(Expression **))f[i][1],
			       (int)f[i][2], (char *)f[i][0]);
}

/* Set up a builtin variable. Builtin variables need to be supplied by
 * reference to their location so the value can be changed between
 * successive runs. If the variable should only be computed if
 * actually used, use builtin functions of arity 0 (see
 * buildinfunction) instead.
 */
void expression_buildinvariable(Program *p, double *value, char *name)
{
  _addvalue(p->builtin, value);
  _addname(p->namespace, name);
}

/* Set up a list v of builtin variables for p; v should be a list of
 * pairs, where the first element in each pair is the name of the
 * variable and the second element is a pointer to the location where
 * the value of the variable is stored. The list should end with a
 * dummy entry where the name is NULL.
 */
void expression_buildinvariables(Program *p, void *f[][2])
{
  int i;

  for (i = 0; f[i][0] != NULL; i++)
    expression_buildinvariable(p, (double *)f[i][1], (char *)f[i][0]);
}

/* List of operator functions with arity */
static void *_operator[][2] = {
  {_or, (void *)2},
  {_and, (void *)2},
  {_xor, (void *)2},
  {_not, (void *)1},
  {_less, (void *)2},
  {_greater, (void *)2},
  {_lessequal, (void *)2},
  {_greaterequal, (void *)2},
  {_equal, (void *)2},
  {_notequal, (void *)2},
  {_add, (void *)2},
  {_sub, (void *)2},
  {_mul, (void *)2},
  {_div, (void *)2},
  {_neg, (void *)1},
  {_exp, (void *)2},
  {NULL, (void *)0}
};

/* Initialise heap and namespace with functions corresponding to
 * operators as nameless builtin functions.
 */
Program *expression_initialise()
{
  Program *p = (Program *)xmalloc(sizeof(Program));
  int i;

  p->builtin = elist_make();
  p->namespace = elist_make();
  p->program = NULL;
  for (i = 0; _operator[i][0] != NULL; i++)
    expression_buildinfunction(p, _operator[i][0], (int)_operator[i][1], NULL);

  return p;
}

/* Recursively release all memory used by expression e and its
 * subexpressions.
 */
static void _destroyexpression(Expression *e, int freedeclarations)
{
  int i;

  /* First free subexpressions and structures allocated within e */
  switch(e->type){
  case EXPRESSION_NUMBER:
  case EXPRESSION_BOOLEAN:
  case EXPRESSION_VARIABLE:
  case EXPRESSION_BUILTIN:
  case EXPRESSION_CONSTANT:
    break;
  case EXPRESSION_DECLARATION:
    if (freedeclarations)
      _destroyexpression(e->content.declaration.body, 1);
    break;
  case EXPRESSION_APPLICATION:
    for (i = 0; i < e->content.application.arity; i++)
      _destroyexpression(e->content.application.arguments[i],
				    freedeclarations);
    free(e->content.application.arguments);
    break;
  case EXPRESSION_CHOICE:
    _destroyexpression(e->content.choice.test, freedeclarations);
    _destroyexpression(e->content.choice.consequence,
				  freedeclarations);
    _destroyexpression(e->content.choice.alternate,
				  freedeclarations);
    break;
  case EXPRESSION_ERROR:
    free(e->content.error);
    break;
  }
  /* All done, now free e */
  free(e);
}

/* Free memory used by p; free_function (if not NULL) is applied to
 * user defined builtin heap functions, free_variable (if not NULL) is
 * applied to user defined builtin variables, and free_name (if not
 * NULL) is applied to names of user defined builtin functions and
 * variables.
 */
void expression_destroy(Program *p,
			void (*free_function)
			(Expression *(*)(Expression **)),
			void (*free_variable)(double *),
			void (*free_name)(char *))
{
  Expression *e;
  Declaration *d;
  int i, j;

  /* Release space used for builtin functions and variables and their names */
  for (i = 0; i < elist_length(p->builtin); i++){
    e = (Expression *)elist_get(p->builtin, i);
    d = (Declaration *)elist_get(p->namespace, i);
    if (e->type == EXPRESSION_BUILTIN){
      /* Builtin function declaration */
      if ((free_function != NULL) || (free_name != NULL)){
	/* Action needs to be taken on user defined builtin functions */
	for (j = 0; _operator[j][0] != NULL; j++)
	  if (_operator[j][0] == e->content.builtin.f)
	    /* One of the builtin operator functions */
	    break;
	if (_operator[j][0] == NULL){
	  /* User defined builtin function */
	  if (free_function != NULL)
	    free_function(e->content.builtin.f);
	  if (free_name != NULL)
	    free_name(d->name);
	}
      }
    }
    else if (e->type == EXPRESSION_CONSTANT){
      /* Builtin variable declaration */
      if (free_variable != NULL)
	free_variable(e->content.constant);
      if (free_name != NULL)
	free_name(d->name);
    }
    else{
      /* Remnant of program evaluation - this should really have been
       * cleared at the end of the evaluation.
       */
      fprintf(stderr, "WARNING: Bug in expression evaluation, please email error report\n");
      _destroyexpression(e, 0);
    }
    free(e);
    free(d);
  }
  for (i = elist_length(p->builtin); i < elist_length(p->namespace); i++){
    d = elist_get(p->namespace, i);
    free(d->name);
    free(d);
  }
  elist_destroy(p->builtin);
  elist_destroy(p->namespace);

  /* Release space used for the actual program */
  if (p->program != NULL){
    for (i = 0; i < elist_length(p->program); i++)
      _destroyexpression((Expression *)elist_get(p->program, i), 1);
    elist_destroy(p->program);
  }

  free(p);
}

/* Find index of last occurrence of name in namespace; if name is not
 * in namespace, -1 is returned.
 */
static int _lookupname(EList *namespace, char *name)
{
  /* Search for most recent addition of name in namespace */
  Declaration *d;
  int i;

  for (i = elist_length(namespace) - 1; i >= 0; i--){
    d = (Declaration *)elist_get(namespace, i);
    if ((d->name != NULL) && (strcmp(d->name, name) == 0))
      break;
  }

  return i;
}

/* Find index of last occurrence of f in heap; if f is not in heap, -1
 * is returned.
 */
static int _lookupfunction(EList *heap, Expression *(*f)(Expression **))
{
  /* Search for most recent addition of f to heap */
  Expression *e;
  int i;

  for (i = elist_length(heap) - 1; i >= 0; i--){
    e = (Expression *)elist_get(heap, i);
    if ((e->type == EXPRESSION_BUILTIN) && (e->content.builtin.f == f))
      break;
  }

  return i;
}

/* Generic function for converting operator to builtin heap function */
static Expression *_functionalise(int f, int arity, ...)
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));
  va_list args;
  int i;

  e->type = EXPRESSION_APPLICATION;
  e->content.application.f = f;
  e->content.application.arity = arity;
  if (arity > 0){
    e->content.application.arguments
      = (Expression **)xmalloc(arity * sizeof(Expression *));
    va_start(args, arity);
    for (i = 0; i < arity; i++)
      e->content.application.arguments[i]
	= (Expression *)va_arg(args, Expression *);
    va_end(args);
  }

  return e;
}

/* Parse or operator */
Expression *expression_parseor(Program *p, Expression *a, Expression *b)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _or);

  return _functionalise(index, 2, a, b);
}

/* Parse and operator */
Expression *expression_parseand(Program *p, Expression *a, Expression *b)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _and);

  return _functionalise(index, 2, a, b);
}

/* Parse xor operator */
Expression *expression_parsexor(Program *p, Expression *a, Expression *b)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _xor);

  return _functionalise(index, 2, a, b);
}

/* Parse Boolean negation operator */
Expression *expression_parsenot(Program *p, Expression *a)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _not);

  return _functionalise(index, 1, a);
}

/* Parse less than operator */
Expression *expression_parseless(Program *p, Expression *a, Expression *b)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _less);

  return _functionalise(index, 2, a, b);
}

/* Parse greater than operator */
Expression *expression_parsegreater(Program *p, Expression *a, Expression *b)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _greater);

  return _functionalise(index, 2, a, b);
}

/* Parse less than or equal to operator */
Expression *expression_parselessequal(Program *p, Expression *a, Expression *b)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _lessequal);

  return _functionalise(index, 2, a, b);
}

/* Parse greater than or equal to operator */
Expression *expression_parsegreaterequal(Program *p, Expression *a, Expression *b)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _greaterequal);

  return _functionalise(index, 2, a, b);
}

/* Parse equal to operator */
Expression *expression_parseequal(Program *p, Expression *a, Expression *b)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _equal);

  return _functionalise(index, 2, a, b);
}

/* Parse not equal to operator */
Expression *expression_parsenotequal(Program *p, Expression *a, Expression *b)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _notequal);

  return _functionalise(index, 2, a, b);
}

/* Parse addition operator */
Expression *expression_parseadd(Program *p, Expression *a, Expression *b)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _add);

  return _functionalise(index, 2, a, b);
}

/* Parse subtraction operator */
Expression *expression_parsesub(Program *p, Expression *a, Expression *b)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _sub);

  return _functionalise(index, 2, a, b);
}

/* Parse multiplication operator */
Expression *expression_parsemul(Program *p, Expression *a, Expression *b)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _mul);

  return _functionalise(index, 2, a, b);
}

/* Parse division operator */
Expression *expression_parsediv(Program *p, Expression *a, Expression *b)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _div);

  return _functionalise(index, 2, a, b);
}

/* Parse arithmetic negation operator */
Expression *expression_parseneg(Program *p, Expression *a)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _neg);

  return _functionalise(index, 1, a);
}

/* Parse exponentiation operator */
Expression *expression_parseexp(Program *p, Expression *a, Expression *b)
{
  static int index = -1;

  /* Only look up index first time */
  if (index == -1)
    index = _lookupfunction(p->builtin, _exp);

  return _functionalise(index, 2, a, b);
}

/* Combine a parse of definitions followed by a final expression into
 * a full program.
 */
Program *expression_parseprogram(Program *p, Expression *e)
{
  /* By convention, last expression in program is the value of the program */
  elist_append(p->program, e);

  return p;
}

/* Add extra definition to program */
Program *expression_parsedefinitions(Program *p, Program *e)
{
  /* Definition should already have been added to p */
  return p;
}

/* Add definition of variable to list of definitions */
Program *expression_parsevariabledefinition(Program *p, char *name,
					    Expression *e)
{
  _addname(p->namespace, name);
  elist_append(p->program, e);

  return p;
}

/* Add function name and parameter names to namespace */
Program *expression_parsefunctiondefinitionhead(Program *p, LList *newnames)
{
  while (Length(newnames) > 0)
    _addname(p->namespace, (char *)Pop(newnames));
  DestroyLList(newnames);

  return p;
}

/* Add body of function definition to program and remove parameter
 * names from namespace.
 */
Program *expression_parsefunctiondefinition(Program *p, Expression *e)
{
  Expression *tmp = (Expression *)xmalloc(sizeof(Expression));
  Declaration *d;
  int i;

  elist_append(p->program, tmp);
  tmp->type = EXPRESSION_DECLARATION;
  tmp->content.declaration.body = e;
  tmp->content.declaration.arity = elist_length(p->namespace)
    - elist_length(p->program) - elist_length(p->builtin);

  /* Remove parameters of function from namespace */
  for (i = 0; i < tmp->content.declaration.arity; i++){
    d = elist_removelast(p->namespace);
    free(d->name);
    free(d);
  }

  return p;
}

/* Add name of function being declared to the beginning of a list of
 * parameter names defined in the local scope of the function.
 */
LList *expression_parsedeclaration(char *name, LList *locals)
{
  Push(locals, name);

  return locals;
}

/* Parse first parameter in a list possibly containing more parameters */
LList *expression_parsefirstparameter(char *name)
{
  LList *tmp = MakeLList();
  Enqueue(tmp, name);

  return tmp;
}

/* Add one more parameter to existing list of parameter names */
LList *expression_parseparameters(LList *p, char *name)
{
  Enqueue(p, name);

  return p;
}

/* Parse a numerical value into an expression */
Expression *expression_parsenumber(double x)
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));

  e->type = EXPRESSION_NUMBER;
  e->content.number = x;

  return e;
}

/* Parse a variable name into an expression */
static int *_expression_builtinused = NULL;
Expression *expression_parsevariable(Program *p, char *name)
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));
  int i = _lookupname(p->namespace, name);

  /* Check whether name was defined */
  if (i < 0){
    fprintf(stderr, "Error in expression - variable\n\n %s\n\nis not defined in current context\n", name);
    exit(1);
  }

  /* Check whether name refers to local (i.e. is a function parameter)
   * or global scope.
   */
  if (i > elist_length(p->builtin) + elist_length(p->program))
    i -= elist_length(p->namespace);
  else if ((i < elist_length(p->builtin))
	   && (_expression_builtinused != NULL))
    _expression_builtinused[i] = 1;

  /* Initialise expression for this variable */
  e->type = EXPRESSION_VARIABLE;
  e->content.variable = i;

  /* Clean up */
  free(name);

  return e;
}

/* Parse expression for pursuing consequence c or alternative a based
 * on the Boolean value of t.
 */
Expression *expression_parsechoice(Expression *t, Expression *c, Expression *a)
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));

  e->type = EXPRESSION_CHOICE;
  e->content.choice.test = t;
  e->content.choice.consequence = c;
  e->content.choice.alternate = a;

  return e;
}

/* Parse application of function, which can be either builtin or
 * previously declared and reside in either local or global scope.
 */
Expression *expression_parseapplication(Program *p, char *name, LList *a)
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));
  int i = _lookupname(p->namespace, name);

  /* Check whether name was defined */
  if (i < 0){
    fprintf(stderr, "Error in expression - function\n\n %s\n\nis not defined in current context\n", name);
    exit(1);
  }

  /* Check whether name refers to local (i.e. is a function parameter)
   * or global scope.
   */
  if (i > elist_length(p->builtin) + elist_length(p->program))
    i -= elist_length(p->namespace);
  else if ((i < elist_length(p->builtin))
	   && (_expression_builtinused != NULL))
    _expression_builtinused[i] = 1;

  /* Initialise expression for this variable */
  e->type = EXPRESSION_APPLICATION;
  e->content.application.f = i;
  e->content.application.arity = Length(a);
  e->content.application.arguments = (Expression **)
    xmalloc(e->content.application.arity * sizeof(Expression *));
  for (i = 0; i < e->content.application.arity; i++)
    e->content.application.arguments[i] = (Expression *)Pop(a);

  /* Clean up */
  free(name);
  DestroyLList(a);

  return e;
}

/* Parse first argument in a list possibly containing more arguments */
LList *expression_parsefirstargument(Expression *e)
{
  LList *tmp = MakeLList();
  Enqueue(tmp, e);

  return tmp;
}

/* Add one more argument to existing list of argumentss */
LList *expression_parsearguments(LList *a, Expression *e)
{
  Enqueue(a, e);

  return a;
}

/* Parse statements into program prg. prg is assumed to be properly
 * initialised; if it furthermore already contains a program, this
 * will be erased. Names will be removed from the namespace, but
 * memory used to store them will not be deallocated.
 */
Program *expression_currentprogram;
int expression_definitioncount;
int expression_currentdefinition;
void expression_parse(Program *prg, char *statements)
{
  int i;
  YY_BUFFER_STATE buffer;

  if (prg->program != NULL)
    /* prg contains a previous program */
    for (i = 0; i < elist_length(prg->program); i++){
      _destroyexpression((Expression *)elist_removelast(prg->program), 1);
      elist_removelast(prg->namespace);
    }
  else
    prg->program = elist_make();

  /* Determine number of definitions in statement (if well formed) */
  expression_definitioncount = expression_currentdefinition = 0;
  for (i = 0; statements[i] != '\0'; i++)
    if (statements[i] == ';')
      expression_definitioncount++;

  /* Make prg available to bison and set up flex to read tokens from
   * statements.
   */
  buffer = yy_scan_string(statements);
  expression_currentprogram = prg;

  /* Invoke bison generated parser */
  yyparse();

  /* Clean up */
  yy_delete_buffer(buffer);

}

/* Parse statements into program prg. prg is assumed to be properly
 * initialised; if it furthermore already contains a program, this
 * will be erased. Names will be removed from the namespace, but
 * memory used to store them will not be deallocated. A list of the
 * names of all builtin functions and variables referenced in
 * statements is returned. The names will be returned as pointers to
 * the strings stored in namespace of prg, and these should thus not
 * be deallocated and will become invalid when prg or its namespace is
 * deallocated.
 */
LList *expression_parse_and_check(Program *prg, char *statements)
{
  LList *checklist = MakeLList();
  Declaration *d;
  int i;

  if (elist_length(prg->builtin) > 0)
    _expression_builtinused
      = (int *)xcalloc(elist_length(prg->builtin), sizeof(int));
  expression_parse(prg, statements);
  if (elist_length(prg->builtin) > 0){
    for (i = 0; i < elist_length(prg->builtin); i++)
      if (_expression_builtinused[i]){
	d = (Declaration *)elist_get(prg->namespace, i);
	Push(checklist, d->name);
      }
    free(_expression_builtinused);
    _expression_builtinused = NULL;
  }

  return checklist;
}

/* Extend s with t, reallocating to ensure sufficient space */
static void _extend_string(char **s, char *t){
  *s = xrealloc(*s, strlen(*s) + strlen(t) + 1);
  strcat(*s, t);
}

/* Evaluate e in the context of stack */
static Expression *_interpret_expression(Expression *e, EList *stack)
{
  Expression *tmp, *f, **p;
  EList *arguments;
  char *s;
  int i;

  switch(e->type){
  case EXPRESSION_NUMBER:
  case EXPRESSION_BOOLEAN:
    /* Basic types cannot be further evaluated */
    /* Return a copy of the basic type expression, as all expressions
     * returned by this function will eventually be freed.
     */
    tmp = (Expression *)xmalloc(sizeof(Expression));
    memcpy(tmp, e, sizeof(Expression));
    break;
  case EXPRESSION_VARIABLE:
    /* Look up variables in the environment */
    /* As for basic type expressions we need to return a copy of the
     * variable that can safely be freed.
     */
    if (e->content.variable >= 0)
      /* Non-negative adresses are heap variables which are located
       * beneath the stack variables, i.e. at the bottom of the stack
       */
      e = elist_get(stack, e->content.variable);
    else
      /* Negative adresses are stack variables, and the address is the
       * distance to the top of the stack with -1 being the topmost
       * element.
       */
      e = elist_get(stack, elist_length(stack) + e->content.variable);
    tmp = _interpret_expression(e, stack);
#ifdef FOO
    if ((e->type != EXPRESSION_BUILTIN) || (e->content.builtin.arity > 0)){
      /* Standard use of variable */
      tmp = xmalloc(sizeof(Expression));
      if (e->type == EXPRESSION_CONSTANT){
	/* Variable refers to builtin constant - copy its value to
	 * expression.
	 */
	tmp->type = EXPRESSION_NUMBER;
	tmp->content.number = *(e->content.constant);
      }
      else
	/* Variable is either peviously declared variable or a builtin
	 * or user-defined function.
	 */
	memcpy(tmp, e, sizeof(Expression));
    }
    else
      /* Variable refers to parameter-less function. These are used
       * for built-in variables that are only computed on demand.
       */
      tmp = ((Expression *(*)())e->content.builtin.f)();
#endif
    break;
  case EXPRESSION_CONSTANT:
    /* Lookup current value of user defined (builtin) variable */
    tmp = (Expression *)xmalloc(sizeof(Expression));
    tmp->type = EXPRESSION_NUMBER;
    tmp->content.number = *(e->content.constant);
    break;
  case EXPRESSION_DECLARATION:
    /* User defined function */
    tmp = e;
    break;
  case EXPRESSION_BUILTIN:
    /* Builtin function */
    if (e->content.builtin.arity > 0)
      /* Standard use */
      tmp = e;
    else
      /* Functions of arity 0 are used for built-in variables that are
       * only computed on demand.
       */
      tmp = ((Expression *(*)())e->content.builtin.f)();
    break;
  case EXPRESSION_APPLICATION:
    /* Application of a function, which is defined in the environment */
    if (e->content.application.f >= 0){
      /* Again, non-negative addresses refer to heap elements */
      f = (Expression *)elist_get(stack, e->content.application.f);
    }
    else{
      /* And negative addresses refer to stack elements */
      f = (Expression *)elist_get(stack, elist_length(stack)
				  + e->content.application.f);
    }
    if ((f->type != EXPRESSION_DECLARATION)
	&& ((f->type != EXPRESSION_BUILTIN)
	    || (f->content.builtin.arity == 0))){
      /* Attempt to apply something that is not a function */
      tmp = xmalloc(sizeof(Expression));
      tmp->type = EXPRESSION_ERROR;
      if (e->content.application.f >= 0)
	tmp->content.error = strdup("Attempt to use previously declared variable as a function, despite it not being a function");
      else
	tmp->content.error = strdup("Attempt to use parameter as a function, despite it not being a function");
    }
    else if (f->type == EXPRESSION_DECLARATION){
      /* Application of user defined function */
      /* Check whether arities match */
      if (f->content.declaration.arity != e->content.application.arity){
	/* Arity error */
	tmp = (Expression *)xmalloc(sizeof(Expression));
	tmp->type = EXPRESSION_ERROR;
	tmp->content.error
	  = strdup("Application of previously declared function ");
	if (e->content.application.f < 0)
	  _extend_string(&(tmp->content.error), "(provided as argument) ");
	_extend_string(&(tmp->content.error), "of ");
	s = i2a(f->content.declaration.arity);
	_extend_string(&(tmp->content.error), s);
	free(s);
	if (f->content.declaration.arity != 1)
	  _extend_string(&(tmp->content.error), " parameters to ");
	else
	  _extend_string(&(tmp->content.error), " parameter to ");
	s = i2a(e->content.application.arity);
	_extend_string(&(tmp->content.error), s);
	free(s);
	if (e->content.application.arity != 1)
	  _extend_string(&(tmp->content.error), "arguments");
	else
	  _extend_string(&(tmp->content.error), "arguments");
      }
      else{
	/* Compute list of arguments and check for errors */
	arguments = elist_make();
	for (i = 0; i < e->content.application.arity; i++){
	  tmp = _interpret_expression(e->content.application.arguments[i],
				      stack);
	  if (tmp->type == EXPRESSION_ERROR)
	    /* Subevaluation failed */
	    return tmp;
	  /* Add argument to stack */
	  elist_append(arguments, tmp);
	}
	elist_extend(stack, arguments);
	/* Evaluate body in context of parameters */
	tmp = _interpret_expression(f->content.declaration.body, stack);
	/* Return stack to previous status */
	for (i = 0; i < e->content.application.arity; i++){
	  f = (Expression *)elist_removelast(stack);
	  if ((f->type != EXPRESSION_DECLARATION)
	      && (f->type != EXPRESSION_BUILTIN))
	    free(f);
	}
      }

    }
    else{
      /* Application of builtin function */
      /* Check whether arities match */
      if (f->content.builtin.arity != e->content.application.arity){
	/* Arity error */
	tmp = (Expression *)xmalloc(sizeof(Expression));
	tmp->type = EXPRESSION_ERROR;
	if (f->content.builtin.arity > 0){
	  tmp->content.error
	    = strdup("Application of builtin function ");
	  if (e->content.application.f < 0)
	    _extend_string(&(tmp->content.error), "(provided as argument) ");
	  _extend_string(&(tmp->content.error), "of ");
	  s = i2a(f->content.builtin.arity);
	  _extend_string(&(tmp->content.error), s);
	  free(s);
	  if (f->content.builtin.arity != 1)
	    _extend_string(&(tmp->content.error), " parameters to ");
	  else
	    _extend_string(&(tmp->content.error), " parameter to ");
	  s = i2a(e->content.application.arity);
	  _extend_string(&(tmp->content.error), s);
	  free(s);
	  if (e->content.application.arity != 1)
	    _extend_string(&(tmp->content.error), " arguments");
	  else
	    _extend_string(&(tmp->content.error), " argument");
	}
	else
	  tmp->content.error = strdup("Use of builtin variable as function");
      }
      else{
	/* Compute list of parameters and check for errors */
	p = (Expression **)
	  xmalloc(e->content.application.arity * sizeof(Expression *));
	for (i = 0; i < e->content.application.arity; i++){
	  p[i] = _interpret_expression(e->content.application.arguments[i],
				       stack);
	  if (p[i]->type == EXPRESSION_ERROR){
	    /* Subevaluation failed */
	    tmp = p[i];
	    for (i--; i >= 0; i --)
	      _destroyexpression(p[i], 0);
	    free(p);
	    return tmp;
	  }
	}
	/* Apply builtin function */
	tmp = f->content.builtin.f(p);
	/* Clean up */
	for (i = 0; i < e->content.application.arity; i++)
	  _destroyexpression(p[i], 0);
	free(p);
      }
    }
    break;
  case EXPRESSION_CHOICE:
    /* Choice operator cannot be hidden away in builtin function, as
     * only the expression chosen should be evaluated (to avoid
     * infinite recursion).
     */
    f = _interpret_expression(e->content.choice.test, stack);
    if (f->type != EXPRESSION_BOOLEAN){
      if (f->type != EXPRESSION_ERROR){
	/* Test does not evaluate to Boolean value */
	f->type = EXPRESSION_ERROR;
	f->content.error = strdup("Choice test does not evaluate to Boolean value");
      }
      tmp = f;
    }
    else{
      if (f->content.boolean)
	/* Pursue consequence branch */
	tmp = _interpret_expression(e->content.choice.consequence, stack);
      else
	tmp = _interpret_expression(e->content.choice.alternate, stack);
      free(f);
    }
    break;
  default:
    fprintf(stderr, "Bug in expression interpreter - please email error report\n");
    exit(1);
    break;
  }

  return tmp;
}

/* Interpret a full program */
double expression_interpret(Program *prg)
{
  int i;
  double r = 0;
  Expression *e;
  /* Build on top of builtins - this is OK as long as we remove
   * exactly what we have added afterwards, and as long as parallel
   * instantiations of this function are not running at the same
   * time. The way builtin variables are handled precludes simple
   * parallelisation of this module, anyway.
   */
  EList *stack = prg->builtin;

  for (i = 0; i < elist_length(prg->program) - 1; i++){
    e = (Expression *)elist_get(prg->program, i);
    if (e->type == EXPRESSION_DECLARATION)
      elist_append(stack, e);
    else{
      e = _interpret_expression(e, stack);
      if (e->type == EXPRESSION_ERROR){
	fprintf(stderr, "%s\nin declaration number %d\n", e->content.error,
		i + 1);
	exit(1);
      }
      elist_append(stack, e);
    }
  }

  /* Check outcome of evaluation */
  e = (Expression *)elist_get(prg->program, elist_length(prg->program) - 1);
  e = _interpret_expression(e, stack);
  switch(e->type){
  case EXPRESSION_ERROR:
    fprintf(stderr, "%s\nin final expression\n", e->content.error);
    exit(1);
    break;
  case EXPRESSION_NUMBER:
    r = e->content.number;
    break;
  case EXPRESSION_BOOLEAN:
    fprintf(stderr, "Final expression evaluates to Boolean\n");
    exit(1);
    break;
  case EXPRESSION_BUILTIN:
  case EXPRESSION_DECLARATION:
    fprintf(stderr, "Final expression evaluates to a function\n");
    exit(1);
    break;
  default:
    fprintf(stderr, "Bug in expression interpretation - please email error report\n");
    exit(1);
  }

  /* Restore builtins and free memory used by interpretation */
  free(e);
  for (i = 1; i < elist_length(prg->program); i++){
    e = (Expression *)elist_removelast(stack);
    if ((e->type != EXPRESSION_DECLARATION)
	&& (e->type != EXPRESSION_BUILTIN))
      free(e);
  }

  return r;
}
