/* Expression for scoring an ancestral state by recombination complexity */
%{
#include <stdlib.h>
#include <stdio.h>
#include "expression.h"
#include "expression.bison.h"

void yyerror(char *s)
{
  fprintf(stderr, "Error in expression parse:\n\n %s\n", s);

  exit(1);
}

int yylex();

%}

%union {
  Program *program;
  Expression *exp;
  LList *list;
  double number;
  char *identifier;
}

%token <number> NUMBER
%token <identifier> IDENTIFIER DEFNAME
%left CHOOSE
%left ALTERNATIVE
%left OR
%left XOR
%left AND
%left NOT
%nonassoc LESS GREATER LESSEQUAL GREATEREQUAL EQUAL NOTEQUAL '='
%left '-' ADD
%left MUL DIV
%left NEG
%right EXP
%type <program> program definitions definition
%type <exp> expression application
%type <list> declaration parameters arguments
%type <identifier> parameter name

%% /* Grammar specification of expression */

/* Complete specification of how score is computed */
program:     /* empty */ { $$ = NULL; }
             | definitions expression { $$ = expression_parseprogram($1, $2); }
;

/* Initial list of auxiliary definitions */
definitions: /* empty */ { $$ = expression_currentprogram; }
         | definitions definition { $$ = expression_parsedefinitions($1, $2); }
;

/* A single definition of either a variable or a function */
definition:  DEFNAME '=' expression ';'
     { expression_parsevariabledefinition(expression_currentprogram, $1, $3); }
           | declaration
 { $$ = expression_parsefunctiondefinitionhead(expression_currentprogram, $1);}
              '=' expression ';'
    { $$ = expression_parsefunctiondefinition(expression_currentprogram, $4); }
;

/* Function being defined */
declaration:  DEFNAME '(' parameters ')'
             { $$ = expression_parsedeclaration($1, $3); }
;

/* List of parameters to function being defined */
parameters:  parameter { $$ = expression_parsefirstparameter($1); }
           | parameters ',' parameter
                                   { $$ = expression_parseparameters($1, $3); }
;

/* A single parameter */
parameter:   IDENTIFIER { $$ = $1; }
;

/* The core rule defining expressions */
expression:  NUMBER { $$ = expression_parsenumber($1); }
           | IDENTIFIER
              { $$ = expression_parsevariable(expression_currentprogram, $1); }
           | application { $$ = $1; }
           | expression ADD expression
               { $$ = expression_parseadd(expression_currentprogram, $1, $3); }
           | expression '-' expression
               { $$ = expression_parsesub(expression_currentprogram, $1, $3); }
           | expression MUL expression
                 {$$ = expression_parsemul(expression_currentprogram, $1, $3);}
           | expression DIV expression
               { $$ = expression_parsediv(expression_currentprogram, $1, $3); }
           | expression EXP expression
                 {$$ = expression_parseexp(expression_currentprogram, $1, $3);}
           | '-' expression %prec NEG
                   { $$ = expression_parseneg(expression_currentprogram, $2); }
           | expression CHOOSE expression ALTERNATIVE expression
                                   { $$ = expression_parsechoice($1, $3, $5); }
           | '(' expression ')' { $$ = $2; }
           | NOT expression
                   { $$ = expression_parsenot(expression_currentprogram, $2); }
           | expression AND expression
               { $$ = expression_parseand(expression_currentprogram, $1, $3); }
           | expression XOR expression
                {$$ = expression_parsexor(expression_currentprogram, $1, $3); }
           | expression OR expression
                { $$ = expression_parseor(expression_currentprogram, $1, $3); }
           | expression LESS expression
              { $$ = expression_parseless(expression_currentprogram, $1, $3); }
           | expression GREATER expression
           { $$ = expression_parsegreater(expression_currentprogram, $1, $3); }
           | expression LESSEQUAL expression
           {$$ = expression_parselessequal(expression_currentprogram, $1, $3);}
           | expression GREATEREQUAL expression
        {$$ = expression_parsegreaterequal(expression_currentprogram, $1, $3);}
           | expression EQUAL expression
             { $$ = expression_parseequal(expression_currentprogram, $1, $3); }
           | expression '=' expression
             { $$ = expression_parseequal(expression_currentprogram, $1, $3); }
           | expression NOTEQUAL expression
          { $$ = expression_parsenotequal(expression_currentprogram, $1, $3); }
	  ;

/* Function being applied */
application:  name '(' arguments ')'
       { $$ = expression_parseapplication(expression_currentprogram, $1, $3); }
;

/* Name of function */
name: IDENTIFIER { $$ = $1; }
;

/* List of arguments to function application */
arguments:   expression { $$ = expression_parsefirstargument($1); }
           | arguments ',' expression { $$ = expression_parsearguments($1, $3); }
;

/* That's it, and that's that */

%%
