#!/usr/bin/python
from sys import stdin, argv, stderr, stdout, exit
from copy import deepcopy
from operator import add
from types import IntType

# getline(f): returns next non-empty, non-comment line from f, None if EOF
def getline(f):
  while 1:
    s = f.readline()
    if s == "":
      return None
    # Strip off trailing comment
    escaped = 0
    for i in range(0, len(s)):
      if s[i] == '#' and not escaped:
        s = s[:i]
        break
      if s[i] == '\\':
        escaped = not escaped
    # Strip off trailing whitespace
    s = s.rstrip()
    if s != "":
      break
  return s

# Set up input and output channels
outfile = None
if len(argv) > 1:
  infile = open(argv[1], 'r')
  if len(argv) > 2:
    outfile = open(argv[2], 'w')
else:
  infile = stdin

# Read data
data = []
while 1:
  line = getline(infile)
  if line == None:
    break
  data.append(line.lstrip())

# Check consistency
for i in range(1, len(data)):
  if len(data[i]) != len(data[0]):
    stderr.write("Sequence " + `i + 1` + " has length different from sequence 1\n")
    exit(1)

# eliminate_uninformative(data): eliminate an uninformative site from data
def eliminate_uninformative(data):
  for i in range(0, len(data[0])):
    n0 = 0
    n1 = 0
    for j in data:
      if j[i] == '0':
        n0 += 1
      elif j[i] == '1':
        n1 += 1
    if n0 <= 1 or n1 <= 1:
      # This column is uninformative - eliminate
      for j in range(0, len(data)):
        data[j] = data[j][:i] + data[j][i + 1:]
      return i
  return None

# coalesce(data): coalesce two compatible sequences in data leading to a
# maximal decrease in ancestral data; if share is true the two sequences
# have to share ancestral material
def coalesce(data, share = 1):
  decrease = -1
  for i in range(0, len(data)):
    if data[i].count('0') + data[i].count('1') == 0:
      # This sequence carries no informative ancestral material
      del data[i]
      return (i, i)
    for j in range(i + 1, len(data)):
      shared = 0
      for k in range(0, len(data[i])):
        if data[i][k] in "01" and data[j][k] in "01":
          if data[i][k] == data[j][k]:
            shared += 1
          else:
            break
      else:
        # Sequences i and j are compatible - coalesce
        if shared > decrease:
          decrease = shared
          pair = (i, j)
  if decrease == -1 or share and decrease == 0:
    # No compatible sequences, or no compatible sequences sharing ancestral
    # information
    return None
  else:
    s = ""
    for k in range(0, len(data[0])):
      if data[pair[0]][k] in "01":
        s += data[pair[0]][k]
      else:
        s += data[pair[1]][k]
    data[pair[0]] = s
    del data[pair[1]]
    return pair

# largest_decrease(data): Find recombination event allowing the largest
# decrease in ancestral material
def largest_decrease(data):
  decrease = -1
  for i in range(0, len(data)):
    for j in range(i + 1, len(data)):
      potential = 0
      for k in range(0, len(data[i])):
        if data[i][k] in "01" and data[j][k] in "01":
          if data[i][k] != data[j][k]:
            if potential > decrease:
              decrease = potential
              pair = (i, j, k)
            break
          else:
            potential += 1
      potential = 0
      for k in range(len(data[i]) - 1, -1, -1):
        if data[i][k] in "01" and data[j][k] in "01":
          if data[i][k] != data[j][k]:
            if potential > decrease:
              decrease = potential
              pair = (j, i, k + 1)
            break
          else:
            potential += 1
  return pair

def recombine(data, event, sequence = 0):
  if event[0] < event[1]:
    # The recombination allows a prefix of the two sequences to coalesce
    for i in range(0, event[2]):
      if data[event[1 - sequence]][i] not in "01":
        data[event[1 - sequence]] = data[event[1 - sequence]][:i] + data[event[sequence]][i] + data[event[1 - sequence]][i + 1:]
    data[event[sequence]] = event[2] * "x" + data[event[sequence]][event[2]:]
  else:
    # The recombination allows a postfix of the two sequences to coalesce
    for i in range(event[2], len(data[0])):
      if data[event[sequence]][i] not in "01":
        data[event[sequence]] = data[event[sequence]][:i] + data[event[1 - sequence]][i] + data[event[sequence]][i + 1:]
    data[event[1 - sequence]] = data[event[1 - sequence]][:event[2]] + (len(data[0]) - event[2]) * "x"

# implode(data): Reduce data set by iteratively eliminating uninformative sites
# and coalescing compatible sequences.
def implode(data):
  if outfile:
    events = []
  change = 1
  while change and data:
    change = 0
    if outfile:
      while 1:
        event = eliminate_uninformative(data)
        if event != None:
          events.append(event)
          change = 1
        else:
          break
      while 1:
        event = coalesce(data)
        if event:
          events.append(event)
          change = 1
        else:
          break
    else:
      while eliminate_uninformative(data) != None:
        change = 1
      while coalesce(data) != None:
        change = 1
  if outfile:
    return events

# Keep applying evolutionary events to data, greedily choosing events leading
# to a maximum decrease in ancestral material, until no data remains
def greedy(data):
  events = implode(data)
  if data:
    data2 = deepcopy(data)
    r = largest_decrease(data)
    recombine(data, r, 0)  # Apply recombination to lowest numbered sequence
    recombine(data2, r, 1) # Apply recombination to highest numbered sequence
    if r[0] > r[1]:
      # Change to indicate postfix coalescence with a negative index
      r = (r[1], r[0], -r[2])
    if outfile:
      return map(add, (1, events), min(map(add, (0, [r]), greedy(data)), map(add, (0, [(r[1], r[0], r[2])]), greedy(data2))))
    else:
      return 1 + min(greedy(data), greedy(data2))
  else:
    if outfile:
      return (0, events)
    else:
      return 0

# Print a human readable listing of the events described for k sequences of
# length n to outfile
def parse_results(k, n, events, outfile = stdout):
  column = range(1, n + 1)
  sequence = range(1, k + 1)
  for event in events:
    if type(event) == IntType:
      # Mutation event
      outfile.write("Mutation in column " + `column[event]` + "\n")
      del column[event]
    elif len(event) == 2:
      # Coalescence event
      if event[0] == event[1]:
        # Ancestral material left in this sequence agrees with all other
        # sequences - we can coalesce it with any other sequence
        outfile.write("Sequence " + `sequence[event[0]]` + " contains no more ancestral material\n")
      else:
        outfile.write("Coalescence of sequences " + `sequence[event[0]]` + " and " + `sequence[event[1]]` + "\n")
      del sequence[event[1]]
    else:
      # Recombination event
      if event[2] < 0:
        k = -event[2]
        coalesce = " postfix "
      else:
        k = event[2]
        coalesce = " prefix "
      outfile.write("Recombination in sequence " + `sequence[event[0]]`)
      outfile.write(" after position " + `column[k - 1]`)
      outfile.write("\n followed by coalescence of" + coalesce + "with sequence " + `sequence[event[1]]` + "\n")

if outfile:
  n = len(data[0])
  k = len(data)
  result = greedy(data)
  print result[0]
  parse_results(k, n, result[1], outfile)
  outfile.close()
else:
  print greedy(data)
