/* greven.c: Implementation of a branch and bound algorithm to
 * enumerate ancestral states that can be encountered in histories
 * with a given maximum number of recombinations in an SNP data set.
 *
 * Rune Lyngsoe, January 2005
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "gene.h"
#include "enumerate.h"
#include "common.h"

void print_usage(FILE *f, char *name)
{
  int linelength = 80;

  fprintf(f, "Usage: %s [options] [input]\n", name);
  pretty_print(f, "The program reads data from the input file specified (or from standard input if no input file is specified) and counts the number of unique ancestral states that can be visited in histories with at most a specified number of recombinations. The default is to count the number of unique ancestral states that can be visited in histories with a minimum number of recombinations.", linelength, 0);
  fprintf(f, "Legal options are:\n");
  print_option(f, "-cN", "count states visitable in histories with at most N recombinations.", linelength, -1);
  print_option(f, "-rN", "count states visitable with at most N more recombinations than the minimum number of recombinations required for the data.", linelength, -1);
  print_option(f, "-k", "Assume that the common ancestral sequence is known, i.e. that we know which is the wild type and which is the mutant in each site. In binary format the all-0 sequence is assumed to be the common ancestral sequence, while the common ancestral sequence has to be specified directly for amino acid and nucleic sequences (see options -a and -n)", 80, -1);
  print_option(f, "-o", "Assume input data is in own format (default is to first try to parse data in own format, and if that fails to try to parse it in fasta format).", linelength, -1);
  print_option(f, "-f", "Assume input data is in fasta format.", linelength, -1);
  print_option(f, "-a", "Assume input consists of amino acid (protein) sequences using the one letter amino acid alphabet; anything not in the amino acid one letter alphabet is treated as an unresolved site (default is to assume sequences in binary, i.e. 0/1, format where anything but a 0 or a 1 is considered an unresolved site). If the most recent common ancestor is assumed known, the first sequence in the input data is considered to specify the wild type of each site and is not included as part of the sample set." , 80, -1);
  print_option(f, "-n", "Assume input consists of nucleic sequences; anything but a/c/g/t/u is considered an unresolved site (default is to assume binary, i.e. 0/1, format where anything but a 0 or a 1 is considered an unresolved site). If the most recent common ancestor is assumed known, the first sequence in the input data is considered to specify the wild type of each site and is not included as part of the sample set.", 80, -1);
  print_option(f, "-l[name]", "Output reachable ancestral states to file name (stdout if no file is specified).", linelength, -1);
  print_option(f, "-h, -H", "print this information and stop.", linelength, -1);
}

/* Print an ancestral state stored in a hash table */
static FILE *state_file = NULL;
static void print_state(PackedGenes *p, void *v, va_list args)
{
  Genes *g = unpack_genes(p);
  output_genes(g, state_file, NULL);
  fprintf(state_file, "--\n");
  free_genes(g);
}

int main(int argc, char **argv)
{
  Genes *g;
  AnnotatedGenes *a;
  HashTable *states;
  /* Default behaviour is to enumerate minACs */
  int i, relative = 1, silent = 0;
  unsigned int bound = 0;
  Gene_Format format = GENE_ANY;
  Gene_SeqType seqtype = GENE_BINARY;

  /* Initialise random number generator */
  initialise_xrandom();

  /* Analyse command line options */
  while ((i = getopt(argc, argv, "c:r:kofanhHl::s")) >= 0){
    switch(i){
    case 'h':
    case 'H':
      print_usage(stdout, argv[0]);
      exit(0);
    case '?':
    case ':':
      print_usage(stderr, argv[0]);
      exit(1);
    case 'c':
      /* Enumerate with an absolute number of recombinations */
      if (sscanf(optarg, "%u", &bound) != 1){
	fprintf(stderr, "Could not parse %s as absolute number of recombinations allowed\n", optarg);
	print_usage(stderr, argv[0]);
	exit(1);
      }
      relative = 0;
      break;
    case 'r':
      /* Enumerate with a relative number of recombinations */ 
      if (sscanf(optarg, "%u", &bound) != 1){
	fprintf(stderr, "Could not parse %s as relative number of recombinations allowed\n", optarg);
	print_usage(stderr, argv[0]);
	exit(1);
      }
      relative = 1;
      break;
    case 'k':
      gene_knownancestor = 1;
      break;
    case 'o':
      format = GENE_BEAGLE;
      break;
    case 'f':
      format = GENE_FASTA;
      break;
    case 'a':
      seqtype = GENE_AMINO;
      break;
    case 'n':
      seqtype = GENE_NUCLEIC;
      break;
    case 'l':
      /* Output reachable ancestral states */
      /* Check whether we already opened an output file that now needs
       * to be closed.
       */
      if ((state_file != NULL) && (state_file != stdout))
	fclose(state_file);
      /* Was a file name specified? */
      if (optarg != 0){
	if ((state_file = fopen(optarg, "w")) == NULL){
	  fprintf(stderr, "Could not open file %s for output\n", optarg);
	  print_usage(stderr, argv[0]);
	  exit(1);
	}
      }
      else
	state_file = stdout;
      break;
    case 's':
      silent = 1;
      break;
    }
  }

#ifdef ENABLE_VERBOSE
  set_verbose(1);
#endif

  /* Read data */
  if (argc > optind){
    if ((a = read_genes(argv[optind], format, seqtype)) == NULL){
      fprintf(stderr, "Could not parse %s as valid data\n", argv[optind]);
      exit(1);
    }
  }
  else{
    if ((a = read_genes(NULL, format, seqtype)) == NULL){
      fprintf(stderr, "Could not parse input as valid data\n");
      exit(1);
    }
  }
  if ((gene_knownancestor) && (seqtype != GENE_BINARY))
    /* First sequnce only included to specify known common ancestor */
    remove_annotatedgene(a, 0);
  g = a->g;

  if (!silent)
    output_annotatedgenes(a, NULL, NULL);

  if (state_file != NULL){
    if (relative)
      states = enumerate_relative(g, bound);
    else
      states = enumerate_absolute(g, bound);
    hashtable_map(states, (void (*)(void *, void *, va_list))print_state);

    i = hashtable_size(states);
    hashtable_destroy(states, (void (*)(void *))free_packedgenes, NULL,
		      (void (*)(void *))free);
  }
  else{
    if (relative)
      i = count_relative(g, bound);
    else
      i = count_absolute(g, bound);
  }

  if (!silent){
    if (relative){
      if (bound == 0)
	printf("Ancestral states visitable with minimum number of recombinations is %d\n", i);
      else
	printf("Ancestral states visitable with at most %d recombination%s more than the minimum\nis %d\n", bound, (bound != 1 ? "s" : ""), i);
    }
    else{
      printf("Ancestral states visitable with at most %d recombination%s is %d\n", bound, (bound != 1 ? "s" : ""), i);
    }
  }

  free_annotatedgenes(a);

  return 0;
}
