#! /usr/local/bin/python

from getopt import getopt
from sys import argv, stdin, stderr, exit, maxint
from popen2 import popen2 as popen
from copy import deepcopy
from string import digits
from operator import and_ as And

# Parse options
options, arguments = getopt(argv[1:], "mka:r:")
ms_format = False
known_mrca = False
limit = 0
relative = True
def parse_positiveint(s):
  i = int(s)
  if i < 0:
    raise ValueError, "Negative value encountered"
  return i
    
for o in options:
  if o[0][1] == 'm':
    ms_format = True
  elif o[0][1] == 'k':
    known_mrca = True
  elif o[0][1] == 'a':
    relative = False
    limit = parse_positiveint(o[1])
  elif o[0][1] == 'r':
    relative = True
    limit = parse_positiveint(o[1])

# Read data
if arguments != []:
  infile = open(arguments[0], 'r')
else:
  infile = stdin
data = []
if ms_format:
  while True:
    s = infile.readline()
    if s[:2] == "//":
      for i in range(3):
        s = infile.readline()
      while s != "" and s[:2] != "//":
        s = s.strip()
        if s != "":
          data.append(s)
          if len(data[-1]) != len(data[0]):
            stderr.write("Data lines are not of the same length\n")
            exit(1)
        s = infile.readline()
    elif s != "":
      continue
    break
else:
  # getline(f): returns next non-empty, non-comment line from f, None if EOF
  def getline(f):
    while 1:
      s = f.readline()
      if s == "":
        return None
      # Strip off trailing comment
      escaped = 0
      for i in range(0, len(s)):
        if s[i] == '#' and not escaped:
          s = s[:i]
          break
        if s[i] == '\\':
          escaped = not escaped
      # Strip off trailing whitespace
      s = s.rstrip()
      if s != "":
        break
    return s

  while 1:
    line = getline(infile)
    if line == None:
      break
    data.append(line)
    if len(data[-1]) != len(data[0]):
      stderr.write("Data lines are not of the same length\n")
      exit(1)
  if arguments != []:
    infile.close()

# Remove uninformative sites
def remove_uninformative(data, known_mrca = False):
  change = False
  i = 0
  while i < len(data[0]):
    c = [0, 0]
    for j in range(len(data)):
      if data[j][i] != 'x':
        c[int(data[j][i])] += 1
    if c[1] < 2 or (not known_mrca and c[0] < 2):
      # This site is not informative
      change = True
      for j in range(len(data)):
        data[j] = data[j][:i] + data[j][i + 1:]
    else:
      i += 1
  return change

# Coalesce identical sequences
def coalesce(data):
  data.sort()
  change = False
  i = 1
  while i < len(data):
    if data[i - 1] == data[i]:
      change = True
      del data[i]
    else:
      i += 1
  return change

# Remove stretches of identical columns
def remove_siamese(data):
  change = False
  i = 1
  while i < len(data[0]):
    if reduce(And, map(lambda x: x[i - 1] == x[i], data), True):
      # Sites i and i - 1 are siamese twins
      j = i + 1
      while j < len(data[0]) and reduce(And, map(lambda x: x[i] == x[j], data), True):
        # And you can add site j to the bunch
        j += 1
      # Remove all but one siamese twins
      for k in range(len(data)):
        data[k] = data[k][:i] + data[k][j:]
      change = True
    i += 1
  return change

# Iteratively remove uninformative sites and coalesce identical sequences
# until this is no longer possible
def implode(data, known_mrca = False, uninformative = True, coalescence = True, siamese = True):
  change = True
  while change:
    change = False
    if uninformative:
      change = remove_uninformative(data, known_mrca) or change
    if coalescence:
      change = coalesce(data) or change
    if siamese:
      change = remove_siamese(data) or change
  return data

# Determine whether sequences s and t are compatible
def compatible(s, t):
  for i in range(len(s)):
    if s[i] != t[i] and 'x' not in [s[i], t[i]]:
      return False
  return True

# Return result of coalescing s and t
def coalesce_pair(s, t):
  u = ""
  for i in range(len(s)):
    if s[i] == 'x':
      u += t[i]
    else:
      u += s[i]
  return u

# Determine the range flanked by the leftmost and rightmost resolved sites in s
def resolved_range(s):
  i = 0
  j = len(s) - 1
  while i <= j and s[i] == 'x':
    i += 1
  while i <= j and s[j] == 'x':
    j -= 1
  return range(i + 1, j + 1)

# Recursion used by find_minimum; determine whether we can find a history for
# state using at most limit recombinations with known_states as prior
# information
def _find_minimum(state, known_states, limit, known_mrca = False):
  implode(state, known_mrca)
  if len(state) <= 1 or len(state[0]) <= 1:
    # Found the most recent common ancestor
    return True
  state.sort()
  key = tuple(state)
  if known_states.has_key(key):
    # Been here, done that
    known = known_states[key]
    if known >= limit:
      return False
    elif known < 0:
      if limit >= -known - 1:
        return True
      else:
        return False
  known_states[tuple(state)] = limit

  # Try all coalesces of compatible sequences
  for i in range(len(state)):
    for j in range(i):
      if compatible(state[i], state[j]):
        new_state = deepcopy(state)
        new_state.append(coalesce_pair(state[i], state[j]))
        del new_state[i]
        del new_state[j]
        if _find_minimum(new_state, known_states, limit, known_mrca):
          # Success
          known_states[key] = -limit - 1
          return True

  # Try all recombinations
  if limit > 0:
    for i in range(len(state)):
      for j in resolved_range(state[i]):
        if state[i][j] != 'x':
          # Only consider one recombination falling in a stretch of unresolved
          # sites
          new_state = deepcopy(state)
          new_state.append(j * 'x' + new_state[i][j:])
          new_state[i] = new_state[i][:j] + (len(state[i]) - j) * 'x'
          if _find_minimum(new_state, known_states, limit - 1, known_mrca):
            # Success
            known_states[key] = -limit - 1
            return True

  # No solution found
  return False

# Determine minimum number of recombinations required for state, using
# known_states as prior information
def find_minimum(state, known_states, known_mrca = False):
  limit = 0
  state = deepcopy(state)
  while True:
    if _find_minimum(state, known_states, limit, known_mrca):
      return limit
    limit += 1

level = 0
# Search one event back from state for a history leading to a most recent
# common ancestor with at most limit recombinations, and with known_states
# as prior information on states
def handle_state(state, known_states, limit, imploded_states, known_mrca = False):
  global level
  level += 1
  # Start by sorting sequences
  state.sort()
  key = tuple(state)

  if known_states.has_key(key) and known_states[key] >= limit:
    # Been here, done that
    return

  if find_minimum(state, imploded_states, known_mrca) > limit:
    # No evolutionary history possible
    return

  known_states[key] = limit
  # Try all possible mutations
  for i in range(len(state[0])):
    count = [-1, -1]
    for j in range(len(state)):
      if state[j][i] == 'x':
        continue
      if count[int(state[j][i])] == -1:
        count[int(state[j][i])] = j
      elif count[int(state[j][i])] >= 0:
        count[int(state[j][i])] = -2
    if count[1] >= 0 and (count[0] != -1 or known_mrca):
      # We can mutate the 1 in site i in sequence count[1]
      new_state = deepcopy(state)
      new_state[count[1]] = new_state[count[1]][:i] + "0" + new_state[count[1]][i + 1:]
      handle_state(new_state, known_states, limit, imploded_states, known_mrca)
    if count[0] >= 0 and not known_mrca and count[1] != -1:
      # We can mutate the 0 in site i in sequence count[0]
      new_state = deepcopy(state)
      new_state[count[0]] = new_state[count[0]][:i] + "1" + new_state[count[0]][i + 1:]
      handle_state(new_state, known_states, limit, imploded_states, known_mrca)

  # Try all possible coalesces
  for i in range(len(state)):
    for j in range(i):
      if compatible(state[i], state[j]):
        new_state = deepcopy(state)
        new_state.append(coalesce_pair(new_state[i], new_state[j]))
        del new_state[i]
        del new_state[j]
        handle_state(new_state, known_states, limit, imploded_states, known_mrca)
  # Try all possible recombinations
  if limit > 0:
    for i in range(len(state)):
      for j in resolved_range(state[i]):
        if state[i][j] != 'x':
          # Only consider one recombination falling in a stretch of unresolved
          # sites
          new_state = deepcopy(state)
          new_state.append(j * 'x' + new_state[i][j:])
          new_state[i] = new_state[i][:j] + (len(state[i]) - j) * 'x'
          handle_state(new_state, known_states, limit - 1, imploded_states, known_mrca)
  level -= 1

reachable = {}
visited = {}
if relative:
  limit += find_minimum(data, visited, known_mrca)
handle_state(data, reachable, limit, visited, known_mrca)
for i in reachable:
  for j in i:
    print j
  print "--"
print "Number of reachable states is", len(reachable)
