/* lower_bounds.c: Wrapper for calling haplotype bound function
 *
 * Rune Lyngsoe, December 2004
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#ifdef HAPLOTYPE_BLOCKS
#include <math.h>
#endif

#include "gene.h"
#include "bounds.h"
#include "common.h"

void print_usage(FILE *f, char *name)
{
  fprintf(f, "Usage: %s [options] [input]\n", name);
  pretty_print(f, "The program reads data from the input file specified (or standard input if no input is specified) and computes the Hudson-Kaplan and haplotype bounds on the minimum number of recombinations required to explain the data set under the infinite sites assumption. A (hopefully faster) heuristic computation can be invoked with the -s, -w, and -i options. Any heuristic restriction not specified, or specified with a non-positive value, is not applied.", 80, 0);
  pretty_print(f, "Legal options are:", 80, 0);
#ifdef HAPLOTYPE_BLOCKS
  print_option(f, "-m[name]", "Output matrix of local bounds to file name (stdout if no file is specified); the Hudson-Kaplan bounds will be in above the diagonal, while the haplotype bounds will be below the diagonal", 80, -1);
#endif
  print_option(f, "-k", "Assume that the common ancestral sequence is known, i.e. that we know which is the wild type and which is the mutant in each site. In binary format the all-0 sequence is assumed to be the common ancestral sequence, while the common ancestral sequence has to be specified directly for amino acid and nucleic sequences (see options -a and -n)", 80, -1);
  print_option(f, "-o", "Assume input data is in own format (default is to first try to parse data in own format, and if that fails to try to parse it in fasta format).", 80, -1);
  print_option(f, "-f", "Assume input data is in fasta format.", 80, -1);
  print_option(f, "-a", "Assume input consists of amino acid (protein) sequences using the one letter amino acid alphabet; anything not in the amino acid one letter alphabet is treated as an unresolved site (default is to assume sequences in binary, i.e. 0/1, format where anything but a 0 or a 1 is considered an unresolved site). If the most recent common ancestor is assumed known, the first sequence in the input data is considered to specify the wild type of each site and is not included as part of the sample set." , 80, -1);
  print_option(f, "-n", "Assume input consists of nucleic sequences; anything but a/c/g/t/u is considered an unresolved site (default is to assume binary, i.e. 0/1, format where anything but a 0 or a 1 is considered an unresolved site). If the most recent common ancestor is assumed known, the first sequence in the input data is considered to specify the wild type of each site and is not included as part of the sample set.", 80, -1);
  print_option(f, "-sN", "Only consider subsets of at most N different types of sites", 80, -1);
  print_option(f, "-wN", "Only consider two sites in conjunction if they are contained in a region of no more than N sites in the reduced data set, i.e. where all non-informative sites, neighbouring identical columns, and identical have been iteratively removed", 80, -1);
  print_option(f, "-iN", "Only continue with a subset of site types if the last type added increased the number of composite types by at least N", 80, -1);
  print_option(f, "-h, -H", "Print this information and stop", 80, -1);
}

int main(int argc, char **argv)
{
  Genes *g;
  AnnotatedGenes *a;
  Sites *s;
  int i;
  Gene_Format geneformat = GENE_ANY;
  Gene_SeqType seqtype = GENE_BINARY;
#ifdef HAPLOTYPE_BLOCKS
  int j, k, **HK, **B, n, entry_width;
  FILE *matrix = NULL;
  SuperColumn *c;
  char *format;
#endif
  int set = INT_MAX, width = INT_MAX, increase = 0, heuristic = 0;

#ifdef ENABLE_VERBOSE
  int v = 0, j;
#endif

  /* Initialise random number generator */
  initialise_xrandom();

  /* Analyse command line options */
#define MYERS_OPTIONS1 "hHkofans:w:i:"
#ifdef HAPLOTYPE_BLOCKS
#define MYERS_OPTIONS MYERS_OPTIONS1 "m::"
#else
#define MYERS_OPTIONS MYERS_OPTIONS1
#endif

  /* Parse command line options */
  while ((i = getopt(argc, argv, MYERS_OPTIONS)) >= 0)
    switch(i){
    case 'h':
    case 'H':
      print_usage(stdout, argv[0]);
      exit(0);
    case '?':
    case ':':
      print_usage(stderr, argv[0]);
      exit(1);
    case 'k':
      gene_knownancestor = 1;
      break;
    case 'o':
      geneformat = GENE_BEAGLE;
      break;
    case 'f':
      geneformat = GENE_FASTA;
      break;
    case 'a':
      seqtype = GENE_AMINO;
      break;
    case 'n':
      seqtype = GENE_NUCLEIC;
      break;
#ifdef HAPLOTYPE_BLOCKS
    case 'm':
      /* Compute matrix of local minimum number of recombinations required */
      /* Check whether we already opened an output file that now needs
       * to be closed.
       */
      if ((matrix != NULL) && (matrix != stdout))
	fclose(matrix);
      /* Was a file name specified? */
      if (optarg != 0){
	if ((matrix = fopen(optarg, "w")) == NULL){
	  fprintf(stderr, "Could not open file %s for output\n", optarg);
	  print_usage(stderr, argv[0]);
	  exit(1);
	}
      }
      else
	matrix = stdout;
      break;
#endif
    case 's':
      if ((set = atoi(optarg)) <= 0)
	set = INT_MAX;
      else
	heuristic = 1;
      break;
    case 'w':
      if ((width = atoi(optarg)) <= 0)
	width = INT_MAX;
      else
	heuristic = 1;
      break;
    case 'i':
      if ((increase = atoi(optarg)) <= 0)
	increase = 0;
      else
	heuristic = 1;
      break;
    }

  /* Read data */
  if (argc > optind){
    if ((a = read_genes(argv[optind], geneformat, seqtype)) == NULL){
      fprintf(stderr, "Could not parse %s as valid data\n", argv[optind]);
      exit(1);
    }
  }
  else{
    if ((a = read_genes(NULL, geneformat, seqtype)) == NULL){
      fprintf(stderr, "Could not parse input as valid data\n");
      exit(1);
    }
  }
  if ((gene_knownancestor) && (seqtype != GENE_BINARY))
    /* First sequnce only included to specify known common ancestor */
    remove_annotatedgene(a, 0);
  g = a->g;

#ifdef HAPLOTYPE_BLOCKS
  if (matrix != NULL){
    n = g->length;
    representativeness = MakeLList();
    for (i = 0; i < g->length; i++){
      c = (SuperColumn *)xmalloc(sizeof(SuperColumn));
      c->left = c->right = i;
      Enqueue(representativeness, (void *)c);
    }
    representativeness_counter = MakeCounter(representativeness, FIRST);
  }
#endif

  implode_genes(g);
#ifdef ENABLE_VERBOSE
  if (v){
    printf("Reduced data set:\n");
    output_genes_indexed(g, NULL);
  }
#endif
  if ((g->length > 0) && (g->n > 0)){
#ifdef HAPLOTYPE_BLOCKS
    if (matrix == NULL){
#endif
      printf("Hudson-Kaplan bound: %d\n", hudson_kaplan_genes(g));
      printf("Haplotype bound: %d\n",
	     (heuristic ? haplotype_heuristic_genes(g, set, width, increase)
	      : haplotype_bound_genes(g)));
#ifdef HAPLOTYPE_BLOCKS
    }
    else{
      /* Compute matrices of bounds for reduced data set */
      s = genes2sites(g);
      HK = hudson_kaplan_local(s);
      local2locals(s->length - 1, HK);
      printf("Hudson-Kaplan bound: %d\n", HK[0][s->length - 2]);
      if (heuristic)
	B = haplotype_heuristic_local(s, set, width, increase);
      else
	B = haplotype_bound_local(s);
      local2locals(s->length - 1, B);
      printf("Haplotype bound: %d\n", B[0][s->length - 2]);

      /* Convert to matrices of bounds for original data set */
      /* Start by creating space for the larger matrices */
      HK = (int **)xrealloc(HK, (n - 1) * sizeof(int *));
      B = (int **)xrealloc(B, (n - 1) * sizeof(int *));
      for (i = 0; i < g->length - 1; i++){
	HK[i] = (int *)xrealloc(HK[i], (n - i - 1) * sizeof(int));
	B[i] = (int *)xrealloc(B[i], (n - i - 1) * sizeof(int));
      }
      for (; i < n - 1; i++){
	HK[i] = (int *)xmalloc((n - i - 1) * sizeof(int));
	B[i] = (int *)xmalloc((n - i - 1) * sizeof(int));
      }
      /* Now do the conversion */
      explode_local(HK, representativeness, n);
      explode_local(B, representativeness, n);

      /* Output matrices to file matrix */
      entry_width = (B[0][n - 2] > 0 ? (int)floor(log10(B[0][n - 2])) + 1 : 1);
      format = (char *)xmalloc(5 + (int)floor(log10(entry_width)));
      sprintf(format, "%%%dd ", (int)entry_width);
      for (i = 0; i < n; i++){
	for (j = 0; j < n; j++)
	  if (j < i)
	    /* Print bound between sites j and i from B */
	    fprintf(matrix, format, B[j][i - j - 1]);
	  else if (j > i)
	    /* Print bound between sites i and j from HK */
	    fprintf(matrix, format, HK[i][j - i - 1]);
	  else{
	    /* Print centered dash */
	    for (k = 0; k < entry_width / 2; k++)
	      fprintf(matrix, " ");
	    fprintf(matrix, "-");
	    for (k = 0; k < (entry_width - 1)/ 2 + 1; k++)
	      fprintf(matrix, " ");
	  }
	fprintf(matrix, "\n");
      }
      /* Clean up */
      for (i = 0; i < n - 1; i++){
	free(B[i]);
	free(HK[i]);
      }
      free(B);
      free(HK);
      free_sites(s);
      free(format);
    }
#endif
  }
  else{
    printf("Hudson-Kaplan bound: 0\n");
    printf("Haplotype bound: 0\n");
#ifdef HAPLOTYPE_BLOCKS
    if (matrix != NULL){
      for (i = 0; i < n; i++){
	for (j = 0; j < n; j++)
	  fprintf(matrix, (i != j ? "0 " : "- "));
	fprintf(matrix, "\n");
      }
    }
#endif
  }

  /* Clean up */
  free_annotatedgenes(a);
#ifdef HAPLOTYPE_BLOCKS
  if (matrix != NULL){
    while (Length(representativeness) > 0)
      free(Pop(representativeness));
    DestroyLList(representativeness);
    DestroyCounter(representativeness_counter);
  }
#endif

  return 0;
}
