/*******************************************************************

    lpinterface.c
  
    Implementation of functions for interfacing with integer and
    linear programming library

    Rune Lyngsø (lyngsoe@stats.ox.ac.uk), December 2006

********************************************************************/

#include <limits.h>
#include "glpk.h"

#include"common.h"

/* Set up problem */
static LPX *setup_problem(int m, int n, int *c, int **A, int (*l)[2])
{
  LPX *lp;
  int i, j, k, *ai, *aj;
  double *ar;

  /* Construct problem */
  lpx_create_prob();
  lpx_set_obj_dir(lp, LPX_MAX);
  /* Add rows */
  lpx_add_rows(lp, m);
  for (i = 0; i < m; i++)
    if (l[i][0] == INT_MIN)
      lp_set_row_bnds(lp, i + 1, LPX_UP, l[i][0], l[i][1]);
    else
      lp_set_row_bnds(lp, i + 1, LPX_DB, l[i][0], l[i][1]);
  /* Add columns */
  lpx_add_cols(lp, n);
  for (i = 0; i < n; i++){
    lp_set_col_bnds(lp, i + 1, LPX_DB, l[m + i][0], l[m + i][1]);
    lp_set_onj_coef(lp, i + 1, c[i]);
  }
  /* Set coefficients of equation system */
  ai = (int *)xmalloc(m * n * sizeof(int));
  aj = (int *)xmalloc(m * n * sizeof(int));
  ar = (double *)xmalloc(m * n * sizeof(double));
  k = 0;
  for (i = 0; i < m; i++)
    for (j = 0; j < n; j++)
      if (A[i][j] != 0){
	k++;
	ai[k] = i;
	aj[k] = j;
	ar[k] = A[i][j];
      }
  lpx_load_matrix(lp, k, ai, aj, ar);

  /* Clean up */
  free(ai);
  free(aj);
  free(ar);

  return lp;
}

/* Solve linear program maximise c·x subject to l[0] <= A·x <= l[1] */
double *lpinterface_solvelp(int m, int n, int *c, int **A, int (*l)[2])
{
  int i;
  double *x;
  LPX *lp = setup_problem(m, n, c, A, l);

  /* Solve problem */
  lpx_set_int_parm(lp, LPX_K_MSGLEV, 1);
  lpx_set_int_parm(lp, LPX_K_ROUND, 1);
  lpx_set_int_parm(lp, LPX_K_PRESOL, 1);
  if (lpx_simplex(lp) != LPX_E_OK)
    /* Problem went unsolved */
    return NULL;

  /* Construct solution vector */
  x = (double *)xmalloc(n * sizeof(double));
  for (i = 0; i < n; i++)
    x[i] = lpx_get_col_prim(lp, i + 1);

  /* Clean up */
  lpx_delete_prob(lp);

  return x;
}

/* Solve integer program maximise c·x subject to l[0] <= A·x <= l[1] */
int *lpinterface_solveip(int m, int n, int *c, int **A, int (*l)[2])
{
  int i;
  int *x;
  LPX *lp = setup_problem(m, n, c, A, l);

  /* Convert to integer programming problem */
  lpx_set_class(lp, LPX_MIP);
  for (i = 0; i < n; i++)
    lpx_set_col_kind(lp, i + 1, LPX_IV);

  /* Solve problem */
  lpx_set_int_parm(lp, LPX_K_MSGLEV, 1);
  lpx_set_int_parm(lp, LPX_K_ROUND, 1);
  if (lpx_integer(lp) != LPX_E_OK)
    /* Problem went unsolved */
    return NULL;

  /* Construct solution vector */
  x = (int *)xmalloc(n * sizeof(int));
  for (i = 0; i < n; i++)
    x[i] = (int)nearbyint(lpx_get_col_prim(lp, i + 1));

  /* Clean up */
  lpx_delete_prob(lp);

  return x;
}
