/***************************************************************************

    testexpression.c

    Small utility program for testing score function definitions by applying
    the specified expression (the kwarg default if no expression is specified)
    to the set of sequences supplied.

    Rune Lyngsø (lyngsoe@stats.ox.ac.uk), July 2007

****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <limits.h>

#include "kwarg.h"
#include "gene.h"
#include "bounds.h"
#include "exact.h"
#include "common.h"
#include "backtrack.h"
#include "expression.h"

static void _print_usage(FILE *f, char *name)
{
  fprintf(f, "Usage: %s [options] [input]\n", name);
  pretty_print(f, "The program reads data from the input file specified (or from standard input if no input is specified) and applies the scoring function specified by the -E or -F option (see these) to it, or the kwarg default scoring function if no scoring function is specified. This allows experimentation to explore time and quality of different strategies.", 80, 0);
  fprintf(f, "Legal options are:\n");
  print_option(f, "-Eexp", "Score configurations according to expression exp. An expression should consist of zero or more definitions followed by a single statement, all separated by semicolons. The score is the value of the final statement, while the definitions allow variables and functions to be defined and used. A variable is defined using the syntax `name = statement' where name is the name of the variable that is set to the value of statement. A function is defined using the syntax `f(x, y, z) = statement' where f is the function name, x, y, and z the parameters of the function (at least one parameter is required, but theere is no upper bound on the number of parameters), and statement the body of the function. Variable, function and parameter names are required to be a sequence of letters and digits, beginning with a letter. Statements can be\n- real or integer numbers\n- variables\n- function calls, with syntax `f(statement1, statement2, statement3)' where the values of arguments statement1, statement2, and statement3 are determined when the function is called; the number of arguments in a function call is required to match the number of parameters in the function definition\n- arithmetic operations: +, -, *, /, and ^ (for exponentiation - this operation is also available using the synonym **) that require numerical arguments and yield numerical results\n- comparisons: <, >, = (with synonym ==), <= (with synonym =<), >= (with synonym =>), and != (with synonyms <> and ><) that require numerical arguments and yield Boolean results\n- Boolean operations: & (and, with synonym &&), | (or, with synonym ||), ! (not, with synonym ¬), and \\ (exclusive or, with synonym \\\\) that require Boolean arguments and yield Boolean values\n- conditionals with the syntax `test ? consequence : alternative' where test, consequence and alternative are statements; the value of test, which is required to be a Boolean, is first determined, and only the relevant branch - consequence if test is True, alternative if test is False - is then evaluated to determine the value of the conditional. White space - space, tabs and newlines - is ignored.\nThe statement defining a variable cannot refer to the variable itself, only previously defined variables and functions. Conversely, a function is allowed to be recursive, so e.g. factorials can be computed as `fac(n) = n < 2 ? 1 : n * fac(n - 1)'. Reusing a variable name does not change the value of the original definition but introduces a new variable with the same name and the new value. So the expression `a = 1; f(x) = a; a = 2; f(0)' will evaluate to 1. Observe that we need to use a dummy parameter x for f as functions are required to have at least one parameter. Arguments to functions are allowed to have any type as long as the corresponding parameter is used in concordance with the argument type. Functions can even take other functions as arguments, e.g. `fixpoint(f, a) = (f(a) = a ? a : fixpoint(f, f(a)))' finds a fixpoint for f by reapplying f starting from a until the value no longer changes.\nSome built-in functions and variables are available for convenience and to allow probing of the configuration being scored. The names of these can of course be be reused to define new variables and functions. This will hide the built-in functions, making them inaccessible for the remainder of the expression. The convenience functions are log (base 10 logarithm) and ln (natural logarithm) and the trigonometric functions sin, cos, tan, asin, acos, and atan. Convenience constants are e and pi. Constants for probing the configuration are\n- r: number of recombinations in step leading to configuration, which is always 0, 1, or 2 - this constant also has synonym recombinations\n- am: the total amount of ancestral material in the sequences of the configuration, with synonym ancestralmaterial\n- maxam: for maximum value of am over all configurations considered at this point, with synonym maximumancestralmaterial\n- minam: minimum value of am, with synonym minimumancestralmaterial\n- seq: number of sequences in configuration, with synonym numberofsequences\n- maxseq: maximum value of seq over all possible configurations considered at this point, with synonym maximumnumberofsequences\n- minseq: minimum value of seq, with synonym minimumnumberofsequences\n- len: number of sites remaining in configuration - uninformative sites are always removed, so this may be smaller than the original number of sites; it has synonym numberofsites\n- maxlen: maximum value of len over all possible configurations considered at this point, with synonym maximumnumberofsites\n- minlen:  minimum value of len, with synonym minimumnumberofsites\n- rmin: minimum number of recombinations required for configuration - note that for all but small data sets this will be extremely slow to compute; it has synonyms Rmin, beagle and Beagle\n- hk: Hudson-Kaplan bound on number of recombinations required for configuration, with synonyms HK, hudsonkaplan and HudsonKaplan\nFunctions for probing configurations are\n- hb: haplotype bound on number of recombinations required for configuration, with synonyms HB, haplotypebound and HaplotypeBound. It takes three arguments specifying the maximum size of subsets of sites considered, the maximum length of an interval the sites in a subset can be spread over, and the minimal increment of number of sequence types required for a subset of sites to be further expanded, in this order; for large data sets haplotype bound computation can be time consuming, and it is suggested to experiment with the haplotypebound program to find suitable values for the arguments\n- eagl: lower bounds based on Rmin values for subintervals combined with Hudson-Kaplan recombination requirements between pairs of sites using the composite method, with synonym Eagl. It takes one argument specifying the maximum length of intervals for which Rmin is computed; again, this can be time consuming for large data sets.\nBuilt-in constants and function calls that are time consuming to compute are only computed when needed and then stored for future reference.\nNumerical values are represented using the C double data type. Expressions where intermediate values are very large or very small should be avoided, as these may incur significant rounding errors. For example defining the combinatorial choose function as `choose(m, n) = fac(m) / (fac(n) * fac(m - n))' using the factorial function fac defined above will not be as robust as defining it as `choose(m, n) = n = 0 ? 1 : (m / n) * choose(m - 1, n - 1)' or even better as `min(a, b) = a < b ? a : b; choose2(m, n) = n = 0 ? 1 : (m / n) * choose2(m - 1, n - 1); choose(m, n) = choose2(m, min(n, m - n))'. One should also beware of the approximate nature of computers' real numbers. The data type (C doubles) used for numbers can behave unexpectedly, as witnessed by the value of `7 * 2.15 = 15.05 ? 1 : 0' on at least some machines (where 7 * 2.15 does not equal 15.05), and if possible tests for equality or inequality should be replaced by less than or greater than comparisons.\nThe expression used as default scoring scheme is `" KWARG_DEFAULTSCORE "'. It computes a proxy for the number of recombinations required for a configuration and among the configurations minimising the proxy chooses one with a minimum amount of ancestral material left. The proxy is computed as twice the Hudson-Kaplan bound as long as the configuration contains many sites, for shorter configurations the proxy is computed as one and a half times the maximum of eagl and haplotypebound heuristics, for smaller configurations still as one and a quarter times the maximum of more comprehensive versions of eagl and haplotypebound heuristics, and for very small configurations as Rmin itself.", 80, -1);
  print_option(f, "-Fname", "Read expression to score configurations by from file name. The format of an expression is as described for the -E option.", 80, -1);
  print_option(f, "-k", "Assume that the common ancestral sequence is known, i.e. that we know which is the wild type and which is the mutant in each site. In binary format the all-0 sequence is assumed to be the common ancestral sequence, while the common ancestral sequence has to be specified directly for amino acid and nucleic sequences (see options -a and -n)", 80, -1);
  print_option(f, "-o", "Assume input data is in own format. Default is to first try to parse data in own format, and if that fails to try to parse it in fasta format. Specifying this option, no attempt will be made to try to parse the data in fasta format.", 80, -1);
  print_option(f, "-f", "Assume input data is in fasta format. No attempt will be made to try to parse the data in own format. Note that the -o and the -f options override each other, so only the last one occurring in the command line will have an effect.", 80, -1);
  print_option(f, "-a", "Assume input consists of amino acid (protein) sequences using the one letter amino acid alphabet; anything not in the amino acid one letter alphabet is treated as an unresolved site (default is to assume sequences in binary, i.e. 0/1, format where anything but a 0 or a 1 is considered an unresolved site). If the most recent common ancestor is assumed known, the first sequence in the input data is considered to specify the wild type of each site and is not included as part of the sample set.", 80, -1);
  print_option(f, "-n", "Assume input consists of nucleic sequences; anything but a/c/g/t/u is considered an unresolved site (default is to assume binary, i.e. 0/1, format where anything but a 0 or a 1 is considered an unresolved site). If the most recent common ancestor is assumed known, the first sequence in the input data is considered to specify the wild type of each site and is not included as part of the sample set.", 80, -1);
  print_option(f, "-r", "Reduce data set by employing all safe events (coalescing identical sequences, merging neighbouring identical columns and removing uninformative sites) before computing score. In the kwarg recursions, ancestral configurations are always reduced this way.", 80, -1);
  print_option(f, "-h, -H", "Print this information and stop.", 80, -1);
}

/* Read entire content of file name into a string */
static char *_read_file(char *name)
{
  FILE *f = stdin;
  char *s = (char *)xmalloc(8 * sizeof(char));
  int size = 0, capacity = 8;

  /* Open file */
  if ((name != NULL) && ((f = fopen(name, "r")) == NULL))
    return NULL;

  /* Read file character by character */
  while ((s[size++] = fgetc(f)) != EOF)
    if (size == capacity - 1){
      /* Ran out of buffer capacity, double its size */
      capacity *= 2;
      s = (char *)xrealloc(s, capacity * sizeof(char));
    }

  /* Zero-terminate string, shrink buffer to fit it, and close file */
  s[size - 1] = '\0';
  s = xrealloc(s, size);
  if (f != stdin)
    fclose(f);

  return s;
}

/* Functions for interfacing with lower bound computations */
static int _score_rmin, _score_hk;
static HashTable *_score_functioncalls = NULL;
static Genes *_score_state;
/* Initialise parameters for hashing function call information into m bins */
static int *_score_initparam(unsigned long m)
{
  int i, *p = (int *)xmalloc(5 * sizeof(int));

  p[0] = m;
  for (i = 1; i < 5; i++)
    p[i] = xrandom() % m;

  return p;
}

/* Compute minimum number of recombinations needed for current state */
static Expression *_rmin()
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));
  HashTable *t;

  if (_score_rmin < 0){
    /* We haven't computed r_min for this configuration yet */
    t = beagle_allocate_hashtable(_score_state, -1);
    _score_rmin = beagle_reusable(_score_state, NULL, t);
    beagle_deallocate_hashtable(t);
  }

  /* Store r_min in expression */
  e->type = EXPRESSION_NUMBER;
  e->content.number = _score_rmin;

  return e;
}

/* Compute Hudson-Kaplan lower bound for current state */
static Expression *_hk()
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));

  if (_score_hk < 0)
    /* We haven't computed Hudson-Kaplan lower bound for this
     * configuration yet.
     */
    _score_hk = hudson_kaplan_genes(_score_state);

  /* Store lower bound in expression */
  e->type = EXPRESSION_NUMBER;
  e->content.number = _score_hk;

  return e;
}

/* Compute haplotype lower bound with the heuristic parameters
 * specified by p.
 */
static Expression *_hb(Expression **p)
{
  int i;
  void **a = (void **)xmalloc(4 * sizeof(void *));
  Expression *e = (Expression *)xmalloc(sizeof(Expression));

  /* Create array specifying this function call */
  for (i = 0; i < 3; i++){
    if (p[i]->type != EXPRESSION_NUMBER){
      /* Type error */
      e->type = EXPRESSION_ERROR;
      e->content.error = strdup("Non-numerical argument (argument 0) passed to haplotype bound computation");
      e->content.error[33] = '1' + i;
      free(a);
      return e;
    }
    a[i + 1] = (void *)(int)p[i]->content.number;
    if ((int)a[i + 1] <= 0)
      a[i + 1] = (void *)INT_MAX;
  }
  a[0] = (void *)_hb;

  /* Check whether this function call has previously been invoked; if
   * not, compute and store value.
   */
  if (!hashtable_lookup(a, _score_functioncalls, (void **)&i)){
    i = haplotype_heuristic_genes(_score_state, (int)a[1], (int)a[2],
				  (int)a[3]);
    hashtable_insert(a, (void *)i, _score_functioncalls);
  }

  /* Store lower bound in expression */
  e->type = EXPRESSION_NUMBER;
  e->content.number = i;

  return e;
}

/* Compute lower bound from local exact minimum number of
 * recombinations combined using the composite method.
 */
static Expression *_eagl(Expression **p)
{
  void **a = (void **)xmalloc(2 * sizeof(void *));
  Expression *e = (Expression *)xmalloc(sizeof(Expression));
  int b, **B;
  Sites *s;

  /* Create array for specifying this function call */
  a[0] = (void *)_eagl;
  if (p[0]->type != EXPRESSION_NUMBER){
    /* Type error */
    e->type = EXPRESSION_ERROR;
    e->content.error = strdup("Non-numerical argument passed to eagl bound computation");
    free(a);
    return e;
  }
  a[1] = (void *)(int)p[0]->content.number;

  /* Check whether this function call has previously been invoked; if
   * not, compute and store value.
   */
  if (!hashtable_lookup(a, _score_functioncalls, (void **)&b)){
    s = genes2sites(_score_state);
    B = hudson_kaplan_local(s);
    free_sites(s);
    b = eagl(_score_state, (int)a[1], B, NULL, NULL);
    hashtable_insert(a, (void *)b, _score_functioncalls);
  }

  /* Store lower bound in expression */
  e->type = EXPRESSION_NUMBER;
  e->content.number = b;

  return e;
}

/* Compute base-10 logarithm */
static Expression *_log(Expression **p)
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));

  if (p[0]->type != EXPRESSION_NUMBER){
    e->type = EXPRESSION_ERROR;
    e->content.error = strdup("Non-numerical argument passed to log");
  }
  else{
    e->type = EXPRESSION_NUMBER;
    e->content.number = log10(p[0]->content.number);
  }

  return e;
}

/* Compute natural logarithm */
static Expression *_ln(Expression **p)
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));

  if (p[0]->type != EXPRESSION_NUMBER){
    e->type = EXPRESSION_ERROR;
    e->content.error = strdup("Non-numerical argument passed to log");
  }
  else{
    e->type = EXPRESSION_NUMBER;
    e->content.number = log(p[0]->content.number);
  }

  return e;
}

/* Compute sine */
static Expression *_sin(Expression **p)
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));

  if (p[0]->type != EXPRESSION_NUMBER){
    e->type = EXPRESSION_ERROR;
    e->content.error = strdup("Non-numerical argument passed to log");
  }
  else{
    e->type = EXPRESSION_NUMBER;
    e->content.number = sin(p[0]->content.number);
  }

  return e;
}

/* Compute cosine */
static Expression *_cos(Expression **p)
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));

  if (p[0]->type != EXPRESSION_NUMBER){
    e->type = EXPRESSION_ERROR;
    e->content.error = strdup("Non-numerical argument passed to log");
  }
  else{
    e->type = EXPRESSION_NUMBER;
    e->content.number = cos(p[0]->content.number);
  }

  return e;
}

/* Compute tangent */
static Expression *_tan(Expression **p)
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));

  if (p[0]->type != EXPRESSION_NUMBER){
    e->type = EXPRESSION_ERROR;
    e->content.error = strdup("Non-numerical argument passed to log");
  }
  else{
    e->type = EXPRESSION_NUMBER;
    e->content.number = tan(p[0]->content.number);
  }

  return e;
}

/* Compute arc sine */
static Expression *_asin(Expression **p)
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));

  if (p[0]->type != EXPRESSION_NUMBER){
    e->type = EXPRESSION_ERROR;
    e->content.error = strdup("Non-numerical argument passed to log");
  }
  else{
    e->type = EXPRESSION_NUMBER;
    e->content.number = asin(p[0]->content.number);
  }

  return e;
}

/* Compute arc cosine */
static Expression *_acos(Expression **p)
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));

  if (p[0]->type != EXPRESSION_NUMBER){
    e->type = EXPRESSION_ERROR;
    e->content.error = strdup("Non-numerical argument passed to log");
  }
  else{
    e->type = EXPRESSION_NUMBER;
    e->content.number = acos(p[0]->content.number);
  }

  return e;
}

/* Compute arc tangent */
static Expression *_atan(Expression **p)
{
  Expression *e = (Expression *)xmalloc(sizeof(Expression));

  if (p[0]->type != EXPRESSION_NUMBER){
    e->type = EXPRESSION_ERROR;
    e->content.error = strdup("Non-numerical argument passed to log");
  }
  else{
    e->type = EXPRESSION_NUMBER;
    e->content.number = atan(p[0]->content.number);
  }

  return e;
}

/* Hash the function call information stored in elm using the
 * parameters in p.
 */
static unsigned long _score_hash(void **elm, int *p)
{
  int i;
  unsigned long v = 0;

  if (*elm == _hb)
    for (i = 1; i <= 3; i++)
      v += p[i + 1] * (int)elm[i];
  else
    v = p[1] + p[2] * (int)elm[1];

  return v % p[0];
}

/* Compare the two function calls a and b */
static int _score_compare(void **a, void **b)
{
  int i;

  if (*a != *b)
    return 0;

  if (*a == _hb){
    for (i = 1; i <= 3; i++)
      if ((int)a[i] != (int)b[i])
	return 0;
  }
  else
    return (int)a[1] == (int)b[1];

  return 1;
}

/* At face value, it may appear that this function has been
 * misplaced. It does not compute the exact number of recombinations
 * required for g, but rather builds a plausible evolutionary history
 * in greedy fashion by choosing events that seem to reduce the
 * complexity of the data set as specified by score (that assigns a
 * score to each next configuration - the number of recombinations in
 * the step leading to this configuration is passed as second argument
 * to this function) and select (that chooses an index from the array
 * of scores passed to it). However, it shares a significant amount of
 * functionality with beagle and is thus placed here. If
 * print_progress is not NULL, a hash mark is output to this file for
 * every recombination used.
 */
static double _score(Genes *g, char *score)
{
  double recombinations, am, minam, maxam, seq, minseq, maxseq, len, minlen,
    maxlen, pi = M_PI, e = M_E, s;
  Program *prg = expression_initialise();
  void *builtin_variables[][2] = {
    { "am", &am },
    { "ancestralmaterial", &am },
    { "r", &recombinations },
    { "recombinations", &recombinations },
    { "minam", &minam },
    { "minimumancestralmaterial", &minam },
    { "maxam", &maxam },
    { "maximumancestralmaterial", &maxam },
    { "seq", &seq },
    { "numberofsequences", &seq },
    { "minseq", &minseq },
    { "minimumnumberofsequences", &minseq },
    { "maxseq", &maxseq },
    { "maximumnumberofsequences", &maxseq },
    { "len", &len },
    { "numberofsites", &len },
    { "minlen", &minlen },
    { "minimumnumberofsites", &minlen },
    { "maxlen", &maxlen },
    { "maximumnumberofsites", &maxlen },
    { "e", &e },
    { "pi", &pi },
    { NULL, NULL}};
  void *builtin_functions[][3] = {
    { "Rmin", _rmin, (void *)0 },
    { "rmin", _rmin, (void *)0 },
    { "beagle", _rmin, (void *)0},
    { "Beagle", _rmin, (void *)0},
    { "hk", _hk, (void *)0 },
    { "HK", _hk, (void *)0 },
    { "hudsonkaplan", _hk, (void *)0 },
    { "HudsonKaplan", _hk, (void *)0 },
    { "hb", _hb, (void *)3 },
    { "HB", _hb, (void *)3 },
    { "haplotypebound", _hb, (void *)3 },
    { "HaplotypeBound", _hb, (void *)3 },
    { "eagl", _eagl, (void *)1 },
    { "Eagl", _eagl, (void *)1 },
    { "log", _log, (void *)1 },
    { "ln", _ln, (void *)1 },
    { "sin", _sin, (void *)1 },
    { "cos", _cos, (void *)1 },
    { "tan", _tan, (void *)1 },
    { "asin", _asin, (void *)1 },
    { "acos", _acos, (void *)1 },
    { "atan", _atan, (void *)1 },
    { NULL, NULL, (void *)0 }};
  
#ifdef ENABLE_VERBOSE
  int v = verbose();
  set_verbose(0);
#endif

  /* Set up builtin functions and variables for scoring a configuration */
  expression_buildinvariables(prg, builtin_variables);
  expression_buildinfunctions(prg, builtin_functions);
  _score_functioncalls = hashtable_new
    (6, (unsigned long (*)(void *, void *))_score_hash,
     (int (*)(void *, void *))_score_compare,
     (void *(*)(unsigned long))_score_initparam);
  _score_rmin = _score_hk = -1;
  _score_state = g;
  minam = maxam = am = ancestral_material(g);
  minseq = maxseq = seq = g->n;
  minlen = maxlen = len = g->length;
  recombinations = 0;

  /* Parse expression for scoring putative previous ancestral states */
  expression_parse(prg, score);
  s = expression_interpret(prg);

  /* Clean up */
  hashtable_destroy(_score_functioncalls, free, NULL, free);
  expression_destroy(prg, NULL, NULL, NULL);

  return s;
}

int main(int argc, char **argv)
{
  Genes *g;
  AnnotatedGenes *a;
  int i,
    reduce = 0,
    score_from_file = 0;
  Gene_Format format = GENE_ANY;
  Gene_SeqType seqtype = GENE_BINARY;
  char *score = NULL;

  /* Initialise random number generator */
  initialise_xrandom();

#ifdef ENABLE_VERBOSE
  set_verbose(1);
#endif

  /* Analyse command line options */
#define KWARG_OPTIONS "kofanE:F::rhH?"

  /* Parse command line options */
  while ((i = getopt(argc, argv, KWARG_OPTIONS)) >= 0){
    switch(i){
    case 'k':
      gene_knownancestor = 1;
      break;
    case 'o':
      format = GENE_BEAGLE;
      break;
    case 'f':
      format = GENE_FASTA;
      break;
    case 'a':
      seqtype = GENE_AMINO;
      break;
    case 'n':
      seqtype = GENE_NUCLEIC;
      break;
    case 'E':
      if (score_from_file){
	free(score);
	score_from_file = 0;
      }
      score = optarg;
      break;
    case 'F':
      if (score_from_file)
	free(score);
      if ((score = _read_file(optarg == 0 ? NULL : optarg)) == NULL){
	_print_usage(stderr, argv[0]);
	fprintf(stderr, "\nCould not open file %s for reading\n", optarg);
	exit(1);
      }
      score_from_file = 1;
      break;
    case 'r':
      reduce = 1;
      break;
    case 'h':
    case 'H':
    case '?':
      _print_usage(stdout, argv[0]);
      /* Clean up */
      if (score_from_file)
	free(score);
      exit(0);
    case ':':
      _print_usage(stderr, argv[0]);
      exit(1);
    }
  }

  /* Set default scoring scheme if none has been specified */
  if (score == NULL)
    score = KWARG_DEFAULTSCORE;
  /* A positive score indicates that scores should be treated as
   * pseudoenergies and next configurations selected from the
   * corresponding Boltzman distribution.
   */

  /* Read data */
  if (argc > optind){
    if ((a = read_genes(argv[optind], format, seqtype)) == NULL){
      fprintf(stderr, "Could not parse %s as valid data\n", argv[optind]);
      exit(1);
    }
  }
  else{
    if ((a = read_genes(NULL, format, seqtype)) == NULL){
      fprintf(stderr, "Could not parse input as valid data\n");
      exit(1);
    }
  }
  if ((gene_knownancestor) && (seqtype != GENE_BINARY))
    /* First sequnce only included to specify known common ancestor */
    remove_annotatedgene(a, 0);
  g = a->g;
  if (reduce)
    implode_genes(g);

#ifdef ENABLE_VERBOSE
  output_annotatedgenes(a, NULL, NULL);
#endif
  printf("Score is %lf\n", _score(g, score));

  /* Clean up */
  free_annotatedgenes(a);
  if (score_from_file)
    free(score);

  return 0;
}
